// Portable implementation of the bipartite table method for a signaling
// less-than-or-equal comparison between a binary64 and a decimal64
// 
// Copyright (C) 2012, 2013, 2014 by
// 
// Laboratoire de l'Informatique du Parallelisme,
// UMR CNRS - ENS Lyon - UCB Lyon 1 - INRIA 5668
// 
// and by
// 
// Laboratoire d'Informatique de Paris 6, equipe PEQUAN,
// UPMC Universite Paris 06 - CNRS - UMR 7606 - LIP6, Paris, France.
// 
// Authors : Christoph Lauter, Marc Mezzarobba
// 
// Contact : christoph.lauter@lip6.fr, marc@mezzarobba.net
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
// 

#include <stdio.h>
#include <stdint.h>

/* Get our constants and types */
#include "comparison.h"
#include <math.h>

/* Get the table we need for the comparison */
#include "comparison-table-alt-new.h"

/* Compile me with something like 

    icc -xSSE4.2 -O3 -c -o comparison-alt.o comparison-alt.c 

*/

#define multiplyUint64ByUint64(multiplyUint64ByUint64_rh,                                                         \
                               multiplyUint64ByUint64_rl,                                                         \
                               multiplyUint64ByUint64_a,                                                          \
                               multiplyUint64ByUint64_b) do {                                                     \
  uint64_t multiplyUint64ByUint64_ah;                                                                             \
  uint64_t multiplyUint64ByUint64_al;                                                                             \
  uint64_t multiplyUint64ByUint64_bh;                                                                             \
  uint64_t multiplyUint64ByUint64_bl;                                                                             \
  uint64_t multiplyUint64ByUint64_resl;                                                                           \
  uint64_t multiplyUint64ByUint64_resm1;                                                                          \
  uint64_t multiplyUint64ByUint64_resm2;			                                                  \
  uint64_t multiplyUint64ByUint64_resm;                                                                           \
  uint64_t multiplyUint64ByUint64_resmh;                                                                          \
  uint64_t multiplyUint64ByUint64_resml;                                                                          \
  uint64_t multiplyUint64ByUint64_resh;	                                                                          \
  multiplyUint64ByUint64_ah = (multiplyUint64ByUint64_a) >> 32;                                                   \
  multiplyUint64ByUint64_al = (multiplyUint64ByUint64_a) & 0x00000000ffffffffull;                                 \
  multiplyUint64ByUint64_bh = (multiplyUint64ByUint64_b) >> 32;                                                   \
  multiplyUint64ByUint64_bl = (multiplyUint64ByUint64_b) & 0x00000000ffffffffull;                                 \
  multiplyUint64ByUint64_resl = multiplyUint64ByUint64_al * multiplyUint64ByUint64_bl;                            \
  multiplyUint64ByUint64_resh = multiplyUint64ByUint64_ah * multiplyUint64ByUint64_bh;                            \
  multiplyUint64ByUint64_resm1 = multiplyUint64ByUint64_ah * multiplyUint64ByUint64_bl;                           \
  multiplyUint64ByUint64_resm2 = multiplyUint64ByUint64_al * multiplyUint64ByUint64_bh;                           \
  multiplyUint64ByUint64_resm = multiplyUint64ByUint64_resm1 + multiplyUint64ByUint64_resm2;                      \
  multiplyUint64ByUint64_resh += ((uint64_t) (multiplyUint64ByUint64_resm < multiplyUint64ByUint64_resm1)) << 32; \
  multiplyUint64ByUint64_resmh = multiplyUint64ByUint64_resm >> 32;                                               \
  multiplyUint64ByUint64_resml = multiplyUint64ByUint64_resm << 32;                                               \
  *(multiplyUint64ByUint64_rl) = multiplyUint64ByUint64_resl + multiplyUint64ByUint64_resml;                      \
  multiplyUint64ByUint64_resh += (uint64_t) (*(multiplyUint64ByUint64_rl) < multiplyUint64ByUint64_resl);         \
  *(multiplyUint64ByUint64_rh) = multiplyUint64ByUint64_resh + multiplyUint64ByUint64_resmh;                      \
} while (0) 

#define multiplyTwoUint64ByUint64(multiplyTwoUint64ByUint64_rh,                                                   \
                                  multiplyTwoUint64ByUint64_rm,                                                   \
                                  multiplyTwoUint64ByUint64_rl,                                                   \
                                  multiplyTwoUint64ByUint64_ah,                                                   \
                                  multiplyTwoUint64ByUint64_al,                                                   \
                                  multiplyTwoUint64ByUint64_b) do {                                               \
  uint64_t multiplyTwoUint64ByUint64_t, multiplyTwoUint64ByUint64_c;                                              \
  multiplyUint64ByUint64(&multiplyTwoUint64ByUint64_c, (multiplyTwoUint64ByUint64_rl),                            \
                         (multiplyTwoUint64ByUint64_al), (multiplyTwoUint64ByUint64_b));                          \
  multiplyUint64ByUint64((multiplyTwoUint64ByUint64_rh), &multiplyTwoUint64ByUint64_t,                            \
                         (multiplyTwoUint64ByUint64_ah), (multiplyTwoUint64ByUint64_b));                          \
  *(multiplyTwoUint64ByUint64_rm) = multiplyTwoUint64ByUint64_c + multiplyTwoUint64ByUint64_t;                    \
  *(multiplyTwoUint64ByUint64_rh) += (uint64_t) (*(multiplyTwoUint64ByUint64_rm) < multiplyTwoUint64ByUint64_c);  \
} while (0)

int compareSignalingLessEqualBinary64Decimal64(_binary64 x, _Decimal64 y) {
  bin64caster xcst;
  dec64caster ycst;
  volatile _binary64 a, b, c;
  int xIsInf, yIsInf, xIsZero, yIsZero, xIsNeg, yIsNeg;
  uint64_t ySignificand;
  int64_t E, F, G, S, H, cTwo19Log2OverLog5, Q, q, r, cTwo12Log5OverLog2, sigma;
  uint64_t m, n, t, theta1hi, theta1lo, theta2, lefth, leftl, righth, rightm, rightl;
  uint64_t resh, resl; 
  int64_t reshsigned;
  bin64caster temp;

  xcst.f = x;
  ycst.f = y;

  /* Handle NaNs first: for a signaling comparison for less-or-equal
     the result always is false and invalid is to be signaled. 
  */
  if (((xcst.i & 0x7fffffffffffffffull) > 0x7ff0000000000000ull) ||
      ((ycst.i & 0x7c00000000000000ull) == 0x7c00000000000000ull)) {
    /* At least one of x or y is NaN.
       Perform two dummy operations and return false.
    */
    a = 0.0;
    b = 0.0;
    c = a / b; /* Signals invalid */
    return 0;
  }

  /* Here, neither x nor y is NaN */

  /* Check if x or y are zero. Also determine their signs. */
  xIsNeg = ((xcst.i & 0x8000000000000000ull) != 0x0ull);
  yIsNeg = ((ycst.i & 0x8000000000000000ull) != 0x0ull);
  xIsZero = ((xcst.i & 0x7fffffffffffffffull) == 0x0ull);
  ySignificand = (((ycst.i & 0x6000000000000000ull) != 0x6000000000000000ull) ? 
		  (ycst.i & 0x001fffffffffffffull) : 
		  ((ycst.i & 0x0007ffffffffffffull) | 0x0020000000000000ull));
  if (ySignificand > 9999999999999999ull) ySignificand = 0x0ull;
  yIsZero = (ySignificand == 0x0ull);
  
  if (xIsZero || yIsZero) {
    /* Here, at least one of x or y is zero. */

    /* Start by checking if x and y are both zero */
    if (xIsZero && yIsZero) {
      /* Here, both x and y are zero. As we for a -or-equal comparison
	 the result is true, regardless of the sign of the zeros. 
      */
      return 1;
    }

    /* Here, only one of x or y is zero */
    if (xIsZero) {
      /* Here, x is zero and y is non-zero. The sign of y determines
	 the result: for a less-or-equal comparison, the result is
	 true iff y is not negative.
      */
      return !yIsNeg;
    }

    /* Here, y is zero and x is non-zero. The sign of x determines the
       result: for a less-or-equal comparison, the result is true iff
       x is negative.
    */
    return xIsNeg;
  }

  /* Here, x and y are both neither zero nor NaN. */
  
  /* Now perform a first check based on the sign of the input: 
     if the signs of x and y are different, the result is already known.
  */
  if (xIsNeg != yIsNeg) {
    /* Here, the signs of x and y are different.  The result is
       already known: the less-or-equal comparison is true iff x is
       negative (and y is positive).
    */
    return xIsNeg;
  }
  
  /* Here, x and y are both neither zero nor NaN and have the same
     sign. 
  */
  
  /* Now check if x or y is +/- Infinity */
  xIsInf = ((xcst.i & 0x7fffffffffffffffull) == 0x7ff0000000000000ull);
  yIsInf = ((ycst.i & 0x7c00000000000000ull) == 0x7800000000000000ull);
  if (xIsInf || yIsInf) {
    /* Here, at least one of x or y is infinity. We do know that they
       have the same signs. So we start by testing if they are both
       infinities.
    */
    if (xIsInf && yIsInf) {
      /* Here, both x and y are infinity. For a less-or-equal
	 comparison, the result is hence true.
      */
      return 1;
    }
    
    /* Here, at least one of x and y is a finite number, the other is
       an infinity.
    */
    if (xIsInf) {
      /* Here x is an infinity, y is finite. The result of the
	 less-than-equal comparison is true iff x is negative. 
      */
      return xIsNeg;
    }

    /* Here x is finite and non-zero, y is an infinity. The result of
       the less-than-equal comparison is true iff y is not negative.
    */
    return !yIsNeg;
  }

  /* Here, x and y are both non-zero finite numbers of the same
     sign. 
  */

  /* We start by decomposing |x| into 

     2^E * m = |x|

     with E integer and m integer satisfying 2^52 <= m <= 2^53 - 1.

  */
  xcst.i &= 0x7fffffffffffffffull;
  E = 0;
  if (xcst.i < 0x0010000000000000ull) {
    /* Handle subnormal numbers */
    E -= 54;
    xcst.f *= 18014398509481984.0; /* Multiply with 2^54 */
  }
  E += ((xcst.i & 0x7ff0000000000000ull) >> 52) - 1023 - 52;
  m = (xcst.i & 0x000fffffffffffffull) | 0x0010000000000000ull;
  
  /* Now we decompose |y| into 

     2^F * 5^G * n = |y|

     with F integer and n integer satisfying 2^53 <= m <= 2^54 - 1.

     We already know the binary encoded significand ySignificand that
     satisfies:

     1 <= ySignificand <= 10^16 - 1.

     We have 10^16 - 1 \approx 2^53.151 hence 2^53 < 10^16 - 1 < 2^54.

     We first compute G such that 

     10^G * ySignificand = |y|

     Then we use the binary FP unit as a leading zero counter in order
     to normalize ySignificand for F and n, where 2^53 <= n <=
     2^54 - 1.

  */
  G = (((ycst.i & 0x6000000000000000ull) != 0x6000000000000000ull) ? 
       ((ycst.i & 0x7fe0000000000000ull) >> 53) : 
       ((ycst.i & 0x1ff8000000000000ull) >> 51)) - 398;

  t = ySignificand >> 32;
  S = 0;
  if (t == 0ull) {
    t = ySignificand;
    S = 32;
  } 
  temp.i = 0x4330000000000000 | t; /* 0x43300...0 = 2^52 */
  temp.f -= 4503599627370496.0;    /* 2^52 */
  S += (53 - 32) - ((temp.i >> 52) - 1023);
  n = ySignificand << S;
  F = G - S;

  /* Now we have to compare 2^E * m with 2^F * 5^G * n with 

     E, F, G integers

     and 

     2^52 <= m <= 2^53 - 1

     as well as

     2^53 <= n <= 2^54 - 1.

     This is equivalent to comparing 

     2^(E - F) * m with 5^G * n.

     Let H be H = E - F.
  */
  H = E - F;

  /* We can eliminate quite a few cases considering only the
     "exponents" G and H due to the following two implications we can
     prove:

     For m and n normalized as indicated, we have

     (1) G > floor(H * log(2)/log(5)) => 5^G * n > 2^H * m
     (2) G < floor(H * log(2)/log(5)) => 5^G * n < 2^H * m

     So we have to compute the quantity 

     Q = floor(H * log(2)/log(5))

     We can show that H is an integer in the range 

     -1495 <= H <= 1437

     and that for all these integers, we have

     floor(H * log(2)/log(5)) = floor(2^(-19) * H * 
                                     nearestint(2^19 * log(2)/log(5))).
     
     Let be

     cTwo19Log2OverLog5 = nearestint(2^19 * log(2)/log(5)) = 225799.

     So we can compute Q by a simple multiplication followed by a shift.

     It is easy to show that there is no integer overflow for at least
     signed 32 bit integers.

  */
  cTwo19Log2OverLog5 = 225799;
  Q = (H * cTwo19Log2OverLog5) >> 19;

  /* Now check if G < Q or G > Q or if we have equality */
  if (Q != G) {
    /* Here we have either Q < G or Q > G */
    if (Q < G) {
      /* Here we have Q < G hence 5^G * n > 2^H * m.

	 So we know that 

	 abs(x) < abs(y)   and sgn(x) = sgn(y)

	 which means that the less-or-equal comparison x <= y 
	 is true iff x and y are not negative.

      */
      return !xIsNeg;
    } 
    
    /* Here we have Q > G hence 5^G * n < 2^H * m 

       So we know that 

       abs(x) > abs(y)   and sgn(x) = sgn(y)

       which means that the less-or-equal comparison x <= y 
       is true iff x and y are negative.
       
    */
    return xIsNeg;
  } 

  q = -((-G) >> 4);
  r = (q << 4) - G;
  cTwo12Log5OverLog2 = 9511ll;
  sigma = ((r * cTwo12Log5OverLog2) >> 12) + H - ((q * cTwo12Log5OverLog2) >> 8);

  m <<= sigma+7;
  n <<= 10;

  theta1hi = comparisonTblTheta1Hi[q - qmin];
  theta1lo = comparisonTblTheta1Lo[q - qmin];
  theta2   = comparisonTblTheta2[r - rmin];

  multiplyTwoUint64ByUint64(&righth, &rightm, &rightl, theta1hi, theta1lo, n);
  multiplyUint64ByUint64(&lefth, &leftl, theta2, m);

  resl = rightm - leftl; /* Account for borrow afterwards. */
  resh = righth - lefth; 

  /* Account for borrow between low-order and high-order words. */
  if (leftl > rightm) resh--;

  /* Get a signed-variable representation of resh. */
  reshsigned = *((int64_t *) &resh);

  /* 2^64 * resh + resl is negative iff reshsigned is negative */
  if (reshsigned < 0ll) {
    /* In this case, we have 2^H * m > 5^G * n, i.e. abs(x) > abs(y).
       We know that sgn(x) = sgn(y).

       For a less-or-equal comparison x <= y, the result is true iff 
       x (and y) is negative.
    */

    return xIsNeg;
  }

  /* If we are here, we have 2^H * m <= 5^G * n, i.e. abs(x) <= abs(y). 
     We know that sgn(x) = sgn(y).   

     We continue by filtering the exact case, i.e. when x = y.
  */
  if ((resh == 0ull) && (resl == 0ull)) {
    /* In this case, we have 2^H * m = 5^G * n, i.e. abs(x) = abs(y).
       We know that sgn(x) = sgn(y). This means, we have x = y.

       For a less-or-equal comparison x <= y, the result is true.

    */

    return 1;
  } 

  /* If we are here, we have 2^H * m < 5^G * n, i.e. abs(x) < abs(y). 
     We know that sgn(x) = sgn(y).   

     For a less-or-equal comparison x <= y, the result is true iff 
     x (and y) is not negative.
  */
  
  return !xIsNeg;
}


/* An empty dummy function in this translation unit for timing purposes */
int emptyFunctionBinary64Decimal64(_binary64 x, _Decimal64 y) {
  return 1;
}
