// Optimized implementation of the bipartite table method for a signaling
// less-than-or-equal comparison between a binary64 and a decimal64
// 
// Copyright (C) 2012, 2013, 2014 by
// 
// Laboratoire de l'Informatique du Parallelisme,
// UMR CNRS - ENS Lyon - UCB Lyon 1 - INRIA 5668
// 
// and by
// 
// Laboratoire d'Informatique de Paris 6, equipe PEQUAN,
// UPMC Universite Paris 06 - CNRS - UMR 7606 - LIP6, Paris, France.
// 
// Authors : Christoph Lauter, Marc Mezzarobba
// 
// Contact : christoph.lauter@lip6.fr, marc@mezzarobba.net
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
// 

#include <stdio.h>
#include <stdint.h>
#include <math.h>

/* Get our constants and types */
#include "comparison.h"

/* Get the table we need for the comparison */
#include "comparison-table-alt-opt-new.h"

/* Compile me with something like 

    icc -xSSE4.2 -O3 -c -o comparison-alt.o comparison-alt.c 

*/

#define uint64rptr uint64_t* __restrict__
#define uint128rptr __uint128_t* __restrict__

static inline void
mul64_64_128(uint64rptr rh, uint64rptr rl, uint64_t a, uint64_t b) {
  __uint128_t tmp = ((__uint128_t) a) * ((__uint128_t) b);
  *rl = (uint64_t) tmp;
  tmp >>= 64;
  *rh = (uint64_t) (tmp);
}

static inline void
mul64_128_192(uint128rptr rhm, uint64rptr rl, uint64_t ah,
    uint64_t al, uint64_t b) {
  uint64_t c;
  mul64_64_128(&c, rl, al, b);
  *rhm = ((__uint128_t) ah) * ((__uint128_t) b) + c;
}

#define likely(COND) __builtin_expect(COND, 1)
#define unlikely(COND) __builtin_expect(COND, 0)

int compareSignalingLessEqualBinary64Decimal64(_binary64 x, _Decimal64 y) {
  bin64caster xcst;
  dec64caster ycst;
  volatile _binary64 a, b, c;
  int_fast8_t xIsInf, yIsInf, xIsZero, yIsZero, xIsNeg, yIsNeg, xIsNaN, yIsNaN;
  int_fast8_t S;
  int_fast16_t sigma, q, r, E, F, G, H;
  int_fast32_t cTwo19Log2OverLog5, cTwo12Log5OverLog2, Q;
  uint_fast32_t yc;
  uint64_t ySignificand;
  uint64_t m, n, t, theta1hi, theta1lo, theta2, lefth, leftl, righth, rightm, rightl;
  uint64_t resh, resl; 
  int64_t reshsigned;
  bin64caster temp;

  xcst.f = x;
  ycst.f = y;

  xIsNeg = ((xcst.i & 0x8000000000000000ull) != 0x0ull);

  xcst.i &= 0x7fffffffffffffffull;

  uint_fast16_t yc0 = ycst.i >> 48;
  yIsNeg = ((yc0 & 0x8000) != 0);

  /* Handle NaNs and infinities */
  if (unlikely((xcst.i >= 0x7ff0000000000000ull) || ((yc0 & 0x78ff) == 0x7800))) {
    yc = (yc0 & 0x7c00);
    yIsInf = (yc == 0x7800);
    yIsNaN = (yc == 0x7c00);
    xIsNaN = (xcst.i > 0x7ff0000000000000ull);
    xIsInf = (xcst.i == 0x7ff0000000000000ull);
    if (xIsNaN || yIsNaN) {
      a = 0.0;
      b = 0.0;
      c = a / b; /* Signals invalid */
      return 0;
    }
    if (xIsNeg != yIsNeg)
      return xIsNeg;
    if (xIsInf && yIsInf)
      return 1;
    if (xIsInf)
      return xIsNeg;
    return !yIsNeg;
  }

  xIsZero = (xcst.i == 0x0ull);
  if ((yc0 & 0x6000) != 0x6000) {
    ySignificand = ycst.i & 0x001fffffffffffffull;
    G = ((ycst.i >> 53) & ((uint_fast16_t) (0x7fe0000000000000ull >> 53)));
  } else {
    ySignificand = ((ycst.i & 0x0007ffffffffffffull) | 0x0020000000000000ull);
    G = ((ycst.i >> 51) & ((uint_fast16_t) (0x1ff8000000000000ull >> 51)));
  }
  G -= 398;

  yIsZero = (ySignificand == 0x0ull) || (ySignificand > 9999999999999999ull);

  if (unlikely(xIsZero || yIsZero)) {
    if (xIsZero && yIsZero) {
      return 1;
    }
    if (xIsZero) {
      return !yIsNeg;
    }
    return xIsNeg;
  }

  if (xIsNeg != yIsNeg) {
    return xIsNeg;
  }

  /* Here, x and y are both non-zero finite numbers of the same sign.
     We start by decomposing |x| into 2^E * m = |x| with E integer and m
     integer satisfying 2^52 <= m <= 2^53 - 1. */

  E = -1023 -52;
  m = (xcst.i & 0x000fffffffffffffull);
  if (unlikely(xcst.i == m)) {
    /* Handle subnormal numbers */
    unsigned int shift = __builtin_clzl(m) - 11;
    m <<= shift;
    E -= shift-1;
  }
  E += ((xcst.i & 0x7ff0000000000000ull) >> 52);
  m |= 0x0010000000000000ull;

  S = __builtin_clzl(ySignificand) - 10;
  n = ySignificand << S;
  F = G - S;

  H = E - F;

  cTwo19Log2OverLog5 = 225799;
  Q = (H * cTwo19Log2OverLog5) >> 19;

  if (likely(Q != G)) {
    if (Q < G) {
      return !xIsNeg;
    }
    return xIsNeg;
  }

  q = -((-G) >> 4);
  r = (q << 4) - G;
  cTwo12Log5OverLog2 = 9511;

  sigma = ((r * cTwo12Log5OverLog2) >> 12) + H - ((q * cTwo12Log5OverLog2) >> 8);

  m <<= 7 + sigma;
  n <<= 10;

  theta1hi = comparisonTblTheta1[2*(q - qmin)];
  theta1lo = comparisonTblTheta1[2*(q - qmin)+1];
  theta2   = comparisonTblTheta2[r - rmin];

  __uint128_t left = ((__uint128_t) theta2) * ((__uint128_t) m);

  __uint128_t righthm;
  mul64_128_192(&righthm, &rightl, theta1hi, theta1lo, n);

  __uint128_t res = righthm - left;
  resh = res >> 64;

  /* Get a signed-variable representation of resh. Remark: on the register
   * level, this operation is a no-op.  It's just needed to convince the
   * compiler to compare with zero in the next step. */
  reshsigned = *((int64_t *) &resh);

  /* 2^64 * resh + resl is negative iff reshsigned is negative */
  if (reshsigned < 0ll) {
    return xIsNeg;
  }

  /* If we are here, we have 2^H * m <= 5^G * n, i.e. abs(x) <= abs(y). We know
   * that sgn(x) = sgn(y).   We continue by filtering the exact case, i.e. when
   * x = y. */
  if (unlikely(res == (__uint128_t) 0)) {
    return 1;
  }
  return !xIsNeg;
}


/* An empty dummy function in this translation unit for timing purposes */
int emptyFunctionBinary64Decimal64(_binary64 x, _Decimal64 y) {
  return 1;
}
