#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#include <fenv.h>

#include <sched.h>

/* Get our constants and types */
#include "comparison.h"

#include <math.h>

/* Get the helper functions */
#include "comparison-helper.h"

/* Some constants */
#define NUM_ELEMENTS_TESTVECTOR 8092
#define READ_BUFFER_SIZE 1024

#define CACHE_HEAT_RUNS 256
#define TIMING_RUNS 1024
#define MAX_EXPECTED_TIMING 999
#define MAX_EXPECTED_FUNCTION_CALL_TIMING 999
#define FUNCTION_CALL_TIMING_REPS 16 

#define FLAG_PRECISION       1
#define FLAG_UNDERFLOW       2
#define FLAG_OVERFLOW        4
#define FLAG_DIVIDE_BY_ZERO  8
#define FLAG_INVALID        16
#define FLAG_ALL_SET (                      \
                      FLAG_PRECISION      | \
                      FLAG_UNDERFLOW      | \
                      FLAG_OVERFLOW       | \
                      FLAG_DIVIDE_BY_ZERO | \
                      FLAG_INVALID          \
		     )

#define PARSE_ERROR 0
#define EMPTY_LINE  1
#define PARSE_OKAY  2

/* A type for the testvector */
typedef struct {
  _binary64 x;
  _Decimal64 y;
  int expectedResult;
  int actualResult;
  _binary64 computeTime;
  _binary64 computeTimeWithoutFuncCall;
  int beforeFlags;
  int actualAfterFlags;
  int expectedAfterFlags;
} testvector_t;

/* Types for the time analytics */
typedef struct {
  _binary64 minTiming;
  _binary64 minTimingXArg;
  _Decimal64 minTimingYArg;
  _binary64 maxTiming;
  _binary64 maxTimingXArg;
  _Decimal64 maxTimingYArg;
  _binary64 avgTiming;
  int avgHits;
} timeanalytics_class_t;

typedef struct {
  timeanalytics_class_t class[BINARY64_NUMBER_CLASSES * DECIMAL64_NUMBER_CLASSES];
  int avgFunctionCallHits;
  _binary64 avgFunctionCallTiming;
  _binary64 minFunctionCallTiming;
  _binary64 maxFunctionCallTiming;
  timeanalytics_class_t extraClass[EXTRA_CLASSES];
} timeanalytics_t;

/* An assembly macro to get the time counter */

/* time must be a pointer to an unsigned long long int */
#define READ_TIME_COUNTER(time)                              \
        __asm__ __volatile__(                                \
                "xorl %%eax,%%eax\n\t"                       \
                "cpuid\n\t"                                  \
                "rdtsc\n\t"                                  \
                "movl %%eax,(%0)\n\t"                        \
                "movl %%edx,4(%0)\n\t"                       \
                "xorl %%eax,%%eax\n\t"                       \
                "cpuid\n\t"                                  \
                : /* nothing */                              \
		: "S"((time))				     \
                : "eax", "ebx", "ecx", "edx", "memory")


/* Reader, helper and testing functions */

void *mycalloc(size_t nmemb, size_t size) {
  void *ptr;

  ptr = calloc(nmemb, size);
  if (ptr == NULL) {
    fprintf(stderr,"Could not allocate any additional memory; exiting\n");
    exit(17);
  } 

  return ptr;
}


char *readTestVectorLine(FILE *fd, char* buf, size_t bufSize, char **rest, size_t *restSize) {
  char *line = NULL;
  size_t lineSize = 0;
  size_t lineUsed = 0;
  char *curr;
  char *temp;
  size_t elementsRead;
  int ferr;
  size_t i;
  
  if (*restSize > 0) {
    line = (char *) mycalloc(*restSize + 1, sizeof(char));
    lineSize = *restSize + 1;
    lineUsed = 0;
    curr = line; 
    for (i=0;i<*restSize;i++) {
      if ((*rest)[i] == '\n') {
	*rest = &((*rest)[i + 1]);
	*restSize -= (i + 1);
	return line;
      }
      *curr = (*rest)[i];
      curr++;
      lineUsed++;
    }
  }

  while (1) {
    clearerr(fd);
    elementsRead = fread(buf, sizeof(char), bufSize, fd);
    ferr = ferror(fd);
    if (elementsRead != 0) {
      if (line == NULL) {
	line = (char *) mycalloc(elementsRead + 1, sizeof(char));
	lineSize = elementsRead + 1;
	lineUsed = 0;
	curr = line; 
      }
      for (i=0;i<elementsRead;i++) {
	if (buf[i] == '\n') {
	  *rest = &(buf[i + 1]);
	  *restSize = elementsRead - (i + 1);
	  return line;
	}
	if (lineUsed > lineSize - 2) {
	  temp = (char *) mycalloc(lineSize * 2 + 1, sizeof(char));
	  lineSize = lineSize * 2 + 1;
	  memcpy(temp, line, lineUsed);
	  free(line);
	  line = temp;
	  curr = line + lineUsed;
	}
	*curr = buf[i];
	curr++;
	lineUsed++;
      }
    } 
    if (elementsRead < bufSize) {
      if (feof(fd)) {
	break;
      } 
      /* Here, we could not read enough characters but we haven't reached eof */
      if (ferr != 0) {
	fprintf(stderr, "Some I/O error occurred while reading in the test vector file; exiting.\n");
	exit(17);
      }
      /* Hugh, no error occurred but we could read enough? Let's try again! */
    }
  }
  
  return line;
}

int tryReadHexadecimal(uint64_t *res, char *str) {
  uint64_t acc, digit, temp;
  char *curr;
  
  if (*str == '\0') return 0; 

  acc = 0ull;
  for (curr=str;*curr!='\0';curr++) {
    if ((*curr >= '0') && (*curr <= '9')) {
      digit = (uint64_t) (*curr - '0');
    } else {
      if ((*curr >= 'a') && (*curr <= 'f')) {
	digit = (uint64_t) ((*curr - 'a') + 0xa);
      } else {
	if ((*curr >= 'A') && (*curr <= 'F')) {
	  digit = (uint64_t) ((*curr - 'A') + 0xa);
	} else {
	  return 0;
	}
      }
    }
    
    /* Integrate the digit into the accumulator
       stop when there is any overflow 
    */
    temp = acc << 4;
    if ((temp >> 4) == acc) {
      temp += digit;
      if (temp >= digit) {
	acc = temp;
      } else {
	return 0;
      }
    } else {
      return 0;
    }
  }

  *res = acc;
  return 1;
}

int tryReadDecimal(uint64_t *res, char *str) {
  uint64_t acc, digit, temp1, temp2;
  char *curr;

  if (*str == '\0') return 0; 
  
  acc = 0ull;
  for (curr=str;*curr!='\0';curr++) {
    if ((*curr >= '0') && (*curr <= '9')) {
      digit = (uint64_t) (*curr - '0');
    } else {
      return 0;
    }
    
    /* Integrate the digit into the accumulator
       stop when there is any overflow 

       acc = acc * 10 + digit

       acc = acc * 8 + acc * 2 + digit

    */
    temp1 = acc << 3;
    temp2 = acc << 1;
    if (((temp1 >> 3) == acc) && 
	((temp2 >> 1) == acc)) {
      temp2 += digit;
      if (temp2 >= digit) {
	temp1 += temp2;
	if (temp1 >= temp2) {
	  acc = temp1;
	} else {
	  return 0;
	}
      } else {
	return 0;
      }
    } else {
      return 0;
    }
  }

  *res = acc;
  return 1;
}

int tryReadDecimalSigned(int *res, char *str) {
  int negative;
  int acc;
  uint64_t abs, bound;
  int64_t temp;
  char *curr;

  if (str[0] == '-') {
    negative = 1;
    curr = str + 1;
  } else {
    negative = 0;
    curr = str;
  }

  if (!tryReadDecimal(&abs, curr)) {
    return 0;
  }

  if (negative) {
    temp = INT_MIN;
    temp = -temp;
    bound = temp;
    if (abs > bound) return 0;
    acc = -abs;
  } else {
    temp = INT_MAX;
    bound = temp;
    if (abs > bound) return 0;
    acc = abs;
  }

  *res = acc;
  return 1;
}

int tryReadFlags(int *res, char *str) {
  int flagsToBeRaised, flagsToBeLowered;
  char *curr;

  if (*str == '\0') return 0;

  flagsToBeRaised = 0;
  flagsToBeLowered = 0;
  for (curr=str;*curr!='\0';curr++) {
    switch (*curr) {
    case 'p':
      flagsToBeLowered |= FLAG_PRECISION;
      break;
    case 'i':
      flagsToBeLowered |= FLAG_INVALID;
      break;
    case 'u':
      flagsToBeLowered |= FLAG_UNDERFLOW;
      break;
    case 'o':
      flagsToBeLowered |= FLAG_OVERFLOW;
      break;
    case 'z':
      flagsToBeLowered |= FLAG_DIVIDE_BY_ZERO;
      break;
    case 'P':
      flagsToBeRaised |= FLAG_PRECISION;
      break;
    case 'I':
      flagsToBeRaised |= FLAG_INVALID;
      break;
    case 'U':
      flagsToBeRaised |= FLAG_UNDERFLOW;
      break;
    case 'O':
      flagsToBeRaised |= FLAG_OVERFLOW;
      break;
    case 'Z':
      flagsToBeRaised |= FLAG_DIVIDE_BY_ZERO;
      break;
    default:
      return 0;
    }
  }

  flagsToBeLowered &= FLAG_ALL_SET;
  flagsToBeRaised &= FLAG_ALL_SET;
  if ((flagsToBeLowered & flagsToBeRaised) != 0) return 0;

  *res = flagsToBeRaised;
  return 1;
}

int parseLine(char *origline, testvector_t *testvector) {
  char *start, *curr;
  int maxWordCount, wordCount, i, parseOkay, currWord;
  char **words;
  char *line;
  uint64_t uint64temp;
  int inttemp;
  int negative;

  line = (char *) mycalloc(strlen(origline) + 1, sizeof(char));
  strcpy(line, origline);

  /* Trim of leading and trailing whitespace */
  for (start=line;*start!='\0';start++) {
    if ((*start != ' ') && 
	(*start != '\t')) break;
  }

  for (curr=start;*curr!='\0';curr++);
  for (;curr>=start;curr--) {
    if ((*curr != ' ') && 
	(*curr != '\t') && 
	(*curr != '\0')) break;
    *curr = '\0';
  }

  /* EZ return for empty lines */
  if (*start == '\0') {
    free(line);
    return EMPTY_LINE;
  }

  /* Check for comment lines */
  if ((*start == '/') && (*(start + 1) == '/')) {
    free(line);
    return EMPTY_LINE;
  }

  /* Cut the line into words delimited by whitespace */
  maxWordCount = strlen(start);
  words = (char **) mycalloc(maxWordCount, sizeof(char *));
  wordCount = 1;
  words[wordCount-1]=start;
  for (curr=start;*curr!='\0';curr++) {
    if ((*curr == ' ') || 
	(*curr == '\t')) {
      for (;*curr!='\0';curr++) {
	if ((*curr != ' ') && 
	    (*curr != '\t')) break;
	*curr = '\0';
      }
      if (*curr != '\0') {
	wordCount++;
	words[wordCount-1]=curr;
      }
    }
  }

  /* Analyze the different words */
  parseOkay = 1;
  currWord = 0;

  /* First read a binary64 in two possible formats:

     - 0xdeadbeefcafebabe (in hexadecimal memory notation)

     - sign expo mantissa (with expo and mantissa in decimal notation)

  */
  if ((wordCount >= 1 + currWord) && /* at least 1 word left to be read */
      (strlen(words[currWord]) == 2 + 16) && /* 0x => 2, 16 hexadecimal digits follow */
      (words[currWord][0] == '0') && 
      (words[currWord][1] == 'x')) {
    if (tryReadHexadecimal(&uint64temp, &(words[currWord][2]))) {
      testvector->x = *((_binary64 *) &uint64temp);
      currWord++;
    } else {
      parseOkay = 0;
    }
  } else {
    if ((wordCount >= 3 + currWord) && /* at least 3 words left for sign, expo and mantissa */
	(strlen(words[currWord]) == 1) && /* 1 character for the sign */
	((words[currWord][0] == '+') || (words[currWord][0] == '-'))) {
      negative = (words[currWord][0] == '-');
      if (tryReadDecimalSigned(&inttemp, words[currWord + 1]) &&
	  tryReadDecimal(&uint64temp, words[currWord + 2])) {
	testvector->x = composeBinary(negative, inttemp, uint64temp);
	currWord += 3;
      } else {
	parseOkay = 0;
      }
    } else {
      parseOkay = 0;
    }
  }
  if (!parseOkay) goto getMeOuttaHere;


  /* Then read a decimal64 in two possible formats:

     - 0xdeadbeefcafebabe (in hexadecimal memory notation)

     - sign expo mantissa (with expo and mantissa in decimal notation)

  */
  if ((wordCount >= 1 + currWord) && /* at least 1 word left to be read */
      (strlen(words[currWord]) == 2 + 16) && /* 0x => 2, 16 hexadecimal digits follow */
      (words[currWord][0] == '0') && 
      (words[currWord][1] == 'x')) {
    if (tryReadHexadecimal(&uint64temp, &(words[currWord][2]))) {
      testvector->y = *((_Decimal64 *) &uint64temp);
      currWord++;
    } else {
      parseOkay = 0;
    }
  } else {
    if ((wordCount >= 3 + currWord) && /* at least 3 words left for sign, expo and mantissa */
	(strlen(words[currWord]) == 1) && /* 1 character for the sign */
	((words[currWord][0] == '+') || (words[currWord][0] == '-'))) {
      negative = (words[currWord][0] == '-');
      if (tryReadDecimalSigned(&inttemp, words[currWord + 1]) &&
	  tryReadDecimal(&uint64temp, words[currWord + 2])) {
	testvector->y = composeDecimal(negative, inttemp, uint64temp);
	currWord += 3;
      } else {
	parseOkay = 0;
      }
    } else {
      parseOkay = 0;
    }
  }
  if (!parseOkay) goto getMeOuttaHere;

  /* Now read two words puioz PUIOZ for the before and after flags */
  if (wordCount >= 2 + currWord) {
    if (tryReadFlags(&(testvector->beforeFlags), words[currWord]) &&
	tryReadFlags(&(testvector->expectedAfterFlags), words[currWord + 1])) {
      currWord += 2;
    } else {
      parseOkay = 0;
    }
  } else {
    parseOkay = 0;
  }
  if (!parseOkay) goto getMeOuttaHere;

  /* Finally read one word true or false for the expected result */
  if (wordCount >= 1 + currWord) {
    if (strcmp(words[currWord],"true") == 0) {
      testvector->expectedResult = 1;
      currWord++;
    } else {
      if (strcmp(words[currWord],"false") == 0) {
	testvector->expectedResult = 0;
	currWord++;
      } else {
	parseOkay = 0;
      }
    }
  } else {
    parseOkay = 0;
  }

 getMeOuttaHere:
  free(words);
  free(line);
  if (parseOkay) return PARSE_OKAY;
  return PARSE_ERROR;
}

int readTestVectorChunk(FILE *fd, testvector_t *testvector, int maxElements) {
  char *currLine;
  int i;
  char readBuffer[READ_BUFFER_SIZE];
  char *bufRest = NULL;
  size_t bufRestSize = 0;
  int parseRes;

  for (i=0;i<maxElements;) {
    /* Read a line */
    if ((currLine = readTestVectorLine(fd, readBuffer, READ_BUFFER_SIZE, &bufRest, &bufRestSize)) == NULL) {
      /* Could not read a line anymore */
      return i;
    }
    /* Parse it */
    if ((parseRes = parseLine(currLine, &(testvector[i]))) != PARSE_OKAY) {
      /* Could not parse the line, skip it */
      if (parseRes != EMPTY_LINE) 
	fprintf(stderr, "Could not parse the test vector line \"%s\", skipping it.\n", currLine);
    } else {
      i++;
    }
    /* Free the line */
    free(currLine);
  }

  if (bufRestSize > 0) {
    if (fseek(fd, -((long) bufRestSize), SEEK_CUR) != 0) {
      fprintf(stderr, "Error during fseek: %s\n", strerror(errno));
    } 
  }

  return maxElements;
}

void flagsToString(char *str, int flags) {
  if (flags & FLAG_PRECISION)      str[0] = 'P'; else str[0] = 'p';
  if (flags & FLAG_UNDERFLOW)      str[1] = 'U'; else str[1] = 'u';
  if (flags & FLAG_OVERFLOW)       str[2] = 'O'; else str[2] = 'o';
  if (flags & FLAG_DIVIDE_BY_ZERO) str[3] = 'Z'; else str[3] = 'z';
  if (flags & FLAG_INVALID)        str[4] = 'I'; else str[4] = 'i';
}

void setFlags(int flags) {
  int excepts = 0;

  if (flags & FLAG_PRECISION)      excepts |= FE_INEXACT;
  if (flags & FLAG_UNDERFLOW)      excepts |= FE_UNDERFLOW;
  if (flags & FLAG_OVERFLOW)       excepts |= FE_OVERFLOW;
  if (flags & FLAG_DIVIDE_BY_ZERO) excepts |= FE_DIVBYZERO;
  if (flags & FLAG_INVALID)        excepts |= FE_INVALID;
 
  feclearexcept(FE_ALL_EXCEPT);
  feraiseexcept(excepts);
}

int getFlags() {
  int excepts, flags;

  excepts = fetestexcept(FE_ALL_EXCEPT);
  
  flags = 0;
  if (excepts & FE_INEXACT)   flags |= FLAG_PRECISION;
  if (excepts & FE_UNDERFLOW) flags |= FLAG_UNDERFLOW;
  if (excepts & FE_OVERFLOW)  flags |= FLAG_OVERFLOW;
  if (excepts & FE_DIVBYZERO) flags |= FLAG_DIVIDE_BY_ZERO;
  if (excepts & FE_INVALID)   flags |= FLAG_INVALID;
 
  return flags;
}

int compareFlags(int flags1, int flags2) {
  return ((flags1 & FLAG_ALL_SET) == (flags2 & FLAG_ALL_SET));
}

int checkTestVectorChunk(testvector_t *testvector, int maxElements) {
  int i, okayVals, okay;
  char actualFlagsString[] = "PUOZI";
  char expectedFlagsString[] = "PUOZI";

  okayVals = 0;
  for (i=0;i<maxElements;i++) {
    setFlags(testvector[i].beforeFlags);
    testvector[i].actualResult = compareSignalingLessEqualBinary64Decimal64(testvector[i].x, testvector[i].y);
    testvector[i].actualAfterFlags = getFlags();
    okay = 1;
    if (!compareFlags(testvector[i].expectedAfterFlags, testvector[i].actualAfterFlags)) {
      flagsToString(actualFlagsString, testvector[i].actualAfterFlags);
      flagsToString(expectedFlagsString, testvector[i].expectedAfterFlags);
      printf("Flag error for comparison between 0x%016llx to 0x%016llx: actual flags are %s, we expected %s\n", 
	     *((unsigned long long int *) &(testvector[i].x)),
	     *((unsigned long long int *) &(testvector[i].y)),
	     actualFlagsString, expectedFlagsString);
      okay = 0;
    }
    if ((!!testvector[i].actualResult) != (!!testvector[i].expectedResult)) {
      printf("Result error for comparison between 0x%016llx to 0x%016llx: actual result is %s, we expected %s\n", 
	     *((unsigned long long int *) &(testvector[i].x)),
	     *((unsigned long long int *) &(testvector[i].y)),
	     (testvector[i].actualResult ? "true" : "false"),
	     (testvector[i].expectedResult ? "true" : "false"));      
      okay = 0;
    }
    if (okay) okayVals++;
  }

  return okayVals;
}

void timeTestVectorChunk(_binary64 *functionCallTimeRes, testvector_t *testvector, int maxElements, 
			 int cacheHeatRuns, int timingRuns, 
			 int maxExpectedTimingInteger, int maxExpectedFunctionCallTimingInteger,
			 int functionCallTimingRepetitions) {
  volatile int resultsSink;
  int i, k, l, okay;
  _binary64 timing;
  _binary64 functionCallTime, fCT;
  _binary64 maxExpectedTiming, maxExpectedFunctionCallTiming;
  unsigned long long int before, after;
  int results, res;

  /* Get the FP representation of the maximum expected timing value */
  maxExpectedTiming = maxExpectedTimingInteger;
  maxExpectedFunctionCallTiming = maxExpectedFunctionCallTimingInteger;

  /* Run the function call timing at least once */
  if (functionCallTimingRepetitions < 1) functionCallTimingRepetitions = 1;

  /* Create a results register that gets used in the end */
  results = 0;

  /* Get the timing needed to call a function */
  functionCallTime = 2.0 * maxExpectedFunctionCallTiming;

  for (l=0;l<functionCallTimingRepetitions;l++) {
    fCT = 0.0;
    /* Run the function call timing cacheHeatRuns times to heat the cache */
    for (k=0;k<((cacheHeatRuns / functionCallTimingRepetitions) + 1);k++) {
      for (i=0;i<maxElements;i++) {
	do {
	  /* The real measurement */
	  READ_TIME_COUNTER(&before);
	  res = emptyFunctionBinary64Decimal64(testvector[i].x, testvector[i].y);
	  READ_TIME_COUNTER(&after);
	  results += res;
	
	  /* Handling of the timing values */
	  if (after > before) {
	    timing = (_binary64) (after - before);
	    if (timing <= maxExpectedFunctionCallTiming) {
	      okay = 1;
	    } else {
	      okay = 0;
	      timing = 0.0;
	    }	  
	  } else {
	    timing = 0.0;
	    okay = 0;
	  }
	} while (!okay);
	fCT += timing;
      }
    }

    /* Normalize the timings for the number of runs */
    fCT /= (((_binary64) ((cacheHeatRuns / functionCallTimingRepetitions) + 1)) * ((_binary64) maxElements));

    if (fCT < functionCallTime) {
      functionCallTime = fCT;
    }
  }

  /* Clear the value again and repeat the timing for real */
  functionCallTime = 2.0 * maxExpectedFunctionCallTiming;

  for (l=0;l<functionCallTimingRepetitions;l++) {
    fCT = 0.0;
    /* Run the function call timing */
    for (k=0;k<((timingRuns / functionCallTimingRepetitions) + 1);k++) {
      for (i=0;i<maxElements;i++) {
	do {
	  /* The real measurement */
	  READ_TIME_COUNTER(&before);
	  res = emptyFunctionBinary64Decimal64(testvector[i].x, testvector[i].y);
	  READ_TIME_COUNTER(&after);
	  results += res;
	
	  /* Handling of the timing values */
	  if (after > before) {
	    timing = (_binary64) (after - before);
	    if (timing <= maxExpectedFunctionCallTiming) {
	      okay = 1;
	    } else {
	      okay = 0;
	      timing = 0.0;
	    }	  
	  } else {
	    timing = 0.0;
	    okay = 0;
	  }
	} while (!okay);
	fCT += timing;
      }
    }

    /* Normalize the timings for the number of runs */
    fCT /= (((_binary64) ((timingRuns / functionCallTimingRepetitions) + 1)) * ((_binary64) maxElements));

    if (fCT < functionCallTime) {
      functionCallTime = fCT;
    }
  }

  /* Clear the timing values */
  for (i=0;i<maxElements;i++) {
    testvector[i].computeTime = 0.0;
  }

  /* Run the timing cacheHeatRuns times to heat the cache */
  for (k=0;k<cacheHeatRuns;k++) {
    for (i=0;i<maxElements;i++) {
      do {
	/* The real measurement */
	READ_TIME_COUNTER(&before);
	res = compareSignalingLessEqualBinary64Decimal64(testvector[i].x, testvector[i].y);
	READ_TIME_COUNTER(&after);
	results += res;
	
	/* Handling of the timing values */
	if (after > before) {
	  timing = (_binary64) (after - before);
	  if (timing <= maxExpectedTiming) {
	    if (timing - functionCallTime >= 1.0) {
	      okay = 1;
	    } else {
	      okay = 0;
	      timing = 0.0;
	    }
	  } else {
	    okay = 0;
	    timing = 0.0;
	  }
	} else {
	  okay = 0;
	  timing = 0.0;
	}
      } while (!okay);
      testvector[i].computeTime += timing;
    }
  }

  /* Clear the timing values again */
  for (i=0;i<maxElements;i++) {
    testvector[i].computeTime = 0.0;
  }

  /* Run the timing TIMING_RUNS times for real */
  for (k=0;k<timingRuns;k++) {
    for (i=0;i<maxElements;i++) {
      do {
	/* The real measurement */
	READ_TIME_COUNTER(&before);
	res = compareSignalingLessEqualBinary64Decimal64(testvector[i].x, testvector[i].y);
	READ_TIME_COUNTER(&after);
	results += res;

	/* Handling of the timing values */
	if (after > before) {
	  timing = (_binary64) (after - before);
	  if (timing <= maxExpectedTiming) {
	    if (timing - functionCallTime >= 1.0) {
	      okay = 1;
	    } else {
	      okay = 0;
	      timing = 0.0;
	    }
	  } else {
	    okay = 0;
	    timing = 0.0;
	  }
	} else {
	  okay = 0;
	  timing = 0.0;
	}
      } while (!okay);
      testvector[i].computeTime += timing;
    }
  }
 
  /* Normalize the timings for the number of runs and subtract function call time */
  for (i=0;i<maxElements;i++) {
    testvector[i].computeTime /= (_binary64) (timingRuns);
    testvector[i].computeTimeWithoutFuncCall = testvector[i].computeTime - functionCallTime;
    if (testvector[i].computeTimeWithoutFuncCall < 0.0) { 
      testvector[i].computeTimeWithoutFuncCall = 0.0; 
      fprintf(stderr,"Warning: measured a negative lapse of time for a function call.\n");
    }
  }

  /* Use the results to be sure that nothing has been optimized out */
  resultsSink = results;

  /* Return the function call timing */
  *functionCallTimeRes = functionCallTime;
}

void printTimingTestVectorChunk(testvector_t *testvector, int maxElements) {
  int i;

  for (i=0;i<maxElements;i++) {
    printf("Comparing 0x%016llx to 0x%016llx took %3.5f time units, %3.5f without the function call.\n", 
	   *((unsigned long long int *) &(testvector[i].x)),
	   *((unsigned long long int *) &(testvector[i].y)),
	   testvector[i].computeTime, testvector[i].computeTimeWithoutFuncCall);
  }
}

void initializeTimeAnalytics(timeanalytics_t *timeanalytics) {
  int i, k;

  timeanalytics->avgFunctionCallHits = 0;
  timeanalytics->avgFunctionCallTiming = 0.0;
  timeanalytics->minFunctionCallTiming = 1e308 * 1e308;
  timeanalytics->maxFunctionCallTiming = -1e308 * 1e308;

  for (i=0;i<BINARY64_NUMBER_CLASSES;i++) {
    for (k=0;k<DECIMAL64_NUMBER_CLASSES;k++) {
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].minTiming = 1e308 * 1e308;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].minTimingXArg = 0.0;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].minTimingYArg = 0.0;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].maxTiming = -1e308 * 1e308;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].maxTimingXArg = 0.0;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].maxTimingYArg = 0.0;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].avgTiming = 0.0;
      timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].avgHits = 0;
    }
  }

  for (i=0;i<EXTRA_CLASSES;i++) {
    timeanalytics->extraClass[i].minTiming = 1e308 * 1e308;
    timeanalytics->extraClass[i].minTimingXArg = 0.0;
    timeanalytics->extraClass[i].minTimingYArg = 0.0;
    timeanalytics->extraClass[i].maxTiming = -1e308 * 1e308;
    timeanalytics->extraClass[i].maxTimingXArg = 0.0;
    timeanalytics->extraClass[i].maxTimingYArg = 0.0;
    timeanalytics->extraClass[i].avgTiming = 0.0;
    timeanalytics->extraClass[i].avgHits = 0;
  }
}

void analyzeTimeTestVectorChunk(timeanalytics_t *timeanalytics, testvector_t *testvector, int maxElements, _binary64 functionCallTiming, int easyCases, int hardCases) {
  int i, binClass, decClass, extraClass;
  int isAnEasyCase;

  timeanalytics->avgFunctionCallHits++;
  timeanalytics->avgFunctionCallTiming += functionCallTiming;
  if (timeanalytics->minFunctionCallTiming > functionCallTiming) {
    timeanalytics->minFunctionCallTiming = functionCallTiming;
  }
  if (timeanalytics->maxFunctionCallTiming < functionCallTiming) {
    timeanalytics->maxFunctionCallTiming = functionCallTiming;
  }  

  for (i=0;i<maxElements;i++) {
    isAnEasyCase = classifyCase(testvector[i].x, testvector[i].y) == PAIR_EASY;

    if ((easyCases && isAnEasyCase) || (hardCases && (!isAnEasyCase))) {
      binClass = classifyBinary(testvector[i].x);
      decClass = classifyDecimal(testvector[i].y);

      if (timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].minTiming > 
	  testvector[i].computeTimeWithoutFuncCall) {
	timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].minTiming = 
	  testvector[i].computeTimeWithoutFuncCall;
	timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].minTimingXArg = 
	  testvector[i].x;
	timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].minTimingYArg = 
	  testvector[i].y;
      }
      if (timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].maxTiming < 
	  testvector[i].computeTimeWithoutFuncCall) {
	timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].maxTiming = 
	  testvector[i].computeTimeWithoutFuncCall;
	timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].maxTimingXArg = 
	  testvector[i].x;
	timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].maxTimingYArg = 
	  testvector[i].y;
      }
      timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].avgTiming += testvector[i].computeTimeWithoutFuncCall;
      timeanalytics->class[binClass * DECIMAL64_NUMBER_CLASSES + decClass].avgHits++;

      extraClass = classifyExtraClass(binClass, decClass, isAnEasyCase == PAIR_EASY);

      if (timeanalytics->extraClass[extraClass].minTiming > 
	  testvector[i].computeTimeWithoutFuncCall) {
	timeanalytics->extraClass[extraClass].minTiming = 
	  testvector[i].computeTimeWithoutFuncCall;
	timeanalytics->extraClass[extraClass].minTimingXArg = 
	  testvector[i].x;
	timeanalytics->extraClass[extraClass].minTimingYArg = 
	  testvector[i].y;
      }

      if (timeanalytics->extraClass[extraClass].maxTiming <
	  testvector[i].computeTimeWithoutFuncCall) {
	timeanalytics->extraClass[extraClass].maxTiming = 
	  testvector[i].computeTimeWithoutFuncCall;
	timeanalytics->extraClass[extraClass].maxTimingXArg = 
	  testvector[i].x;
	timeanalytics->extraClass[extraClass].maxTimingYArg = 
	  testvector[i].y;
      }
      timeanalytics->extraClass[extraClass].avgTiming += testvector[i].computeTimeWithoutFuncCall;
      timeanalytics->extraClass[extraClass].avgHits++;

    }
  }
}

void printTimeAnalytics(timeanalytics_t *timeanalytics, int showCondensed, int showDetailed) {
  int i, k;
  int minT, maxT, avgT, minFT, avgFT, maxFT;
  int minTOkay, maxTOkay, avgTOkay;
  int minFTOkay, maxFTOkay, avgFTOkay;
  _binary64 avgTFloat, avgFTFloat, timingOffset;

  avgFTFloat = (timeanalytics->avgFunctionCallTiming) / ((_binary64) timeanalytics->avgFunctionCallHits);
  if ((avgFTFloat < 1000.0) && (avgFTFloat >= 0.0)) {
    avgFT = ((int) avgFTFloat) + 1;
    if (avgFT <= 999) avgFTOkay = 1;
  } else {
    avgFTOkay = 0;
  }
  if ((timeanalytics->minFunctionCallTiming < 1000.0) && 
      (timeanalytics->minFunctionCallTiming >= 0.0)) {
    minFT = ((int) timeanalytics->minFunctionCallTiming) + 1;
    if (minFT <= 999) minFTOkay = 1;
  } else {
    minFTOkay = 0;
  }
  if ((timeanalytics->maxFunctionCallTiming < 1000.0) && 
      (timeanalytics->maxFunctionCallTiming >= 0.0)) {
    maxFT = ((int) timeanalytics->maxFunctionCallTiming) + 1;
    if (maxFT <= 999) maxFTOkay = 1;
  } else {
    maxFTOkay = 0;
  }
  
  timingOffset = (_binary64) (timeanalytics->maxFunctionCallTiming - timeanalytics->minFunctionCallTiming);

  if (showCondensed) {
    printf("------------------------+---------------------+\n");
    for (i=0;i<EXTRA_CLASSES;i++) {
      printf("%s|", extraClassToName(i));
      if (timeanalytics->extraClass[i].avgHits > 0) {
	avgTFloat = timeanalytics->extraClass[i].avgTiming / 
	  ((_binary64) timeanalytics->extraClass[i].avgHits);
	if ((avgTFloat + timingOffset < 1000.0) && (avgTFloat + timingOffset >= 0.0)) {
	  avgT = ((int) (avgTFloat + timingOffset)) + 1;
	  if (avgT <= 999) avgTOkay = 1;
	} else {
	  avgTOkay = 0;
	}
	if ((timeanalytics->extraClass[i].minTiming + timingOffset < 1000.0) && 
	    (timeanalytics->extraClass[i].minTiming + timingOffset >= 0.0)) {
	  minT = ((int) (timeanalytics->extraClass[i].minTiming + timingOffset)) + 1;
	  if (minT <= 999) minTOkay = 1;
	} else {
	  minTOkay = 0;
	}
	if ((timeanalytics->extraClass[i].maxTiming + timingOffset < 1000.0) && 
	    (timeanalytics->extraClass[i].maxTiming + timingOffset >= 0.0)) {
	  maxT = ((int) (timeanalytics->extraClass[i].maxTiming + timingOffset)) + 1;
	  if (maxT <= 999) maxTOkay = 1;
	} else {
	  maxTOkay = 0;
	}
	if (minTOkay) {
	  printf("%3d/",minT);
	} else {
	  printf("---/");
	}
	if (avgTOkay) {
	  printf("%3d/",avgT);
	} else {
	  printf("---/");
	}
	if (maxTOkay) {
	  printf("%3d ",maxT);
	} else {
	  printf("--- ");
	}
      } else {
	printf("   /   /    ");
      }
      printf("(%7d)|", timeanalytics->extraClass[i].avgHits);
      printf("\n");
    }
    printf("------------------------+---------------------+\n");
    printf("\n");
  }

  if (showDetailed) {
    printf(" B\\D |");
    for (k=0;k<DECIMAL64_NUMBER_CLASSES;k++) {
      printf("%s                |",decimalClassToName(k));
    }
    printf("\n");
    printf("-----+");
    for (k=0;k<DECIMAL64_NUMBER_CLASSES;k++) {
      printf("---------------------+");
    }
    printf("\n");
    for (i=0;i<BINARY64_NUMBER_CLASSES;i++) {
      printf("%s|",binaryClassToName(i));
      for (k=0;k<DECIMAL64_NUMBER_CLASSES;k++) {
	if (timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].avgHits > 0) {
	  avgTFloat = timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].avgTiming / 
	    ((_binary64) timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].avgHits);
	  if ((avgTFloat + timingOffset < 1000.0) && (avgTFloat + timingOffset >= 0.0)) {
	    avgT = ((int) (avgTFloat + timingOffset)) + 1;
	    if (avgT <= 999) avgTOkay = 1;
	  } else {
	    avgTOkay = 0;
	  }
	  if ((timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].minTiming + timingOffset < 1000.0) && 
	      (timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].minTiming + timingOffset >= 0.0)) {
	    minT = ((int) (timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].minTiming + timingOffset)) + 1;
	    if (minT <= 999) minTOkay = 1;
	  } else {
	    minTOkay = 0;
	  }
	  if ((timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].maxTiming + timingOffset < 1000.0) && 
	      (timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].maxTiming + timingOffset >= 0.0)) {
	    maxT = ((int) (timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].maxTiming + timingOffset)) + 1;
	    if (maxT <= 999) maxTOkay = 1;
	  } else {
	    maxTOkay = 0;
	  }
	  if (minTOkay) {
	    printf("%3d/",minT);
	  } else {
	    printf("---/");
	  }
	  if (avgTOkay) {
	    printf("%3d/",avgT);
	  } else {
	    printf("---/");
	  }
	  if (maxTOkay) {
	    printf("%3d ",maxT);
	  } else {
	    printf("--- ");
	  }
	} else {
	  printf("   /   /    ");
	}
	printf("(%7d)|", timeanalytics->class[i * DECIMAL64_NUMBER_CLASSES + k].avgHits);
      }
      printf("\n");
    }
    printf("-----+");
    for (k=0;k<DECIMAL64_NUMBER_CLASSES;k++) {
      printf("---------------------+");
    }
    printf("\n");

    printf("Time for a function call: ");
    if (minFTOkay) {
      printf("%3d/",minFT);
    } else {
      printf("---/");
    }
    if (avgFTOkay) {
      printf("%3d/",avgFT);
    } else {
      printf("---/");
    }
    if (maxFTOkay) {
      printf("%3d\n",maxFT);
    } else {
      printf("---\n");
    }
  }

}


/* Simple main testing loop */

int main(int argc, char **argv) {
  FILE *fd;
  int numTests, overallTests, okayTests;
  testvector_t testvector[NUM_ELEMENTS_TESTVECTOR];
  int cacheHeatRuns, timingRuns, maxExpectedTiming, maxExpectedFunctionCallTiming;
  int functionCallTimingRepetitions;
  timeanalytics_t timeanalyticsAll, timeanalyticsEasy, timeanalyticsHard;
  _binary64 functionCallTiming;
  int showEasy, showHard, showCombined, showCondensed, showDetailed;

  if (argc < 2) {
    fprintf(stderr, "Not enough arguments given.\nUsage:\n%s <testvectorfile> <timings for easy, hard, both, combined, all, none> <display mode: condensed, detailed, both> <timing runs> <cache heat runs> <max expected time> <max expected function call time> <number of function call time measurements>", argv[0]);
    exit(17);
  } 

  if ((fd = fopen(argv[1],"r")) == NULL) {
    fprintf(stderr, "Could not open the test vector file \"%s\": %s\n", argv[1], strerror(errno));
    exit(17);
  }

  if (argc < 3) {
    showEasy = 1;
    showHard = 1;
    showCombined = 1;
  } else {
    if (strcmp(argv[2], "all") == 0) {
      showEasy = 1;
      showHard = 1;
      showCombined = 1;
    } else {
      if (strcmp(argv[2], "both") == 0) {
	showEasy = 1;
	showHard = 1;
	showCombined = 0;
      } else {
	if (strcmp(argv[2], "easy") == 0) {
	  showEasy = 1;
	  showHard = 0;
	  showCombined = 0;
	} else {
	  if (strcmp(argv[2], "hard") == 0) {
	    showEasy = 0;
	    showHard = 1;
	    showCombined = 0;
	  } else {
	    if (strcmp(argv[2], "combined") == 0) {
	      showEasy = 0;
	      showHard = 0;
	      showCombined = 1;
	    } else {
	      if (strcmp(argv[2], "none") == 0) {
		showEasy = 0;
		showHard = 0;
		showCombined = 0;
	      } else {
		showEasy = 1;
		showHard = 1;
		showCombined = 1;
		fprintf(stderr, "Could not read the requested display mode, using default mode.\n");
	      }
	    }
	  }
	}
      }
    }
  }
  
  if (argc < 4) {
    showCondensed = 1;
    showDetailed = 1;
  } else {
    if (strcmp(argv[3], "both") == 0) {
      showCondensed = 1;
      showDetailed = 1;
    } else {
      if (strcmp(argv[3], "detailed") == 0) {
	showCondensed = 0;
	showDetailed = 1;
      } else {
	if (strcmp(argv[3], "condensed") == 0) {
	  showCondensed = 1;
	  showDetailed = 0;
	} else {
	  showCondensed = 1;
	  showDetailed = 1;
	  fprintf(stderr, "Could not read the requested display mode, using default mode.\n");
	}
      }
    }
  }

  if (argc < 5) {
    timingRuns = TIMING_RUNS;
  } else {
    if (!tryReadDecimalSigned(&timingRuns, argv[4])) {
      timingRuns = TIMING_RUNS;
      fprintf(stderr, "Could not read the number of requested timing runs, using default value %d.\n", timingRuns);
    }
  }

  if (argc < 6) {
    cacheHeatRuns = CACHE_HEAT_RUNS;
  } else {
    if (!tryReadDecimalSigned(&cacheHeatRuns, argv[5])) {
      cacheHeatRuns = CACHE_HEAT_RUNS;
      fprintf(stderr, "Could not read the number of requested cache heat runs, using default value %d.\n", cacheHeatRuns);
    }
  }

  if (argc < 7) {
    maxExpectedTiming = MAX_EXPECTED_TIMING;
  } else {
    if (!tryReadDecimalSigned(&maxExpectedTiming, argv[6])) {
      maxExpectedTiming = MAX_EXPECTED_TIMING;
      fprintf(stderr, "Could not read the maximum expected timing value, using default value %d.\n", maxExpectedTiming);
    }
  }

  if (argc < 8) {
    maxExpectedFunctionCallTiming = MAX_EXPECTED_FUNCTION_CALL_TIMING;
  } else {
    if (!tryReadDecimalSigned(&maxExpectedFunctionCallTiming, argv[7])) {
      maxExpectedFunctionCallTiming = MAX_EXPECTED_FUNCTION_CALL_TIMING;
      fprintf(stderr, "Could not read the maximum expected function call timing value, using default value %d.\n", maxExpectedFunctionCallTiming);
    }
  }

  if (argc < 9) {
    functionCallTimingRepetitions = FUNCTION_CALL_TIMING_REPS;
  } else {
    if (!tryReadDecimalSigned(&functionCallTimingRepetitions, argv[8])) {
      functionCallTimingRepetitions = FUNCTION_CALL_TIMING_REPS;
      fprintf(stderr, "Could not read the function call timing repetitions value, using default value %d.\n", functionCallTimingRepetitions);
    }
  }

  /* Attempt to bind the process to a single core */
  cpu_set_t cpuMask;
  CPU_ZERO(&cpuMask);
  CPU_SET(0, &cpuMask);
  sched_setaffinity(0, sizeof(cpuMask), &cpuMask);

  if (timingRuns > 0) {
    initializeTimeAnalytics(&timeanalyticsAll);
    initializeTimeAnalytics(&timeanalyticsEasy);
    initializeTimeAnalytics(&timeanalyticsHard);
  }
  overallTests = 0;
  okayTests = 0;
  while (1) {
    numTests = readTestVectorChunk(fd, testvector, NUM_ELEMENTS_TESTVECTOR);
    okayTests += checkTestVectorChunk(testvector, numTests);
    overallTests += numTests;
    if (timingRuns > 0) {
      timeTestVectorChunk(&functionCallTiming, testvector, numTests, 
			  cacheHeatRuns, timingRuns, maxExpectedTiming, maxExpectedFunctionCallTiming, functionCallTimingRepetitions);
      analyzeTimeTestVectorChunk(&timeanalyticsAll, testvector, numTests, functionCallTiming, 1, 1);
      analyzeTimeTestVectorChunk(&timeanalyticsEasy, testvector, numTests, functionCallTiming, 1, 0);
      analyzeTimeTestVectorChunk(&timeanalyticsHard, testvector, numTests, functionCallTiming, 0, 1);
    }
    if (numTests < NUM_ELEMENTS_TESTVECTOR) 
      break;
  }

  if (fclose(fd) != 0) {
    fprintf(stderr, "Could not close the test vector file \"%s\": %s\n", argv[1], strerror(errno));
    exit(17);
  }

  if (timingRuns > 0) {
    if (showCombined) {
      printf("\n");
      printf("All test cases:\n");
      printTimeAnalytics(&timeanalyticsAll, showCondensed, showDetailed);
      printf("\n");
    }
    if (showEasy) {
      printf("\n");
      printf("Easy test cases only:\n");
      printTimeAnalytics(&timeanalyticsEasy, showCondensed, showDetailed);
      printf("\n");
    }
    if (showHard) {
      printf("\n");
      printf("Hard test cases only:\n");
      printTimeAnalytics(&timeanalyticsHard, showCondensed, showDetailed);
      printf("\n");
    }
  }

  if (okayTests < overallTests) {
    printf("FAILURE - only %d out of %d tests have yield to a correct result.\n", okayTests, overallTests);
    return 1;
  } else {
    printf("SUCCESS - all %d tests have yield to a correct result.\n", okayTests);
  }
  
  return 0;
}


