
#ifndef FP_UTILS_H
#define FP_UTILS_H

#include <stdio.h>
#include <stdint.h>
#ifndef __STDC_WANT_DEC_FP__
#define __STDC_WANT_DEC_FP__ 1
#endif
#include <math.h>

#include "wide_integers.h"

/* An enumeration of the different classes of FP data */

enum __comp_fp_class_t_enum {
  COMP_FP_CLASS_NONZERO_REAL,
  COMP_FP_CLASS_SIGNALING_NAN,
  COMP_FP_CLASS_QUIET_NAN,
  COMP_FP_CLASS_INFINITY,
  COMP_FP_CLASS_ZERO
};
typedef enum __comp_fp_class_t_enum comp_fp_class_t;

/* Typedefs for the FP types */

typedef float         comp_binary_32_t;
typedef double        comp_binary_64_t;
typedef __float128    comp_binary_128_t;

typedef _Decimal64    comp_decimal_64_t;
typedef _Decimal128   comp_decimal_128_t;

/* Types to access the fields of the FP types */

typedef union {
  uint32_t           i;
  comp_binary_32_t   f;
} __comp_binary_32_caster_t;

typedef union {
  uint64_t           i;
  comp_binary_64_t   f;
} __comp_binary_64_caster_t;

typedef union {
  __uint128_t        i;
  comp_binary_128_t  f;
} __comp_binary_128_caster_t;

typedef union {
  uint64_t           i;
  comp_decimal_64_t  f;
} __comp_decimal_64_caster_t;

typedef union {
  __uint128_t        i;
  comp_decimal_128_t f;
} __comp_decimal_128_caster_t;

/* Symbolic constants for the FP types */

#define __COMP_BINARY_32_SIGNIFICAND_LENGTH          (23)
#define __COMP_BINARY_64_SIGNIFICAND_LENGTH          (52)
#define __COMP_BINARY_128_SIGNIFICAND_LENGTH         (112)

#define __COMP_BINARY_32_EXPONENT_WIDTH              (8)
#define __COMP_BINARY_64_EXPONENT_WIDTH              (11)
#define __COMP_BINARY_128_EXPONENT_WIDTH             (15)

#define __COMP_DECIMAL_64_SIGNIFICAND_LENGTH         (50)
#define __COMP_DECIMAL_128_SIGNIFICAND_LENGTH        (110)

#define __COMP_DECIMAL_64_COMBINATION_FLD_LENGTH     (13)
#define __COMP_DECIMAL_128_COMBINATION_FLD_LENGTH    (17)

#define __COMP_DECIMAL_64_BITS_AFTER_CLASS_BITS      (64 - 1 - 6)
#define __COMP_DECIMAL_128_BITS_AFTER_CLASS_BITS     (128 - 1 - 6)

#define __COMP_DECIMAL_64_INFINITY_CLASS             ((uint64_t) (30))
#define __COMP_DECIMAL_128_INFINITY_CLASS            ((__uint128_t) (30))

#define __COMP_DECIMAL_64_NAN_CLASS                  ((uint64_t) (31))
#define __COMP_DECIMAL_128_NAN_CLASS                 ((__uint128_t) (31))

#define __COMP_DECIMAL_64_SIGNALING_NAN_BIT          ((uint64_t) (1))
#define __COMP_DECIMAL_128_SIGNALING_NAN_BIT         ((uint64_t) (1))

#define __COMP_DECIMAL_64_MAX_SIGNIFICAND            ((uint64_t) (UINT64_C(9999999999999999)))
#define __COMP_DECIMAL_128_MAX_SIGNIFICAND           ((((__uint128_t) (UINT64_C(542101086242752))) << 64) + ((__uint128_t) (UINT64_C(4003012203950112767))))

#define __COMP_DECIMAL_64_BIAS                       (398)
#define __COMP_DECIMAL_128_BIAS                      (6176)

#define __COMP_DECIMAL_64_BITS_AFTER_G0G1_BITS       (64 - 1 - 2)
#define __COMP_DECIMAL_128_BITS_AFTER_G0G1_BITS      (128 - 1 - 2)

#define __COMP_DECIMAL_64_SIGNIFICAND_LEN_CASE_ONE   (53)
#define __COMP_DECIMAL_64_SIGNIFICAND_LEN_CASE_TWO   (51)
#define __COMP_DECIMAL_64_SIGNIFICAND_EXTRA_BIT      (53)
#define __COMP_DECIMAL_128_SIGNIFICAND_LEN_CASE_ONE  (113)
#define __COMP_DECIMAL_128_SIGNIFICAND_LEN_CASE_TWO  (111)
#define __COMP_DECIMAL_128_SIGNIFICAND_EXTRA_BIT     (113)

#define __COMP_DECIMAL_64_EXPONENT_LEN_CASE_ONE      (10)
#define __COMP_DECIMAL_64_EXPONENT_LEN_CASE_TWO      (10)
#define __COMP_DECIMAL_128_EXPONENT_LEN_CASE_ONE     (14)
#define __COMP_DECIMAL_128_EXPONENT_LEN_CASE_TWO     (14)


/* Decomposition functions */

STATIC INLINE void comp_fp_decompose_binary_32(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
					       int64_t * RESTRICT exponent, comp_uint32_t * RESTRICT significand, 
					       comp_binary_32_t x) {
  __comp_binary_32_caster_t x_cast;
  uint32_t X, E, M;
  int log2M, lzc, s;

  /* Get memory representation of the IEEE754 FP number */
  x_cast.f = x;
  X = x_cast.i;
  
  /* Extract the sign */
  *sign_negative = !!((X >> (32 - 1)) != ((uint32_t) 0));

  /* Get the exponent field */
  E = (X << 1) >> 1;                           /* Forget about the sign */
  E >>= __COMP_BINARY_32_SIGNIFICAND_LENGTH;   /* Forget about the significand */
  
  /* Get the significand */
  M = X & ((((uint32_t) 1) << __COMP_BINARY_32_SIGNIFICAND_LENGTH) - ((uint32_t) 1));

  /* Check for Inf or NaN */
  if (E == ((((uint32_t) 1) << __COMP_BINARY_32_EXPONENT_WIDTH) - ((uint32_t) 1))) {
    /* The input is Inf or NaN */
    if (M == ((uint32_t) 0)) {
      /* The input is +/-Inf, set that class; the sign has already
	 been set 
      */
      *class = COMP_FP_CLASS_INFINITY;
      return;
    }
    /* The input is sNaN or qNaN */
    if (M == (((uint32_t) 1) << (__COMP_BINARY_32_SIGNIFICAND_LENGTH - 1))) {
      /* The input is a qNaN */
      *class = COMP_FP_CLASS_QUIET_NAN;
      return;
    } 
    /* The input is a sNaN */
    *class = COMP_FP_CLASS_SIGNALING_NAN;
    return;
  }

  /* The input is a real number. 

     Check for subnormal numbers and zero 
  */
  if (E == ((uint32_t) 0)) {
    /* The input is subnormal 

       Check for zero.

    */
    if (M == ((uint32_t) 0)) {
      /* The input is zero */
      *class = COMP_FP_CLASS_ZERO;
      return;
    }
    /* The input is a non-zero subnormal */
    *class = COMP_FP_CLASS_NONZERO_REAL;
    
    /* Get the leading-zero count on the mantissa */
    lzc = __comp_leading_zeros_uint32(M);
    
    /* Deduce floor(log2(M)) */
    log2M = (32 - 1) - lzc;

    /* Normalize the significand */
    s = __COMP_BINARY_32_SIGNIFICAND_LENGTH - log2M;
    M <<= s;
    
    /* Set the significand */
    *significand = comp_uint32_to_comp_uint32(M);

    /* Set the exponent */
    *exponent = (-((int64_t)((((uint32_t) 1) << (__COMP_BINARY_32_EXPONENT_WIDTH - 1)) + 
			     ((uint32_t) __COMP_BINARY_32_SIGNIFICAND_LENGTH) - ((uint32_t) 2)))) - 
      ((int64_t) s);
    return;
  }

  /* The input is a normal number, hence nonzero and real */
  *class = COMP_FP_CLASS_NONZERO_REAL;

  /* The input is a normal number 

     Subtract the exponent bias from the exponent field, add the
     hidden 1 to the significand.

  */
  M |= ((uint32_t) 1) << __COMP_BINARY_32_SIGNIFICAND_LENGTH;
  *exponent = ((int64_t) E) - 
    ((int64_t) ((((uint32_t) 1) << (__COMP_BINARY_32_EXPONENT_WIDTH - 1)) - ((uint32_t) 1))) - 
    ((int64_t) __COMP_BINARY_32_SIGNIFICAND_LENGTH);
  *significand = comp_uint32_to_comp_uint32(M);
}

STATIC INLINE void comp_fp_decompose_binary_64(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
					       int64_t * RESTRICT exponent, comp_uint64_t * RESTRICT significand, 
					       comp_binary_64_t x) {
  __comp_binary_64_caster_t x_cast;
  uint64_t X, E, M;
  int log2M, lzc, s;

  /* Get memory representation of the IEEE754 FP number */
  x_cast.f = x;
  X = x_cast.i;
  
  /* Extract the sign */
  *sign_negative = !!((X >> (64 - 1)) != ((uint64_t) 0));

  /* Get the exponent field */
  E = (X << 1) >> 1;                           /* Forget about the sign */
  E >>= __COMP_BINARY_64_SIGNIFICAND_LENGTH;   /* Forget about the significand */
  
  /* Get the significand */
  M = X & ((((uint64_t) 1) << __COMP_BINARY_64_SIGNIFICAND_LENGTH) - ((uint64_t) 1));

  /* Check for Inf or NaN */
  if (E == ((((uint64_t) 1) << __COMP_BINARY_64_EXPONENT_WIDTH) - ((uint64_t) 1))) {
    /* The input is Inf or NaN */
    if (M == ((uint64_t) 0)) {
      /* The input is +/-Inf, set that class; the sign has already
	 been set 
      */
      *class = COMP_FP_CLASS_INFINITY;
      return;
    }
    /* The input is sNaN or qNaN */
    if (M == (((uint64_t) 1) << (__COMP_BINARY_64_SIGNIFICAND_LENGTH - 1))) {
      /* The input is a qNaN */
      *class = COMP_FP_CLASS_QUIET_NAN;
      return;
    } 
    /* The input is a sNaN */
    *class = COMP_FP_CLASS_SIGNALING_NAN;
    return;
  }

  /* The input is a real number. 

     Check for subnormal numbers and zero 
  */
  if (E == ((uint64_t) 0)) {
    /* The input is subnormal 

       Check for zero.

    */
    if (M == ((uint64_t) 0)) {
      /* The input is zero */
      *class = COMP_FP_CLASS_ZERO;
      return;
    }
    /* The input is a non-zero subnormal */
    *class = COMP_FP_CLASS_NONZERO_REAL;
    
    /* Get the leading-zero count on the mantissa */
    lzc = __comp_leading_zeros_uint64(M);
    
    /* Deduce floor(log2(M)) */
    log2M = (64 - 1) - lzc;

    /* Normalize the significand */
    s = __COMP_BINARY_64_SIGNIFICAND_LENGTH - log2M;
    M <<= s;
    
    /* Set the significand */
    *significand = comp_uint64_to_comp_uint64(M);

    /* Set the exponent */
    *exponent = (-((int64_t)((((uint64_t) 1) << (__COMP_BINARY_64_EXPONENT_WIDTH - 1)) + 
			     ((uint64_t) __COMP_BINARY_64_SIGNIFICAND_LENGTH) - ((uint64_t) 2)))) - 
      ((int64_t) s);
    return;
  }

  /* The input is a normal number, hence nonzero and real */
  *class = COMP_FP_CLASS_NONZERO_REAL;

  /* The input is a normal number 

     Subtract the exponent bias from the exponent field, add the
     hidden 1 to the significand.

  */
  M |= ((uint64_t) 1) << __COMP_BINARY_64_SIGNIFICAND_LENGTH;
  *exponent = ((int64_t) E) - 
    ((int64_t) ((((uint64_t) 1) << (__COMP_BINARY_64_EXPONENT_WIDTH - 1)) - ((uint64_t) 1))) - 
    ((int64_t) __COMP_BINARY_64_SIGNIFICAND_LENGTH);
  *significand = comp_uint64_to_comp_uint64(M);
}

STATIC INLINE void comp_fp_decompose_binary_128(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
						int64_t * RESTRICT exponent, comp_uint128_t * RESTRICT significand, 
						comp_binary_128_t x) {
  __comp_binary_128_caster_t x_cast;
  __uint128_t X, E, M;
  int log2M, lzc, s;

  /* Get memory representation of the IEEE754 FP number */
  x_cast.f = x;
  X = x_cast.i;
  
  /* Extract the sign */
  *sign_negative = !!((X >> (128 - 1)) != ((__uint128_t) 0));

  /* Get the exponent field */
  E = (X << 1) >> 1;                           /* Forget about the sign */
  E >>= __COMP_BINARY_128_SIGNIFICAND_LENGTH;   /* Forget about the significand */
  
  /* Get the significand */
  M = X & ((((__uint128_t) 1) << __COMP_BINARY_128_SIGNIFICAND_LENGTH) - ((__uint128_t) 1));

  /* Check for Inf or NaN */
  if (E == ((((__uint128_t) 1) << __COMP_BINARY_128_EXPONENT_WIDTH) - ((__uint128_t) 1))) {
    /* The input is Inf or NaN */
    if (M == ((__uint128_t) 0)) {
      /* The input is +/-Inf, set that class; the sign has already
	 been set 
      */
      *class = COMP_FP_CLASS_INFINITY;
      return;
    }
    /* The input is sNaN or qNaN */
    if (M == (((__uint128_t) 1) << (__COMP_BINARY_128_SIGNIFICAND_LENGTH - 1))) {
      /* The input is a qNaN */
      *class = COMP_FP_CLASS_QUIET_NAN;
      return;
    } 
    /* The input is a sNaN */
    *class = COMP_FP_CLASS_SIGNALING_NAN;
    return;
  }

  /* The input is a real number. 

     Check for subnormal numbers and zero 
  */
  if (E == ((__uint128_t) 0)) {
    /* The input is subnormal 

       Check for zero.

    */
    if (M == ((__uint128_t) 0)) {
      /* The input is zero */
      *class = COMP_FP_CLASS_ZERO;
      return;
    }
    /* The input is a non-zero subnormal */
    *class = COMP_FP_CLASS_NONZERO_REAL;
    
    /* Get the leading-zero count on the mantissa */
    lzc = __comp_leading_zeros_uint128(M);
    
    /* Deduce floor(log2(M)) */
    log2M = (128 - 1) - lzc;

    /* Normalize the significand */
    s = __COMP_BINARY_128_SIGNIFICAND_LENGTH - log2M;
    M <<= s;
    
    /* Set the significand */
    *significand = comp_uint128_to_comp_uint128(M);

    /* Set the exponent */
    *exponent = (-((int64_t)((((__uint128_t) 1) << (__COMP_BINARY_128_EXPONENT_WIDTH - 1)) + 
			     ((__uint128_t) __COMP_BINARY_128_SIGNIFICAND_LENGTH) - ((__uint128_t) 2)))) - 
      ((int64_t) s);
    return;
  }

  /* The input is a normal number, hence nonzero and real */
  *class = COMP_FP_CLASS_NONZERO_REAL;

  /* The input is a normal number 

     Subtract the exponent bias from the exponent field, add the
     hidden 1 to the significand.

  */
  M |= ((__uint128_t) 1) << __COMP_BINARY_128_SIGNIFICAND_LENGTH;
  *exponent = ((int64_t) E) - 
    ((int64_t) ((((__uint128_t) 1) << (__COMP_BINARY_128_EXPONENT_WIDTH - 1)) - ((__uint128_t) 1))) - 
    ((int64_t) __COMP_BINARY_128_SIGNIFICAND_LENGTH);
  *significand = comp_uint128_to_comp_uint128(M);
}
					      
STATIC INLINE void comp_fp_decompose_decimal_64(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
						int64_t * RESTRICT exponent, comp_uint64_t * RESTRICT significand, 
						comp_decimal_64_t x) {
  __comp_decimal_64_caster_t x_cast;
  uint64_t X, C, CL, M, E, G01;

  /* Get memory representation of the IEEE754 FP number */
  x_cast.f = x;
  X = x_cast.i;
  
  /* Extract the sign */
  *sign_negative = !!((X >> (64 - 1)) != ((uint64_t) 0));

  /* Extract classifying fields C and CL */
  C = (X << 1) >> 1;
  C >>= __COMP_DECIMAL_64_BITS_AFTER_CLASS_BITS;
  CL = C >> 1;

  /* Check for infinity */
  if (CL == __COMP_DECIMAL_64_INFINITY_CLASS) {
    /* The input is infinity 

       The sign has already been set.
    
    */
    *class = COMP_FP_CLASS_INFINITY;
    return;
  }

  /* Check for NaN */
  if (CL == __COMP_DECIMAL_64_NAN_CLASS) {
    /* The input is NaN

       Check for signaling or quiet NaNs.
    
    */
    if ((C & __COMP_DECIMAL_64_SIGNALING_NAN_BIT) != ((uint64_t) 0)) {
      *class = COMP_FP_CLASS_SIGNALING_NAN;
    } else {
      *class = COMP_FP_CLASS_QUIET_NAN;
    }
    return;
  }
  
  /* The input is a real number 

     This implies that G_0 thru G_3 cannot be 1111.

     First get the bits G_0 and G_1, then decide which encoding case
     applies and finally get the biased exponent and the significand.

  */
  G01 = (X << 1) >> 1;
  G01 >>= __COMP_DECIMAL_64_BITS_AFTER_G0G1_BITS;
  G01 &= (uint64_t) 0x03;
  if (G01 != ((uint64_t) 0x03)) {
    /* Case (i) in IEEE 754-2008, Section 3.5.2 c) 2) */
    M = X & ((((uint64_t) 1) << __COMP_DECIMAL_64_SIGNIFICAND_LEN_CASE_ONE) - ((uint64_t) 1));
    E = (X >> __COMP_DECIMAL_64_SIGNIFICAND_LEN_CASE_ONE) & 
      ((((uint64_t) 1) << __COMP_DECIMAL_64_EXPONENT_LEN_CASE_ONE) - ((uint64_t) 1));
  } else {
    /* Case (ii) in IEEE 754-2008, Section 3.5.2 c) 2) */
    M = (X & ((((uint64_t) 1) << __COMP_DECIMAL_64_SIGNIFICAND_LEN_CASE_TWO) - ((uint64_t) 1))) | 
      (((uint64_t) 1) << __COMP_DECIMAL_64_SIGNIFICAND_EXTRA_BIT);
    E = (X >> __COMP_DECIMAL_64_SIGNIFICAND_LEN_CASE_TWO) & 
      ((((uint64_t) 1) << __COMP_DECIMAL_64_EXPONENT_LEN_CASE_TWO) - ((uint64_t) 1));
  }
  if (M > __COMP_DECIMAL_64_MAX_SIGNIFICAND) {
    M = ((uint64_t) 0);
  }

  /* Check if the input is zero */
  if (M == ((uint64_t) 0)) {
    /* The input is zero */
    *class = COMP_FP_CLASS_ZERO;
    return;
  }

  /* The input is a non-zero real

     Set the significand and the unbiased exponent.

  */
  *class = COMP_FP_CLASS_NONZERO_REAL; 
  *significand = comp_uint64_to_comp_uint64(M);
  *exponent = ((int64_t) E) - ((int64_t) (__COMP_DECIMAL_64_BIAS));
}

STATIC INLINE void comp_fp_decompose_decimal_128(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
						 int64_t * RESTRICT exponent, comp_uint128_t * RESTRICT significand, 
						 comp_decimal_128_t x) {
  __comp_decimal_128_caster_t x_cast;
  __uint128_t X, C, CL, M, E, G01;

  /* Get memory representation of the IEEE754 FP number */
  x_cast.f = x;
  X = x_cast.i;
  
  /* Extract the sign */
  *sign_negative = !!((X >> (128 - 1)) != ((__uint128_t) 0));

  /* Extract classifying fields C and CL */
  C = (X << 1) >> 1;
  C >>= __COMP_DECIMAL_128_BITS_AFTER_CLASS_BITS;
  CL = C >> 1;

  /* Check for infinity */
  if (CL == __COMP_DECIMAL_128_INFINITY_CLASS) {
    /* The input is infinity 

       The sign has already been set.
    
    */
    *class = COMP_FP_CLASS_INFINITY;
    return;
  }

  /* Check for NaN */
  if (CL == __COMP_DECIMAL_128_NAN_CLASS) {
    /* The input is NaN

       Check for signaling or quiet NaNs.
    
    */
    if ((C & __COMP_DECIMAL_128_SIGNALING_NAN_BIT) != ((__uint128_t) 0)) {
      *class = COMP_FP_CLASS_SIGNALING_NAN;
    } else {
      *class = COMP_FP_CLASS_QUIET_NAN;
    }
    return;
  }
  
  /* The input is a real number 

     This implies that G_0 thru G_3 cannot be 1111.

     First get the bits G_0 and G_1, then decide which encoding case
     applies and finally get the biased exponent and the significand.

  */
  G01 = (X << 1) >> 1;
  G01 >>= __COMP_DECIMAL_128_BITS_AFTER_G0G1_BITS;
  G01 &= (__uint128_t) 0x03;
  if (G01 != ((__uint128_t) 0x03)) {
    /* Case (i) in IEEE 754-2008, Section 3.5.2 c) 2) */
    M = X & ((((__uint128_t) 1) << __COMP_DECIMAL_128_SIGNIFICAND_LEN_CASE_ONE) - ((__uint128_t) 1));
    E = (X >> __COMP_DECIMAL_128_SIGNIFICAND_LEN_CASE_ONE) & 
      ((((__uint128_t) 1) << __COMP_DECIMAL_128_EXPONENT_LEN_CASE_ONE) - ((__uint128_t) 1));
  } else {
    /* Case (ii) in IEEE 754-2008, Section 3.5.2 c) 2) */
    M = (X & ((((__uint128_t) 1) << __COMP_DECIMAL_128_SIGNIFICAND_LEN_CASE_TWO) - ((__uint128_t) 1))) | 
      (((__uint128_t) 1) << __COMP_DECIMAL_128_SIGNIFICAND_EXTRA_BIT);
    E = (X >> __COMP_DECIMAL_128_SIGNIFICAND_LEN_CASE_TWO) & 
      ((((__uint128_t) 1) << __COMP_DECIMAL_128_EXPONENT_LEN_CASE_TWO) - ((__uint128_t) 1));
  }
  if (M > __COMP_DECIMAL_128_MAX_SIGNIFICAND) {
    M = ((__uint128_t) 0);
  }

  /* Check if the input is zero */
  if (M == ((__uint128_t) 0)) {
    /* The input is zero */
    *class = COMP_FP_CLASS_ZERO;
    return;
  }

  /* The input is a non-zero real

     Set the significand and the unbiased exponent.

  */
  *class = COMP_FP_CLASS_NONZERO_REAL; 
  *significand = comp_uint128_to_comp_uint128(M);
  *exponent = ((int64_t) E) - ((int64_t) (__COMP_DECIMAL_128_BIAS));
}

/* Functions to quietize NaNs, while signaling the invalid exception
   for signaling NaNs.

   The functions return something undefined if the input is not a NaN.

*/
STATIC INLINE comp_binary_32_t comp_fp_quietize_nan_binary_32(comp_binary_32_t x) {
  VOLATILE comp_binary_32_t tmp;
  tmp = (comp_binary_32_t) 1.0f;
  return (x + tmp);
}

STATIC INLINE comp_binary_64_t comp_fp_quietize_nan_binary_64(comp_binary_64_t x) {
  VOLATILE comp_binary_64_t tmp;
  tmp = (comp_binary_64_t) 1.0;
  return (x + tmp);
}

STATIC INLINE comp_binary_128_t comp_fp_quietize_nan_binary_128(comp_binary_128_t x) {
  VOLATILE comp_binary_128_t tmp;
  tmp = (comp_binary_128_t) 1.0q;
  return (x + tmp);
}

STATIC INLINE comp_decimal_64_t comp_fp_quietize_nan_decimal_64(comp_decimal_64_t x) {
  VOLATILE comp_decimal_64_t tmp;
  tmp = (comp_decimal_64_t) 1.0dd;
  return (x + tmp);
}

STATIC INLINE comp_decimal_128_t comp_fp_quietize_nan_decimal_128(comp_decimal_128_t x) {
  VOLATILE comp_decimal_128_t tmp;
  tmp = (comp_decimal_128_t) 1.0dl;
  return (x + tmp);
}


/* Functions to signal the invalid exception for the different types */

STATIC INLINE void comp_fp_signal_invalid_binary_32() {
  VOLATILE comp_binary_32_t tmp1;
  VOLATILE comp_binary_32_t tmp2;
  VOLATILE comp_binary_32_t tmp3;
  tmp1 = (comp_binary_32_t) 0.0f;
  tmp2 = (comp_binary_32_t) 0.0f;
  tmp3 = tmp1 / tmp2;
  (void) tmp3;
}

STATIC INLINE void comp_fp_signal_invalid_binary_64() { 
  VOLATILE comp_binary_64_t tmp1;
  VOLATILE comp_binary_64_t tmp2;
  VOLATILE comp_binary_64_t tmp3;
  tmp1 = (comp_binary_64_t) 0.0;
  tmp2 = (comp_binary_64_t) 0.0;
  tmp3 = tmp1 / tmp2; 
  (void) tmp3;
}

STATIC INLINE void comp_fp_signal_invalid_binary_128() {
  VOLATILE comp_binary_128_t tmp1;
  VOLATILE comp_binary_128_t tmp2;
  VOLATILE comp_binary_128_t tmp3;
  tmp1 = (comp_binary_128_t) 0.0;
  tmp2 = (comp_binary_128_t) 0.0;
  tmp3 = tmp1 / tmp2; 
  (void) tmp3;
}

STATIC INLINE void comp_fp_signal_invalid_decimal_64() {
  VOLATILE comp_decimal_64_t tmp1;
  VOLATILE comp_decimal_64_t tmp2;
  VOLATILE comp_decimal_64_t tmp3;
  tmp1 = (comp_decimal_64_t) 0.0dd;
  tmp2 = (comp_decimal_64_t) 0.0dd;
  tmp3 = tmp1 / tmp2; 
  (void) tmp3;
}

STATIC INLINE void comp_fp_signal_invalid_decimal_128() {
  VOLATILE comp_decimal_128_t tmp1;
  VOLATILE comp_decimal_128_t tmp2;
  VOLATILE comp_decimal_128_t tmp3;
  tmp1 = (comp_decimal_128_t) 0.0dl;
  tmp2 = (comp_decimal_128_t) 0.0dl;
  tmp3 = tmp1 / tmp2; 
  (void) tmp3;
}


/* External (dynamically linkable, non-static) versions of the functions */

void comp_ext_fp_decompose_binary_32(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
				     int64_t * RESTRICT exponent, comp_uint32_t * RESTRICT significand, 
				     comp_binary_32_t x);
void comp_ext_fp_decompose_binary_64(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
				     int64_t * RESTRICT exponent, comp_uint64_t * RESTRICT significand, 
				     comp_binary_64_t x);
void comp_ext_fp_decompose_binary_128(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
				      int64_t * RESTRICT exponent, comp_uint128_t * RESTRICT significand, 
				      comp_binary_128_t x);
void comp_ext_fp_decompose_decimal_64(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
				      int64_t * RESTRICT exponent, comp_uint64_t * RESTRICT significand, 
				      comp_decimal_64_t x);
void comp_ext_fp_decompose_decimal_128(int * RESTRICT sign_negative, comp_fp_class_t * RESTRICT class, 
				       int64_t * RESTRICT exponent, comp_uint128_t * RESTRICT significand, 
				       comp_decimal_128_t x);
comp_binary_32_t comp_ext_fp_quietize_nan_binary_32(comp_binary_32_t x);
comp_binary_64_t comp_ext_fp_quietize_nan_binary_64(comp_binary_64_t x);
comp_binary_128_t comp_ext_fp_quietize_nan_binary_128(comp_binary_128_t x);
comp_decimal_64_t comp_ext_fp_quietize_nan_decimal_64(comp_decimal_64_t x);
comp_decimal_128_t comp_ext_fp_quietize_nan_decimal_128(comp_decimal_128_t x);
void comp_ext_fp_signal_invalid_binary_32();
void comp_ext_fp_signal_invalid_binary_64();
void comp_ext_fp_signal_invalid_binary_128();
void comp_ext_fp_signal_invalid_decimal_64();
void comp_ext_fp_signal_invalid_decimal_128();


#endif

