#!/usr/bin/env python
# coding: utf-8

# This script requires Sollya[1] and the pythonsollya[2] interface
#
# [1] http://sollya.gforge.inria.fr/
# [2] https://scm.gforge.inria.fr/anonscm/git/metalibm/pythonsollya.git

from __future__ import print_function

import collections, itertools, math, textwrap, sys

import pythonsollya
from pythonsollya import log, floor, ceil, nearestint

################################################################################

class Printer(object):
    def __init__(self):
        self._indent = 0
    def emit(self, *words):
        text = ' '.join(str(w) for w in words)
        for line in str.splitlines(textwrap.dedent(text)):
            if line:
                print('    '*self._indent, end='')
            print(line)
    def indent(self):
        self._indent += 1
    def dedent(self):
        self._indent -= 1

pr = Printer()
emit = pr.emit

################################################################################

class FloatingPointType(object):
    def __init__(self, radix_name, bits):
        self.bits = bits
        self.radix_name = radix_name
    def __str__(self):
        return "comp_{}_{}_t".format(self.radix_name, self.bits)
    def emin(self):
        return -self.emax() + 1

class BinaryFPtype(FloatingPointType):
    def __init__(self, bits):
        super(self.__class__, self).__init__('binary', bits)
    def prec(self):
        return {32:24, 64:53, 128:113}[self.bits]
    def emax(self):
        return {32:127, 64:1023, 128:16383}[self.bits]

class DecimalFPtype(FloatingPointType):
    def __init__(self, bits):
        super(self.__class__, self).__init__('decimal', bits)
    def prec(self):
        return {64:16, 128:34}[self.bits]
    def emax(self):
        return {64:384, 128:6144}[self.bits]

class uint(object):
    def __init__(self, bits):
        self.bits = bits
    def __str__(self):
        return "comp_uint{}_t".format(self.bits)
    def const(self, val):
        if self.bits % 64 != 0:
            raise NotImplementedError
        wd_bits = 64
        wd_modulus = 1 << wd_bits
        args = ', '.join([
            str((val >> (k*wd_bits))%wd_modulus)
            for k in xrange(self.bits//wd_bits-1, -1, -1)])
        return 'COMP_UINT{}_CONSTANT({})'.format(self.bits, args)

################################################################################

class Variable(object):
    def __init__(self, type, name):
        self.type = type
        self.name = name
    def __str__(self):
        return self.name
    @property
    def bits(self):
        return self.type.bits

class VariableManager(object):
    def __init__(self):
        self.variables = []
    def new(self, type, name):
        var = Variable(type, name)
        self.variables.append(var)
        return var
    def emit_declarations(self):
        bytype = collections.defaultdict(list)
        for var in self.variables:
            bytype[var.type].append(var.name)
        for type in bytype.iterkeys():
            emit(str(type) + ' ' + ', '.join(bytype[type]) + ';')

def fp_decompose(isneg, cls, expo, mant, src):
    return ('''\
            {isneg} = 0; {expo} = 0; memset(&{mant}, 0, sizeof({mant})); // silence compiler warnings
            comp_fp_decompose_{radix}_{bits}(&{isneg}, &{cls}, &{expo}, &{mant}, {src});
            '''.format(
                radix=src.type.radix_name, bits=src.type.bits, isneg=isneg,
                cls=cls, expo=expo, mant=mant, src=src))

def shift(dest, src, amount, dir):
    if dest.bits != src.bits:
        raise ValueError('{}b src, {}b dest'.format(src.bits, dest.bits))
    cmt = '// ' if dest is src and amount == 0 else ''
    return "{cmt}comp_shift_{dir}_{bits}(&{dest}, {src}, {amount});".format(
            cmt=cmt, dir=dir, bits=src.bits, dest=dest, src=src, amount=amount)

def shift_left(*args): return shift(*args, dir='left')
def shift_right(*args): return shift(*args, dir='right')

def leading_zeros(src):
    return "comp_leading_zeros_{}({})".format(src.bits, src)

def mul(dest, src1, src2):
    if dest.bits != src1.bits + src2.bits:
        raise ValueError("sizes do not match")
    return "comp_mul_{}_{}_{}(&{}, {}, {});".format(
            src1.bits, src2.bits, dest.bits,
            dest, src1, src2)

def downconvert(dest, src, part):
    if dest.bits < src.bits:
        return "comp_downconvert_{}_{}_{}(&{}, {});".format(
                part, src.bits, dest.bits, dest, src)
    elif dest.bits == src.bits:
        return "{} = {};".format(dest, src)
    else:
        raise ValueError("cannot downconvert to larger type")

def upconvert(dest, src, part):
    if dest.bits > src.bits:
        return "comp_upconvert_{}_{}_{}(&{}, {});".format(
                part, src.bits, dest.bits, dest, src)
    elif dest.bits == src.bits:
        return "{} = {};".format(dest, src)
    else:
        raise ValueError("cannot upconvert to smaller type")

def sub(dest, left, right):
    if not (dest.bits == left.bits == right.bits):
        raise ValueError
    return "comp_sub_{}(&{}, {}, {});".format(dest.bits, dest, left, right)

def sign(val):
    return "comp_sign_{}({})".format(val.bits, val)

################################################################################

pythonsollya.prec(1000)
pythonsollya.display(pythonsollya.powers)
five = pythonsollya.SollyaObject(5)
two = pythonsollya.SollyaObject(2)
log5_of_2 = log(two)/log(five)
log2_of_5 = log(five)/log(two)

def ceil_theta1(p, q):
    theta1 = pythonsollya.ceil(five**(p.gamma*q) * two**(p.lambda1 - 1 - psi(p.gamma*q)))
    tmp = str(theta1)
    return long(eval(tmp.replace('^', '**')))

def phi(h): return int(floor(h*log5_of_2))
def psi(g): return int(floor(g*log2_of_5))

def params_psi(p):
    def search(argrange):
        for shift in itertools.count(0):
            cst = int(nearestint(log2_of_5 * (1 << shift)))
            def approx(g):
                return (cst*g) >> shift
            if all(psi(g) == approx(g) for g in argrange):
                return shift, cst
    shift1, magic1 = search(range(p.gamma*p.qmin, p.gamma*p.qmax+1, p.gamma))
    shift2, magic2 = search(range(0, p.gamma))
    if shift1 > shift2:
        return (shift1, magic1)
    else:
        return (shift2, magic2)

################################################################################

def emit_table_theta1(p):
    emit('static const {uint} theta1_bin{bin}_dec{dec}[] = {{'.format(
        uint=uint(p.Wlambda1), bin=p.bintype.bits, dec=p.dectype.bits))
    pr.indent()
    for q in range(p.qmin, p.qmax+1):
        val = ceil_theta1(p, q)
        emit('{}{}'.format(
            uint(p.Wlambda1).const(val),
            ',' if q < p.qmax else ''))
    pr.dedent()
    emit('};')

def emit_table_theta2(p):
    emit('static const {uint} theta2_bin{bin}_dec{dec}[] = {{'.format(
        uint=uint(p.Wlambda2), bin=p.bintype.bits, dec=p.dectype.bits))
    pr.indent()
    for r in range(0, p.gamma):
        val = (5**r) * (1 << (p.lambda2 - 1 - psi(r)))
        emit('{}{}'.format(
            uint(p.Wlambda2).const(val),
            ',' if r < p.gamma - 1 else ''))
    pr.dedent()
    emit('};')

def emit_compare(p):
    emit('''\
        int comp_le_bin{binfmt}_dec{decfmt}({bintype} x, {dectype} y) {{\n
        '''.format(
            binfmt=p.bintype.bits, decfmt=p.dectype.bits,
            bintype=p.bintype, dectype=p.dectype))
    pr.indent()

    var = VariableManager()
    xIsNeg = var.new('int', 'xIsNeg')
    yIsNeg = var.new('int', 'yIsNeg')
    xClass = var.new('comp_fp_class_t', 'xClass')
    yClass = var.new('comp_fp_class_t', 'yClass')
    nu = var.new('int_fast8_t', 'nu')
    e2 = var.new(p.small_int, 'e2')
    e10 = var.new(p.small_int, 'e10')
    h = var.new(p.small_int, 'h')
    g = var.new(p.small_int, 'g')
    q = var.new(p.small_int, 'q')
    r = var.new(p.small_int, 'r')
    phi = var.new(p.small_int, 'phi')
    sigma = var.new(p.small_int, 'sigma')
    M10 = var.new(uint(p.Wpp10), 'M10')
    n = var.new(uint(p.Wpp10), 'n')
    tmp_M2 = var.new(uint(p.bintype.bits), 'tmp_M2')
    m = var.new(uint(p.Wp2), 'm')
    if p.epsilon == 1:
        tmpA = var.new(uint(p.Wlambda1 + p.Wpp10), 'tmpA')
        tmpB = var.new(uint(p.Wlambda2 + p.Wp2), 'tmpB')
    else:
        tmpA = var.new(uint(p.Wlambda1 + p.Wp2), 'tmpA')
        tmpB = var.new(uint(p.Wlambda2 + p.Wpp10), 'tmpB')
    A = var.new(uint(p.Wlambda1), 'A')
    B = var.new(uint(p.Wlambda1), 'B')
    ceil_theta1_q = var.new(uint(p.Wlambda1), 'ceil_theta1_q')
    theta2 = var.new(uint(p.Wlambda2), 'theta2')
    var.emit_declarations()
    emit('\n')

    emit(fp_decompose(xIsNeg, xClass, e2, tmp_M2, Variable(p.bintype, 'x')))
    emit(upconvert(m, tmp_M2, 'low'))
    emit("e2 = e2 + {} - 1;".format(p.p2))
    emit(fp_decompose(yIsNeg, yClass, e10, M10, Variable(p.dectype, 'y')))
    emit("e10 = e10 + {} - 1;".format(p.p10))
    emit('''
        /* Step 0: Zeros, infinities, NaNs, and finite numbers of different sign */
        if (   xClass == COMP_FP_CLASS_SIGNALING_NAN
            || xClass == COMP_FP_CLASS_QUIET_NAN
            || yClass == COMP_FP_CLASS_SIGNALING_NAN
            || yClass == COMP_FP_CLASS_QUIET_NAN)
        {
        ''' +
        "comp_fp_signal_invalid_binary_{}();".format(p.bintype.bits)
        + '''
            return 0;
        }
        if (xClass == COMP_FP_CLASS_ZERO || yClass == COMP_FP_CLASS_ZERO) {
            if (xClass == COMP_FP_CLASS_ZERO && yClass == COMP_FP_CLASS_ZERO)
                return 1;
            if (xClass == COMP_FP_CLASS_ZERO)
                return !yIsNeg;
            return xIsNeg;
        }
        if (xIsNeg != yIsNeg)
            return xIsNeg;
        if (xClass == COMP_FP_CLASS_INFINITY || yClass == COMP_FP_CLASS_INFINITY) {
            if (xClass == COMP_FP_CLASS_INFINITY && yClass == COMP_FP_CLASS_INFINITY)
                return 1;
            if (xClass == COMP_FP_CLASS_INFINITY)
                return xIsNeg;
            return !yIsNeg;
        }\n
    ''')

    emit("/* Step 1: Do the exponents provide enough information to conclude? */\n")
    # Normalisation -- ν = lzc(M10) - (bits(M10) - p'10)
    emit(nu, "=", leading_zeros(M10), "-", M10.bits - p.pp10, ";")
    emit(shift_left(n, M10, nu))
    emit(h, "=", e2, "-", e10, "+", nu, "+", p.p10, "-", p.pp10, "+ 1;")
    emit(g, "=", e10, "-", p.p10, "+ 1;")
    L = int(nearestint(2**p.s*log5_of_2))
    emit("phi = ({}*{}) >> {};".format(L, h, p.s))
    emit('''\
        if (likely(phi != g)) {
            if (phi < g)
                return !xIsNeg;
            else
                return xIsNeg;
        }
        ''')
    emit('\n')

    emit("/* Step 2: Look at the mantissas */")
    # 2.1 Compute q, r and sigma
    lg_gamma = int.bit_length(p.gamma) - 1
    if p.gamma == 1 << lg_gamma:
        emit(q, "=", "-((-({})*g) >> {});".format(p.epsilon, lg_gamma))
        emit(r, "=", "{}*q - ({})*g;".format(p.gamma, p.epsilon))
    else:
        raise NotImplementedError("gamma not power of two")
    # 2.2 Read θ1
    emit(ceil_theta1_q, '=', 'theta1_bin{}_dec{}[q-({})];'.format(
        p.bintype.bits, p.dectype.bits, p.qmin))
    # 2.3 Compute A
    emit(shift_left(ceil_theta1_q, ceil_theta1_q, max(0, p.tau)))
    if p.epsilon == +1:
        emit(shift_left(n, n, p.Wpp10 - p.pp10 - max(0, -p.tau)))
        emit(mul(tmpA, ceil_theta1_q, n))
    else:
        emit(shift_left(m, m, p.Wp2 - p.bintype.prec() - max(0, -p.tau) - 1))
        emit(mul(tmpA, ceil_theta1_q, m))
    emit(downconvert(A, tmpA, 'high'))
    # 2.4 Compute B
    emit(theta2, "=", 'theta2_bin{}_dec{}[r];'.format(
        p.bintype.bits, p.dectype.bits))
    psi_shift, psi_magic = params_psi(p)
    psis = ("((({gamma}*q*{magic}) >> {sh}) - ((r*{magic}) >> {sh}))"
            .format(gamma=p.gamma, magic=psi_magic, sh=psi_shift))
    emit(sigma, "=", "h - ({})*{};".format(p.epsilon, psis))
    if p.epsilon == +1:
        emit(shift_left(m, m, "{} + sigma".format(p.omega)))
        emit(mul(tmpB, theta2, m))
    else:
        emit(shift_left(n, n, "{} - sigma".format(p.omega)))
        emit(mul(tmpB, theta2, n))
    emit(downconvert(B, tmpB, 'high'))
    # 2.5 Subtract
    emit(sub(A, A, B))
    # 2.6 Mask
    if not p.omit_step_6:
        emit(shift_right(A, A, p.tau + 1))
        emit(shift_left(A, A, p.tau + 1))
    # 2.7 Conclude
    emit('''\
        switch({sgn}) {{
            case 0: return 1;
            case {eps}: return !xIsNeg;
            case {meps}: return xIsNeg;
            default: return 42;
        }}'''
        .format(sgn=sign(A), eps=p.epsilon, meps=-p.epsilon))
    pr.dedent()
    emit('\n}\n')

def emit_header():
    emit('''\
        /* Automatically generated. */

        #include <stdio.h>
        #include <string.h>
        #include "fp_utils.h"

        #define likely(COND) __builtin_expect(COND, 1)
        #define unlikely(COND) __builtin_expect(COND, 0)
        ''')

class ImplementationParameters(object):
    def __init__(self, **fields):
        self.__dict__.update(fields)
        self.p2 = self.bintype.prec()
        self.p10 = self.dectype.prec()
        self.pp10 = int(math.ceil(self.p10*log(10)/log(2)))
        self.w = self.pp10 - self.p2 - 1
        self.small_int = 'int64_t'
        if self.epsilon == 1:
            self.omega = (self.Wp2 - self.p2 + self.Wlambda2 - self.lambda2
                    - self.Wlambda1 + self.lambda1 + self.tau - 1)
        else:
            self.omega = (self.Wpp10 - self.pp10 + self.Wlambda2 - self.lambda2
                    - self.Wlambda1 + self.lambda1 + self.tau)
        # not 100% rigorous
        hmin2 = int(math.ceil(
            (self.bintype.emin() - self.bintype.prec() - self.pp10 + 3)
            /(1 + log5_of_2)))
        hmax2 = int(math.floor((self.bintype.emax() + 2)/(1 + log5_of_2)))
        gmin2 = phi(hmin2)
        gmax2 = phi(hmax2)
        self.qmin, self.qmax = sorted([
            int(math.ceil(self.epsilon*gmin2*1./self.gamma)),
            int(math.ceil(self.epsilon*gmax2*1./self.gamma))])

def emit_impl(p):
    emit_table_theta1(p)
    emit('\n')
    emit_table_theta2(p)
    emit('\n')
    emit_compare(p)
    emit('\n')
    emit('''\
        int empty_func_bin{binbits}_dec{decbits}({bintype} x, {dectype} y) {{
            return 1;
        }}
        '''.format(
            binbits=p.bintype.bits, decbits=p.dectype.bits,
            bintype=p.bintype, dectype=p.dectype)
        )


parameters_optimizing_table_size_64 = {
    'b32d64': ImplementationParameters(
        bintype=BinaryFPtype(32), dectype=DecimalFPtype(64),
        s = 18, # actually not that hard to compute on the fly
        Wp2=64, Wpp10=64,
        epsilon=+1, gamma=16,
        lambda1=125, Wlambda1=128,
        lambda2=64, Wlambda2=64,
        tau=0, omit_step_6=True),
    'b32d128': ImplementationParameters(
        bintype=BinaryFPtype(32), dectype=DecimalFPtype(128),
        s = 23, # actually not that hard to compute on the fly
        Wp2=64, Wpp10=128,
        epsilon=-1, gamma=16,
        lambda1=191, Wlambda1=192,
        lambda2=64, Wlambda2=64,
        tau=0, omit_step_6=True),
    'b64d64': ImplementationParameters(
        bintype=BinaryFPtype(64), dectype=DecimalFPtype(64),
        s = 19, # actually not that hard to compute on the fly
        Wp2=64, Wpp10=64,
        epsilon=+1, gamma=16,
        lambda1=125, Wlambda1=128,
        lambda2=64, Wlambda2=64,
        tau=0, omit_step_6=True),
    'b64d128': ImplementationParameters(
        bintype=BinaryFPtype(64), dectype=DecimalFPtype(128),
        s = 23, # actually not that hard to compute on the fly
        Wp2=64, Wpp10=128,
        epsilon=-1, gamma=32,
        lambda1=191, Wlambda1=192,
        lambda2=128, Wlambda2=128,
        tau=0, omit_step_6=False),
    'b128d64': ImplementationParameters(
        bintype=BinaryFPtype(128), dectype=DecimalFPtype(64),
        s = 27, # actually not that hard to compute on the fly
        Wp2=128, Wpp10=64,
        epsilon=-1, gamma=64,
        lambda1=191, Wlambda1=192,
        lambda2=192, Wlambda2=192,
        tau=0, omit_step_6=False),
    'b128d128': ImplementationParameters(
        bintype=BinaryFPtype(128), dectype=DecimalFPtype(128),
        s = 27, # actually not that hard to compute on the fly
        Wp2=128, Wpp10=128,
        epsilon=+1, gamma=64,
        lambda1=253, Wlambda1=256,
        lambda2=192, Wlambda2=192,
        tau=0, omit_step_6=False),
}

def gen():
    emit_header()
    # p = parameters_optimizing_table_size_64['b32d128']
    # emit_impl(p)
    for p in parameters_optimizing_table_size_64.itervalues():
       emit('\n')
       emit_impl(p)

gen()

