#!/usr/bin/env maple

# We are looking for rational approximations (not necessarily irreducible) p/q
# of various numbers alpha with min_numer <= abs(p) <= max_numer and
# min_denom <= q <= max_denom, and with abs(alpha - p/q) <= 1/max_denom^2.

printlevel := -1;
# infolevel[checkcfrac] := 2;

# binfmt := 32;
# decfmt := 64;

if binfmt = 32 then
    prec2 := 24;
    e2min := -126;
    e2max := 127;
elif binfmt = 64 then
    prec2 := 53;
    e2min := -1022;
    e2max := 1023;
elif binfmt = 128 then
    prec2 := 113;
    e2min := -16382;
    e2max := 16383;
end if;

if decfmt = 64 then
    prec10 := 16;
    e10min := -383;
    e10max := 384;
elif decfmt = 128 then
    prec10 := 34;
    e10min := -6143;
    e10max := 6144;
end if;

pp10 := ceil(prec10 * log[2](10));

min_numer := 2^(prec2-1);
max_numer := 2^prec2-1;
min_denom := 2^(pp10-1);
max_denom := 2^pp10-1;

hmin := ceil((e2min-prec2-pp10+3)/(1+log[5](2)));
hmax := ceil((e2max+2)/(1+log[5](2))) - 1;

h0 := ceil((1+e2max-pp10+2)/(1+log[5](2)));

admissible_n := (n, h) ->
    (n >= 10^prec10 implies irem(n, 2) = 0)
    and
    (h >= h0 implies irem(n, 2^(h + phi(h) - e2max + pp10 - 2)) = 0);

dist := table();
worst := table();

update := proc(p, q)
    local j;
    global bound;
    global dist;
    global worst;
    if abs(alpha-p/q) < dist[h] then
        if q > bound then
            error "bug h=%1 p=%2 q=%3", h, p, q;
        elif q < min_denom then
            # Try to scale both p and q up to the allowed ranges. We try
            # all possible scalings to avoid funny interactions with the
            # filtering done on q.
            userinfo(1, checkcfrac, "h"=h, "denom too small", "q"=q);
            for j from ceil(min_denom/q) to floor(bound/q) do
                userinfo(1, checkcfrac, "h"=h, "scaling up", "j"=j);
                update(j*p, j*q, h);
            end do;
        elif abs(p) > max_numer then
            # Hopefully never happens in our examples.
            error "bad luck"
        elif abs(p) < min_numer then
            # The closest admissible rational is at dist >= 1/max_denom^2.
            userinfo(1, checkcfrac, "h"=h, "numer too small", "p"=p, "q"=q);
        elif not admissible_n(q, h) then
            # Idem.
            userinfo(1, checkcfrac, "h"=h, "not admissible", "p"=p, "q"=q);
        else
            # We have found a new worst case.
            dist[h] := abs(alpha-p/q);
            worst[h] := [p, q];
            userinfo(1, checkcfrac, "h"=h, "new worst case", "p"=p, "q"=q,
                "lgdist"=evalf[6](log[2](dist[h])));
        end if;
    end if;
end proc:

phi := h -> floor(h * log(2)/log(5));

for h from hmin to hmax do
    alpha := 5^phi(h)/2^(h-prec2+pp10-1);
    # Since abs(alpha-p/q) <= max_denom^2 and abs(p) <= max_numer, the
    # denominator q must be bounded by the following.
    bound := min(max_denom, (max_numer + 1/max_denom)/abs(alpha));
    dist[h] := 1/max_denom^2;
    # The min dist between two rationals of denom <= max_denom is 1/max_denom².
    # In particular, when denom(alpha) itself is <= max_denom, there is no
    # "interesting" approximation.
    if denom(alpha) < max_denom then next end if;
    use numtheory in
        cf := cfrac(alpha);
        # Look for the "last" convergent pn/qn with qn <= bound.
        for n while nthdenom(cf, n+1) <= bound do end do;
        pl, pm, pr := seq(nthnumer(cf, n+k), k=-1..1);
        ql, qm, qr := seq(nthdenom(cf, n+k), k=-1..1);
    end use;
    update(pm, qm);
    # Now consider semiconvergents in order to find the rational nearest to
    # alpha under the constraint q <= bound.
    for k from 1 while ql+k*qm <= bound do
        if igcd(pl+k*pm, ql+k*qm) <> 1 then error "bug2" end if;
        update(pl+k*pm, ql+k*qm);
    end do;
end do:

#print(dist);

worsth := sort(map(op, [indices(dist)]), proc(i,j) dist[i]<dist[j] end)[1..10];
printf("\nworst h: %a\nbit prec: %a, [p,q]: %a\n\n",
    worsth,
    evalf(-log[2](dist[worsth[1]])),
    worst[worsth[1]]);

plots:-pointplot([seq([h, evalf(log[2](dist[h]))], h=hmin..hmax)]):
