# Format parameters

all_word_sizes := [32, 64]:

all_sizes := [seq(seq([bsz, dsz], dsz in [64, 128]), bsz in [32, 64, 128])]:

binfmt := proc(size)  # p2, e2min, e2max
    if size = 32 then
        24, -126, 127;
    elif size = 64 then
        53, -1022, 1023;
    elif size = 128 then
        113, -16382, 16383;
    end if;
end proc:

decfmt := proc(size)  # p10, e10min, e10max
    if size = 64 then
        16, -383, 384;
    elif size = 128 then
        34, -6143, 6144;
    end if;
end proc:

# Notations

phi := h -> floor(h*log[5](2)):
psi := g -> floor(g*log[2](5)):

# For ranking the solutions we find

sort_helper := proc()
    local i, conds;
    conds := [_passed];
    for i to nops(conds) do
        if conds[i] = 0 then
            next;
        else
            return conds[i] = 1
        end if;
    end do;
    return false;
end proc:

mulcountfirst := proc(a,b)
    sort_helper(
        signum(b[muls]-a[muls]),
        signum(b[totaltblbytes]-a[totaltblbytes]),
        `if`(a[ok_1rnd_tau], 1, 0) - `if`(b[ok_1rnd_tau], 1, 0),
        # favor τ=0 over τ=-1 if two choices of (ε,γ) yield otherwise equivalent solutions
        a[tau]-b[tau],
        NULL
    );
end proc:

tblszfirst := proc(a,b)
    sort_helper(
        signum(b[totaltblbytes]-a[totaltblbytes]),
        signum(b[muls]-a[muls]),
        `if`(a[ok_1rnd_tau], 1, 0) - `if`(b[ok_1rnd_tau], 1, 0),
        a[tau]-b[tau],
        NULL
    );
end proc:

# Stuff that is too expensive to recompute every time

smintbl[32,64]:=18:
smintbl[32,128]:=23:
smintbl[64,64]:=19:
smintbl[64,128]:=23:
smintbl[128,64]:=27:
smintbl[128,128]:=27:

wcbitstbl[32,64]:=112:
wcbitstbl[32,128]:=230:
wcbitstbl[64,64]:=114:
wcbitstbl[64,128]:=234:
wcbitstbl[128,64]:=127:
wcbitstbl[128,128]:=238:

# Main proc: given a binary format, a decimal format and a machine word size,
# compute a module providing lots of information on possible instantiations of
# the algorithm.
params := proc(binsz, decsz, wordsz, nbest:=1)
    option remember;
    module()
        local W;
        export
            p2, e2min, e2max, p10, e10min, e10max, pp10, w, hmin1, hmax1,
            smin, phi_bits,
            hmin2, hmax2, gmin2, gmax2, gcount2,
            gmineq, gmaxeq, gcounteq, hmineq, hmaxeq,
            h0,
            lambda1min, W_lambda1, gamma_max,
            with_split, all_with_gamma_2pow, all_with_any_gamma, best,
            best_with_gamma_2pow_mulcountfirst,
            best_with_gamma_2pow_tblszfirst, best_with_any_gamma_mulcountfirst,
            best_with_any_gamma_tblszfirst,
            W_p2, W_pp10;

        W := proc(k, $)
            ceil(k/wordsz)*wordsz;
        end proc:

        p2, e2min, e2max := binfmt(binsz);
        p10, e10min, e10max := decfmt(decsz);

        pp10 := ceil(p10 * log[2](10));
        w := pp10 - p2 - 1;

        W_p2 := W(p2);
        W_pp10 := W(pp10);

        # Range of h for Step 1
        hmin1 := (e2min - p2 + 1) - e10max + p10 - pp10 + 1;
        hmax1 := e2max - e10min + p10;

        # Used in the computation of phi
        # (a bit tricky to compute fast)
        smin := smintbl[binsz, decsz];

        # 1+ to allow signed types...
        phi_bits := 2^(ceil(log[2](1+ceil(log[2](
            round(2^smin*log[5](2))*max(hmax1, -hmin1))))));

        # Range of h for which we may reach Step 2
        hmin2 := ceil((e2min - p2 - pp10 + 3)/(1 + log[5](2)));
        hmax2 := floor((e2max + 2)/(1 + log[5](2)));
        gmin2 := phi(hmin2);
        gmax2 := phi(hmax2);
        gcount2 := gmax2-gmin2+1;

        # Range of h and g for which we may have x2 = x10.
        gmaxeq := floor(p2*log[5](2));
        hmaxeq := p2;
        gmineq := -floor(p10*log[5](10));
        hmineq := -floor(p10*(1+log[2](5)));
        #if p2 >= 1 then
        #    gmineq := max(gmineq, - ceil((pp10 - 2)*log[5](2)));
        #    hmineq := max(hmineq, -pp10 + 1);
        #end if
        gcounteq := gmaxeq-gmineq+1;

        # Min value of h for which n is necessarily even (if Step 2 is reached)
        h0 := ceil((e2max - pp10 + 3)/(1 + log[5](2)));

        # Min number of bits of 5^... to store in the approximate table (?) and
        # use in arithmetic operations. Note that the *word size* on which
        # table entries are represented will typically be slightly larger.
        # This currently does NOT take into account "inverse worst cases";
        # doing it might lead to slightly better final bounds.
        lambda1min := wcbitstbl[binsz, decsz] - w + 3;
        # A reasonable word size for lambda1 (TBI?)
        W_lambda1 := W(lambda1min);
        # WARNING: check conditions of Algo 2

        gamma_max := phi(lambda1min)+1;

        with_split := proc(_gamma, _epsilon)
            local m := module()
                export gamma := _gamma;
                export epsilon := _epsilon;

                export qbounds :=sort([
                    ceil(epsilon*gmin2/_gamma),
                    ceil(epsilon*gmax2/_gamma)]);
                export tbl1bytes := (qbounds[2]-qbounds[1]+1)*W_lambda1/8;

                # Min number of bits in the exact table, corresponding table
                # size
                export lambda2min := psi(_gamma-1)+1;
                export W_lambda2 := W(lambda2min);
                export tbl2bytes := _gamma*W_lambda2/8;

                export totaltblbytes := tbl1bytes + tbl2bytes;

                export muls := `if`(epsilon=1,
                    ceil(W_lambda1/wordsz)*ceil(W(pp10)/wordsz)
                    +ceil(W_lambda2/wordsz)*ceil(W(p2)/wordsz),
                    ceil(W_lambda1/wordsz)*ceil(W(p2)/wordsz)
                    +ceil(W_lambda2/wordsz)*ceil(W(pp10)/wordsz));

                export lambda1, lambda2, tau, ok_1rnd_tau;
            end module;

            # At this point, there remains to choose λ1, λ2, and τ. We prefer
            # to do it in such a way that the condition to omit the masking
            # step holds. This requires τ<=0, and τ=-1 only makes that
            # condition stricter (but might help satisfy the second condition
            # on W in funny cases with W(λ1)=W(λ2) and λ2 slightly *larger*
            # than λ1).

            local constraints, _lambda1, _lambda2, _tau, lpsol;

            constraints := {
                lambda1min <= _lambda1, _lambda1 <= W_lambda1,
                m[lambda2min] <= _lambda2, _lambda2 <= m[W_lambda2],
                `if`(_epsilon=1, W(pp10)-pp10 >= 1, W(p2)-p2 >=2), # sanity check (constants)
                W_lambda1-_lambda1 >= m[W_lambda2]-_lambda2 + _tau + `if`(_epsilon=1, 3, 1),
                `if`(_epsilon=1,
                    W(p2)-p2+m[W_lambda2]-_lambda2 >= W_lambda1-_lambda1-_tau+1,
                    W(pp10)-pp10+m[W_lambda2]-_lambda2 >= W_lambda1-_lambda1-_tau+3),
                NULL
            };

            # WARNING: for some reason, LPSolve doesn't work with the CLI on
            # my machine (only with maple -x)!
            try
                # First attempt to find a solution (with given γ and ε) that
                # does not require the masking step
                lpsol := Optimization:-LPSolve(
                    # maximize λ1, λ2 for table reuse opportunities,
                    # τ because cases with τ=0 are easier to understand
                    -_lambda1-_lambda2-1000*_tau,
                    constraints union
                        {_lambda1-_lambda2+_tau >= `if`(_epsilon=1, p2+1, pp10+3)},
                    _tau = -1..0,
                    assume=integer);
                m:-ok_1rnd_tau := true;
            catch:
                try
                    # Now try with the masking step
                    m:-ok_1rnd_tau := false;
                    lpsol := Optimization:-LPSolve(
                        -_lambda1-_lambda2+1000*_tau, # minimize τ but avoid τ=-1
                        constraints union {_tau >= 0},
                        assume=integer);
                catch:
                    # Finally, the problem may not be feasible at all
                    return NULL;
                end try;
            end try;

            m:-lambda1 := subs(eval(lpsol)[2], _lambda1);
            m:-lambda2 := subs(eval(lpsol)[2], _lambda2);
            m:-tau := subs(eval(lpsol)[2], _tau);

#            m:-omega := `if`(epsilon=1,
#                W(p2)-p2+W_lambda2-lambda2-W_lambda1+lambda1+tau-1,
#                W(pp10)-pp10+W_lambda2-lambda2-W_lambda1+lambda1+tau);
#
#            m:-bits_to_drop_in_B := -`if`(epsilon=1,
#                W_lambda1 - W_lambda2 - W(p2),
#                W_lambda1 - W_lambda2 - W(pp10));
#
            return m;
        end proc:

        all_with_gamma_2pow := [seq(
            seq(with_split(2^_lggamma, epsilon), epsilon in [-1,1]),
            _lggamma=0..ilog2(gamma_max))];

        all_with_any_gamma := [seq(
            seq(with_split(_gamma, epsilon), epsilon in [-1,1]),
            _gamma=1..gamma_max)];

        best_with_gamma_2pow_mulcountfirst := sort(all_with_gamma_2pow, mulcountfirst)[1..nbest];
        best_with_gamma_2pow_tblszfirst := sort(all_with_gamma_2pow, tblszfirst)[1..nbest];
        best_with_any_gamma_mulcountfirst := sort(all_with_any_gamma, mulcountfirst)[1..nbest];
        best_with_any_gamma_tblszfirst := sort(all_with_any_gamma, tblszfirst)[1..nbest];

        best := [
            best_with_gamma_2pow_mulcountfirst,
            best_with_gamma_2pow_tblszfirst,
            best_with_any_gamma_mulcountfirst,
            best_with_any_gamma_tblszfirst,
            NULL ];

    end module;
end proc:

# Actually perform the computation

for wsz in all_word_sizes do
    for sz in all_sizes do
        p[wsz][sz] := params(op(sz), wsz, 3);
    end do:
end do:

# Display the results

for wsz in all_word_sizes do
    for sz in all_sizes do
        printf("b%d/d%d\t", op(sz));
    end do;
    printf(cat("\n", "--------\t"$(nops(all_sizes)), "\n"));
    for field in
        {exports(p[all_word_sizes[1]][all_sizes[1]], 'string')}
        minus {"with_split", "all_with_gamma_2pow", "all_with_any_gamma",
        "best_with_gamma_2pow_mulcountfirst", "best_with_gamma_2pow_tblszfirst",
        "best_with_any_gamma_mulcountfirst", "best_with_any_gamma_tblszfirst",
        "best"}
    do
        for sz in all_sizes do
            printf("%s = %a\t", field, p[wsz][sz][field]);
        end do:
        printf("\n");
    end do:
    printf("\n");
    example_list := p[all_word_sizes[1]][all_sizes[1]][best]:
    for i to nops(example_list) do
        printf("----------------------------------------------------------------------------------------------------------------------\n\n");
        for j to nops(example_list[1]) do
            for field in {exports(example_list[1][1], 'string')} minus {"constraints"} do
                for sz in all_sizes do
                    printf("%s = %a\t", field, p[wsz][sz][best][i][j][field]);
                end do:
                printf("\n");
            end do:
            printf("\n");
        end do:
    end do;
    printf("\n\n");
end do:

# LaTeX tables for the paper

table_boilerplate := proc(wsz, sizes)
    local sz, row, header, footer;
    row := proc(field, {fmt:="$%a$"})
        cat(seq(sprintf(cat(" & ", fmt), p[wsz][sz][field]), sz in sizes), "\\\\\n")
    end proc:
    header := cat(
        "% generated by parameters.mpl\n",
        "\\begin{tabular}{rcccccc}\n",
        "\\toprule\n",
        cat(seq(sprintf("& b%d/d%d ", op(sz)), sz in sizes), "\\\\\n"),
        "\\midrule\n",
        NULL);
    footer := cat(
        "\\bottomrule\n",
        "\\end{tabular}\n",
        NULL);
    row, header, footer;
end proc:

table_params_step1 := proc(wsz, sizes)
    local row, header, footer;
    row, header, footer := table_boilerplate(wsz, sizes);
    cat(
        header,
        "$p_2$", row('p2'),
        "$p_{10}$", row('p10'),
        "$p'_{10}$", row('pp10'),
        "$w$", row('w'),
        "\\noalign{\\smallskip}\n",
        "$\\hmini, \\hmaxi$",
            cat(seq(
                sprintf(" & $%a, %a$", p[wsz][sz][hmin1], p[wsz][sz][hmax1]),
                sz in sizes), "\\\\\n"),
        "$s_{\\mathrm{min}}$", row(smin),
        "datatype", row(phi_bits, fmt="\\texttt{int%a}"),
        footer,
        NULL);
end proc:

table_params_step2 := proc(wsz, sizes)
    local row, header, footer;
    row, header, footer := table_boilerplate(wsz, sizes);
    cat(
        header,
        "$\\hminii, \\hmaxii$",
            cat(seq(
                sprintf(" & $%a, %a$", p[wsz][sz][hmin2], p[wsz][sz][hmax2]),
                sz in sizes), "\\\\\n"),
        "$\\gminii, \\gmaxii$",
            cat(seq(
                sprintf(" & $%a, %a$", p[wsz][sz][gmin2], p[wsz][sz][gmax2]),
                sz in sizes), "\\\\\n"),
        "\\#$g$", row('gcount2'),
        "\\noalign{\\smallskip}\n",
        "$h_0$", row('h0'),
        footer,
        NULL);
end proc:

subtable := proc(wsz, sizes, sol, {boldface:=NULL}, $)
    local row, header, footer, row_best, sz;
    row, header, footer := table_boilerplate(wsz, sizes);
    row_best := proc(field, {fmt:="$%a$"})
        cat(seq(sprintf(cat(" & ", fmt), p[wsz][sz][sol][1][field]), sz in sizes), "\\\\\n")
    end proc:
    for sz in sizes do
        if p[wsz][sz][sol][1][tau] <> 0 then
            WARNING("tau != 0, %1", [wsz, sz, sol]);
        end if;
    end do;
    "$\\epsilon$",
        cat(seq(
            `if`(p[wsz][sz][sol][1][epsilon]=1, " & $+1$", " & $-1$"),
            sz in sizes), "\\\\\n"),
    "$\\gamma$", row_best('gamma'),
    "range of $q$",
        cat(seq(
            sprintf(" & $%a, %a$", op(p[wsz][sz][sol][1][qbounds])),
            sz in sizes), "\\\\\n"),
    "min $\\lambda_1$ satisfying \\eqref{eq:suitable lambda 1}", row('lambda1min'),
    "min $\\lambda_2$ satisfying \\eqref{eq:suitable lambda 2}", row_best('lambda2min'),
    "chosen $\\lambda_1, W(\\lambda_1)$",
        cat(seq(
            sprintf(" & $%a, %a$",
                p[wsz][sz][sol][1][lambda1],
                p[wsz][sz][W_lambda1]),
            sz in sizes), "\\\\\n"),
    "chosen $\\lambda_2, W(\\lambda_2)$",
        cat(seq(
            sprintf(" & $%a, %a$",
                p[wsz][sz][sol][1][lambda2],
                p[wsz][sz][sol][1][W_lambda2]),
            sz in sizes), "\\\\\n"),
    "$\\tau$", row_best('tau'),
    "Step~\\ref{step:mask} of Algo.~\\ref{algo:bip} can be omitted",
        cat(seq(
            `if`(p[wsz][sz][sol][1][ok_1rnd_tau], "& \\Checkmark", "& \\XSolidBrush"),
            sz in sizes), "\\\\\n"),
#    "$\\omega$ (Algorithm~\\ref{algo:bip})",
#        cat(seq(
#            sprintf(" & $%a$", op(p[wsz][sz][sol][1][omega])),
#            sz in sizes), "\\\\\n"),
#    "bits to drop in $B$",
#        cat(seq(
#            sprintf(" & $%a$", p[wsz][sz][sol][1][bits_to_drop_in_B]),
#            sz in sizes), "\\\\\n"),
    "total table size in bytes", row_best('totaltblbytes', fmt=`if`(boldface="size", "$\\mathbf{%a}$", "$%a$")),
    sprintf("$%a$-bit muls", wsz), row_best('muls', fmt=`if`(boldface="muls", "$\\mathbf{%a}$", "$%a$")),
    NULL;
end proc:

table_inst_step2 := proc(wsz, sizes)
    local row, header, footer, k, sep;
    sep := "\\midrule\n";
    row, header, footer := table_boilerplate(wsz, sizes);
    cat(
        header,
        "$W(p_2)$", row('W_p2'),
        "$W(p'_{10})$", row('W_pp10'),
        sep,
        "\\multicolumn{6}{l}{$\\gamma=2^k$, optimizing multiplication count} \\\\\n",
        subtable(wsz, sizes, best_with_gamma_2pow_mulcountfirst, boldface="muls"),
        sep,
        "\\multicolumn{6}{l}{$\\gamma=2^k$, optimizing table size} \\\\\n",
        subtable(wsz, sizes, best_with_gamma_2pow_tblszfirst, boldface="size"),
        sep,
        "\\multicolumn{6}{l}{any $\\gamma$, optimizing multiplication count} \\\\\n",
        subtable(wsz, sizes, best_with_any_gamma_mulcountfirst, boldface="muls"),
        sep,
        "\\multicolumn{6}{l}{any $\\gamma$, optimizing table size} \\\\\n",
        subtable(wsz, sizes, best_with_any_gamma_tblszfirst, boldface="size"),
        footer,
        NULL);
end proc:

fd := fopen("../IEEETC/table-step1.tex", `WRITE`):
fprintf(fd, "%s", table_params_step1(all_word_sizes[1], all_sizes)):
fclose(fd);

fd := fopen("../IEEETC/table-step2.tex", `WRITE`):
fprintf(fd, "%s", table_params_step2(all_word_sizes[1], all_sizes)):
fclose(fd);

fd := fopen("../IEEETC/table-inst-step2-32.tex", `WRITE`):
fprintf(fd, "%s", table_inst_step2(32, all_sizes)):
fclose(fd);

fd := fopen("../IEEETC/table-inst-step2-64.tex", `WRITE`):
fprintf(fd, "%s", table_inst_step2(64, all_sizes)):
fclose(fd);
