
verbosity = 2!;
dieonerrormode = on!;

/* Some helper functions and procedures */

phi = floor(x * log(2)/log(5));
psi = floor(x * log2(5));


procedure tableSizeDirectIndexByG(hmin, hmax, bitsPerEntry) {
	  var gmin, gmax, numEntries, bytesPerEntry;

	  gmin = phi(hmin);
	  gmax = phi(hmax);

	  numEntries = gmax - gmin + 1;

	  bytesPerEntry = ceil(bitsPerEntry / 8);

	  return numEntries * bytesPerEntry;
};

procedure tableSizeDirectIndexByH(hmin, hmax, bitsPerEntry) {
	  var numEntries, bytesPerEntry;

	  numEntries = hmax - hmin + 1;

	  bytesPerEntry = ceil(bitsPerEntry / 8);

	  return numEntries * bytesPerEntry;
};

procedure chooseBitsPerEntry(sizeRMax, bitsPerEntryChoices) {
	  var found, size, s;

	  bitsPerEntryChoices = sort(bitsPerEntryChoices);

	  found = false;
	  while (!found) do {
	  	s = head(bitsPerEntryChoices);
		bitsPerEntryChoices = tail(bitsPerEntryChoices);
		if (s >= sizeRMax) then {
		   found = true;
		   size = s;
		};
	  };

	  if (!found) then {
	     size = sizeRMax;
	  };

	  return size;
};

procedure tableSizeBipartite(hmin, hmax, bitsPerEntryQ, splitFactor, bitsPerEntryRChoices, ceilingEuclid) {
	  var gmin, gmax, numEntriesQ, bytesPerEntryQ;
	  var qmin, qmax, rmin, rmax;
	  var numEntriesR, bytesPerEntryR;

	  bytesPerEntryQ = ceil(bitsPerEntryQ / 8);

	  gmin = phi(hmin);
	  gmax = phi(hmax);

	  if (ceilingEuclid) then {
	      qmin = floor(gmin / splitFactor) + 1;
	      qmax = floor(gmax / splitFactor) + 1;
	      rmin = 1;            /* Check if bound is always attained */
	      rmax = splitFactor;
	  } else {
	      qmin = floor(gmin / splitFactor);
	      qmax = floor(gmax / splitFactor);
	      rmin = 0;            
	      rmax = splitFactor - 1;
	  };

	  numEntriesQ = qmax - qmin + 1;
	  numEntriesR = rmax - rmin + 1;

	  sizeRMax = ceil(log2(5^rmax));

	  bitsPerEntryR = chooseBitsPerEntry(sizeRMax, bitsPerEntryRChoices);
	  
	  bytesPerEntryR = ceil(bitsPerEntryR / 8);	

	  return { .tableSizeQ = numEntriesQ * bytesPerEntryQ, 
	           .tableSizeR = numEntriesR * bytesPerEntryR, 
		   .overallSize = numEntriesQ * bytesPerEntryQ + numEntriesR * bytesPerEntryR,
		   .bitsPerEntryR = bitsPerEntryR,
		   .splitFactorPowerOfTwo = 2^floor(log2(splitFactor)) == splitFactor };
};

procedure getPossibleSplitFactors(gmin, gmax, onlyPowersOfTwo, bitsPerEntryChoices) {
	  var splitFactorMax;
	  var sF, possibleSplitFactors;

	  splitFactorMax = min(abs(gmin), abs(gmax), floor(max(bitsPerEntryChoices) / log2(5)));

	  possibleSplitFactors = [||];
	  sF = 2;
	  while (sF <= splitFactorMax) do {
	  	possibleSplitFactors = sF .: possibleSplitFactors;
		if (onlyPowersOfTwo) then {
		   sF = 2 * sF;
		} else {
		   sF = sF + 1;
		};
	  };

	  return possibleSplitFactors;
};

procedure optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntryQ, bitsPerEntryRChoices, onlyPowersOfTwo, ceilingEuclid) {
	  var gmin, gmax;
	  var possibleSplitFactors, SF, tSB, tableSize, bestSplitFactor;

	  gmin = phi(hmin);
	  gmax = phi(hmax);

	  possibleSplitFactors = getPossibleSplitFactors(gmin, gmax, onlyPowersOfTwo, bitsPerEntryRChoices);

	  tableSize = infty;
	  bestSplitFactor = 0;
	  for sF in possibleSplitFactors do {
	      tSB = tableSizeBipartite(hmin, hmax, bitsPerEntryQ, sF, bitsPerEntryRChoices, ceilingEuclid);
	      overallSize = tSB.overallSize;
	      if (overallSize < tableSize) then {
	      	 tableSize = overallSize;
		 bestSplitFactor = sF;
	      };
	  };

	  return { .bestSplitFactor = bestSplitFactor, .tableSizeBipartite = tableSizeBipartite(hmin, hmax, bitsPerEntryQ, bestSplitFactor, bitsPerEntryRChoices, ceilingEuclid) };
};

procedure formatSize(size) {
	  var res, kB, kBI, kbF, kbFS;

	  if (size <= 1024) then {
	     res = size @ "B";
	  } else {
	     kB = size / 1024;
	     kBI = floor(kB);
	     kbF = nearestint(100 * (kB - kBI));
	     if (kbF < 10) then {
	     	kbFS = "0" @ kbF;
	     } else {
	     	kbFS = "" @ kbF;
	     };	     
	     res = kBI @ "." @ kbFS @ "kB";
	  };

	  return res;
};

procedure optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, onlyPowersOfTwo, mSize, nSize) {
	  var solA, solB, res;

	  solA = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, onlyPowersOfTwo, true);
	  solB = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, onlyPowersOfTwo, false);

	  entrySize = optRes.tableSizeBipartite.bitsPerEntryR;    
	  
	  if (abs((solA.tableSizeBipartite.bitsPerEntryR + mSize) - (bitsPerEntry + nSize)) <= 
	      abs((solB.tableSizeBipartite.bitsPerEntryR + nSize) - (bitsPerEntry + mSize))) then {
	      res = solA;
	      res.multiplyExactWithM = true;
	  } else {
	      res = solB;
	      res.multiplyExactWithM = false;
	  };

	  return res;
};

/* In the generic case we have 

   * in the case when 

     g = splitFactor * q - r  with q = floor(g/splitFactor) + 1

     that sigma is

     sigma(h) = psi(r) - psi(splitFactor * q) + h

   * in the case when

     g = splitFactor * (-q) + r with q = -floor(g/splitFactor)

     that sigma is

     sigma(h) = psi(splitFactor * (-q)) - psi(r) + h 

*/

procedure boundSigma(hmin, hmax, splitFactor, ceilingEuclid) {



};


/* The parameters for the different comparisons */

parameters = [| { .name = "b32/d64",   .hmin = -140,   .hmax = 90,    .wc = 112,   .p2 = 24,   .pp10 = 54   },
                { .name = "b32/d128",  .hmin = -181,   .hmax = 90,    .wc = 230,   .p2 = 24,   .pp10 = 113  },
                { .name = "b64/d64",   .hmin = -787,   .hmax = 716,   .wc = 114,   .p2 = 53,   .pp10 = 54   },
                { .name = "b64/d128",  .hmin = -828,   .hmax = 716,   .wc = 234,   .p2 = 53,   .pp10 = 113  },
                { .name = "b128/d64",  .hmin = -11565, .hmax = 11452, .wc = 127,   .p2 = 113,  .pp10 = 54   },
                { .name = "b128/d128", .hmin = -11606, .hmax = 11452, .wc = 238,   .p2 = 113,  .pp10 = 113  } 
             |];

/* Go over all comparisons, compute the direct table sizes and the optimal table sizes for bipartite tables */

wordSizes = [| 32, 64, 128, 192, 256 |];

/* Direct table, indexed by h */
write("Size of direct table, indexed with $h$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    tblSize = tableSizeDirectIndexByH(hmin, hmax, bitsPerEntry);
    
    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Direct table, indexed by g */
write("Size of direct table, indexed with $g$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    tblSize = tableSizeDirectIndexByG(hmin, hmax, bitsPerEntry);

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, show entry size inexact table */
write("Entry size of inexact table & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, splitFactor a power of 2, show size */
write("Size, $\\gamma= 2^t$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, true);
    tblSize = optRes.tableSizeBipartite.overallSize;

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show splitfactor */
write("Split factor $\\gamma$, $\\gamma= 2^t$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, true);
    splitFactor = optRes.bestSplitFactor;

    write(splitFactor);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show entry size exact table */
write("Entry size of exact table, $\\gamma= 2^t$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, true);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;

    write(entrySize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show multiplications */
write("Multiplications, $\\gamma= 2^t$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, true);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + mSize, " vs. ", bitsPerEntry + nSize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show accuracies */
write("Accuracies, $\\gamma= 2^t$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, true);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + mSize, " vs. ", bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, any splitFactor, show size */
write("Size, any $\\gamma$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, true);
    tblSize = optRes.tableSizeBipartite.overallSize;

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show splitfactor */
write("Split factor $\\gamma$, any $\\gamma$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, true);
    splitFactor = optRes.bestSplitFactor;

    write(splitFactor);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show entry size exact table */
write("Entry size of exact table, any $\\gamma$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, true);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;

    write(entrySize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show multiplications */
write("Multiplications, any $\\gamma$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, true);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + mSize, " vs. ", bitsPerEntry + nSize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show accuracies */
write("Accuracies, any $\\gamma$, $\\theta_2$ with $m$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, true);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + mSize, " vs. ", bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, splitFactor a power of 2, show size */
write("Size, $\\gamma= 2^t$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, false);
    tblSize = optRes.tableSizeBipartite.overallSize;

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show splitfactor */
write("Split factor $\\gamma$, $\\gamma= 2^t$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, false);
    splitFactor = optRes.bestSplitFactor;

    write(splitFactor);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show entry size exact table */
write("Entry size of exact table, $\\gamma= 2^t$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, false);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;

    write(entrySize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show multiplications */
write("Multiplications, $\\gamma= 2^t$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, false);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + nSize, " vs. ", bitsPerEntry + mSize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show accuracies */
write("Accuracies, $\\gamma= 2^t$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, true, false);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + nSize, " vs. ", bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, any splitFactor, show size */
write("Size, any $\\gamma$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, false);
    tblSize = optRes.tableSizeBipartite.overallSize;

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show splitfactor */
write("Split factor $\\gamma$, any $\\gamma$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, false);
    splitFactor = optRes.bestSplitFactor;

    write(splitFactor);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show entry size exact table */
write("Entry size of exact table, any $\\gamma$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, false);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;

    write(entrySize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show multiplications */
write("Multiplications, any $\\gamma$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, false);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + nSize, " vs. ", bitsPerEntry + mSize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show accuracies */
write("Accuracies, any $\\gamma$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSize(hmin, hmax, bitsPerEntry, wordSizes, false, false);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);

    write(entrySize + nSize, " vs. ", bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, splitFactor a power of 2, show size */
write("Size, $\\gamma= 2^t$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    tblSize = optRes.tableSizeBipartite.overallSize;

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show splitfactor */
write("Split factor $\\gamma$, $\\gamma= 2^t$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    splitFactor = optRes.bestSplitFactor;

    write(splitFactor);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show entry size exact table */
write("Entry size of exact table, $\\gamma= 2^t$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;

    write(entrySize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show multiplications */
write("Multiplications, $\\gamma= 2^t$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    match (optRes.multiplyExactWithM) with
       true    : { exSize = mSize; inexSize = nSize; }
       default : { exSize = nSize; inexSize = mSize; };

    write(entrySize + exSize, " vs. ", bitsPerEntry + inexSize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show accuracies */
write("Accuracies, $\\gamma= 2^t$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    match (optRes.multiplyExactWithM) with
       true    : { exSize = mSize; inexSize = nSize; }
       default : { exSize = nSize; inexSize = mSize; };

    write(entrySize + exSize, " vs. ", bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show which table to multiply with */
write("Multiply $\\theta_2$ with, $\\gamma= 2^t$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    match (optRes.multiplyExactWithM) with
       true    : { multWith = "$\\theta_2$ with $m$"; }
       default : { multWith = "$\\theta_2$ with $n$"; };

    write(multWith);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");

/* Bipartite table, any splitFactor, show size */
write("Size, any $\\gamma$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, false, mSize, nSize);
    tblSize = optRes.tableSizeBipartite.overallSize;

    write(formatSize(tblSize));
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show splitfactor */
write("Split factor $\\gamma$, any $\\gamma$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, false, mSize, nSize);
    splitFactor = optRes.bestSplitFactor;

    write(splitFactor);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show entry size exact table */
write("Entry size of exact table, any $\\gamma$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, false, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;

    write(entrySize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show multiplications */
write("Multiplications, any $\\gamma$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, false, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    match (optRes.multiplyExactWithM) with
       true    : { exSize = mSize; inexSize = nSize; }
       default : { exSize = nSize; inexSize = mSize; };

    write(entrySize + exSize, " vs. ", bitsPerEntry + inexSize);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, splitFactor a power of 2, show accuracies */
write("Accuracies, any $\\gamma$, $\\theta_2$ with $n$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, false, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    match (optRes.multiplyExactWithM) with 
       true    : { exSize = mSize; inexSize = nSize; }
       default : { exSize = nSize; inexSize = mSize; };

    write(entrySize + exSize, " vs. ", bitsPerEntry);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

/* Bipartite table, any splitFactor, show which table to multiply with */
write("Multiply $\\theta_2$ with, any $\\gamma$ & ");
for i from 0 to length(parameters) - 1 do {
    p = parameters[i];
    hmin = p.hmin;
    hmax = p.hmax;
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    optRes = optimizeSplitFactorForBipartiteTableSizeMinimizeBitDifference(hmin, hmax, bitsPerEntry, wordSizes, true, mSize, nSize);
    entrySize = optRes.tableSizeBipartite.bitsPerEntryR;
    mSize = chooseBitsPerEntry(p.p2, wordSizes);
    nSize = chooseBitsPerEntry(p.pp10, wordSizes);
    bitsPerEntry = chooseBitsPerEntry(p.wc, wordSizes);
    match (optRes.multiplyExactWithM) with
       true    : { multWith = "$\\theta_2$ with $m$"; }
       default : { multWith = "$\\theta_2$ with $n$"; };

    write(multWith);
    if (i != length(parameters) - 1) then write(" & ") else write(" \\\\\n");
};

write("\\noalign{\\medskip}\n");





quit;

