
#ifndef WIDE_INTEGERS_H
#define WIDE_INTEGERS_H

#include <stdint.h>

/* Helper defines for specifiers */
#define STATIC   static
#define INLINE   inline
#define CONST    const
#define RESTRICT __restrict__
#define VOLATILE volatile

/* Macros for the natural endianness of the system */

/* #define LITTLE_ENDIAN 1 */
/* #define BIG_ENDIAN 1 */

#if (!(defined(LITTLE_ENDIAN) || defined(BIG_ENDIAN)))
#define LITTLE_ENDIAN 1
#endif

#if defined(LITTLE_ENDIAN)
/* The code is written for big-endian (the natural way to think in a
   floating-point environment), so for little-endian systems return
   all array accesses. 
*/
#define __COMP_ENDIANNESS_IDX_NUMBER_ELEMS(x, n) (((n) - 1) - (x))
#else
/* Just take the indices as they are */
#define __COMP_ENDIANNESS_IDX_NUMBER_ELEMS(x, n) (x)
#endif

/* Wrapper around the endianness macro */
#define __COMP_ENDIANNESS_IDX_ARR_SIZE(x, a) __COMP_ENDIANNESS_IDX_NUMBER_ELEMS((x), ((sizeof((a)))/(sizeof(((a)[0])))))

/* Unsigned integer types 32 thru 512 bits (by 32 bit
   increments) 
*/

typedef uint32_t comp_uint32_t;
typedef uint64_t comp_uint64_t;
typedef struct {
  uint32_t w[3];
} comp_uint96_t;
typedef struct {
  uint64_t d[2];
} comp_uint128_t;
typedef struct {
  uint32_t w[5];
} comp_uint160_t;
typedef struct {
  uint64_t d[3];
} comp_uint192_t;
typedef struct {
  uint32_t w[7];
} comp_uint224_t;
typedef struct {
  uint64_t d[4];
} comp_uint256_t;
typedef struct {
  uint32_t w[9];
} comp_uint288_t;
typedef struct {
  uint64_t d[5];
} comp_uint320_t;
typedef struct {
  uint32_t w[11];
} comp_uint352_t;
typedef struct {
  uint64_t d[6];
} comp_uint384_t;
typedef struct {
  uint32_t w[13];
} comp_uint416_t;
typedef struct {
  uint64_t d[7];
} comp_uint448_t;
typedef struct {
  uint32_t w[15];
} comp_uint480_t;
typedef struct {
  uint64_t d[8];
} comp_uint512_t;

/* Macros to create literals for the different types 

   Types that are multiples of 64 bits must be given as chunks of 64
   bits, MSB in the first argument of the macro. The other types, which
   are multiple of 32 bits but not of 64 bits, must be given as chunks
   of 32 bits, MSB in the first argument of the macro.

   For example 

   COMP_UINT96_CONSTANT(1664, 42, 17) means 1664 * 2^(32 * 2) + 42 * 2^(32 * 1) + 17 * 2^(32 * 0)

*/

#if defined(LITTLE_ENDIAN)

#define COMP_UINT32_CONSTANT(w0) (UINT32_C(w0))
#define COMP_UINT64_CONSTANT(w0) (UINT64_C(w0))
#define COMP_UINT96_CONSTANT(w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2) } }
#define COMP_UINT128_CONSTANT(w1, w0) { { UINT64_C(w0), UINT64_C(w1) } }
#define COMP_UINT160_CONSTANT(w4, w3, w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2), UINT32_C(w3), UINT32_C(w4) } }
#define COMP_UINT192_CONSTANT(w2, w1, w0) { { UINT64_C(w0), UINT64_C(w1), UINT64_C(w2) } }
#define COMP_UINT224_CONSTANT(w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2), UINT32_C(w3), UINT32_C(w4), UINT32_C(w5), UINT32_C(w6) } }
#define COMP_UINT256_CONSTANT(w3, w2, w1, w0) { { UINT64_C(w0), UINT64_C(w1), UINT64_C(w2), UINT64_C(w3) } }
#define COMP_UINT288_CONSTANT(w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2), UINT32_C(w3), UINT32_C(w4), UINT32_C(w5), UINT32_C(w6), UINT32_C(w7), UINT32_C(w8) } }
#define COMP_UINT320_CONSTANT(w4, w3, w2, w1, w0) { { UINT64_C(w0), UINT64_C(w1), UINT64_C(w2), UINT64_C(w3), UINT64_C(w4) } }
#define COMP_UINT352_CONSTANT(w10, w9, w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2), UINT32_C(w3), UINT32_C(w4), UINT32_C(w5), UINT32_C(w6), UINT32_C(w7), UINT32_C(w8), UINT32_C(w9), UINT32_C(w10) } }
#define COMP_UINT384_CONSTANT(w5, w4, w3, w2, w1, w0) { { UINT64_C(w0), UINT64_C(w1), UINT64_C(w2), UINT64_C(w3), UINT64_C(w4), UINT64_C(w5) } }
#define COMP_UINT416_CONSTANT(w12, w11, w10, w9, w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2), UINT32_C(w3), UINT32_C(w4), UINT32_C(w5), UINT32_C(w6), UINT32_C(w7), UINT32_C(w8), UINT32_C(w9), UINT32_C(w10), UINT32_C(w11), UINT32_C(w12) } }
#define COMP_UINT448_CONSTANT(w6, w5, w4, w3, w2, w1, w0) { { UINT64_C(w0), UINT64_C(w1), UINT64_C(w2), UINT64_C(w3), UINT64_C(w4), UINT64_C(w5), UINT64_C(w6) } }
#define COMP_UINT480_CONSTANT(w14, w13, w12, w11, w10, w9, w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w0), UINT32_C(w1), UINT32_C(w2), UINT32_C(w3), UINT32_C(w4), UINT32_C(w5), UINT32_C(w6), UINT32_C(w7), UINT32_C(w8), UINT32_C(w9), UINT32_C(w10), UINT32_C(w11), UINT32_C(w12), UINT32_C(w13), UINT32_C(w14) } }
#define COMP_UINT512_CONSTANT(w7, w6, w5, w4, w3, w2, w1, w0) { { UINT64_C(w0), UINT64_C(w1), UINT64_C(w2), UINT64_C(w3), UINT64_C(w4), UINT64_C(w5), UINT64_C(w6), UINT64_C(w7) } }

#else

#define COMP_UINT32_CONSTANT(w0) (UINT32_C(w0))
#define COMP_UINT64_CONSTANT(w0) (UINT64_C(w0))
#define COMP_UINT96_CONSTANT(w2, w1, w0) { { UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT128_CONSTANT(w1, w0) { { UINT64_C(w1), UINT64_C(w0) } }
#define COMP_UINT160_CONSTANT(w4, w3, w2, w1, w0) { { UINT32_C(w4), UINT32_C(w3), UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT192_CONSTANT(w2, w1, w0) { { UINT64_C(w2), UINT64_C(w1), UINT64_C(w0) } }
#define COMP_UINT224_CONSTANT(w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w6), UINT32_C(w5), UINT32_C(w4), UINT32_C(w3), UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT256_CONSTANT(w3, w2, w1, w0) { { UINT64_C(w3), UINT64_C(w2), UINT64_C(w1), UINT64_C(w0) } }
#define COMP_UINT288_CONSTANT(w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w8), UINT32_C(w7), UINT32_C(w6), UINT32_C(w5), UINT32_C(w4), UINT32_C(w3), UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT320_CONSTANT(w4, w3, w2, w1, w0) { { UINT64_C(w4), UINT64_C(w3), UINT64_C(w2), UINT64_C(w1), UINT64_C(w0) } }
#define COMP_UINT352_CONSTANT(w10, w9, w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w10), UINT32_C(w9), UINT32_C(w8), UINT32_C(w7), UINT32_C(w6), UINT32_C(w5), UINT32_C(w4), UINT32_C(w3), UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT384_CONSTANT(w5, w4, w3, w2, w1, w0) { { UINT64_C(w5), UINT64_C(w4), UINT64_C(w3), UINT64_C(w2), UINT64_C(w1), UINT64_C(w0) } }
#define COMP_UINT416_CONSTANT(w12, w11, w10, w9, w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w12), UINT32_C(w11), UINT32_C(w10), UINT32_C(w9), UINT32_C(w8), UINT32_C(w7), UINT32_C(w6), UINT32_C(w5), UINT32_C(w4), UINT32_C(w3), UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT448_CONSTANT(w6, w5, w4, w3, w2, w1, w0) { { UINT64_C(w6), UINT64_C(w5), UINT64_C(w4), UINT64_C(w3), UINT64_C(w2), UINT64_C(w1), UINT64_C(w0) } }
#define COMP_UINT480_CONSTANT(w14, w13, w12, w11, w10, w9, w8, w7, w6, w5, w4, w3, w2, w1, w0) { { UINT32_C(w14), UINT32_C(w13), UINT32_C(w12), UINT32_C(w11), UINT32_C(w10), UINT32_C(w9), UINT32_C(w8), UINT32_C(w7), UINT32_C(w6), UINT32_C(w5), UINT32_C(w4), UINT32_C(w3), UINT32_C(w2), UINT32_C(w1), UINT32_C(w0) } }
#define COMP_UINT512_CONSTANT(w7, w6, w5, w4, w3, w2, w1, w0) { { UINT64_C(w7), UINT64_C(w6), UINT64_C(w5), UINT64_C(w4), UINT64_C(w3), UINT64_C(w2), UINT64_C(w1), UINT64_C(w0) } }

#endif



/* Helper functions to link the traditional uint32_t, uint64_t and __uint128_t with our comp_uint*_t types */

STATIC INLINE comp_uint32_t comp_uint32_to_comp_uint32(uint32_t a) {
  return a;
}

STATIC INLINE uint32_t comp_uint32_from_comp_uint32(comp_uint32_t a) {
  return a;
}

STATIC INLINE comp_uint64_t comp_uint64_to_comp_uint64(uint64_t a) {
  return a;
}

STATIC INLINE uint64_t comp_uint64_from_comp_uint64(comp_uint64_t a) {
  return a;
}

STATIC INLINE comp_uint128_t comp_uint128_to_comp_uint128(__uint128_t a) {
  __uint128_t tmp1;
  uint64_t tmp2, tmp3;
  comp_uint128_t tmp4;
  tmp1 = a;
  tmp2 = (uint64_t) tmp1;
  tmp1 >>= 64;
  tmp3 = (uint64_t) tmp1;
  tmp4.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0, tmp4.d)] = tmp3;
  tmp4.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1, tmp4.d)] = tmp2;
  return tmp4;
}

STATIC INLINE __uint128_t comp_uint128_from_comp_uint128(comp_uint128_t a) {
  uint64_t tmp1, tmp2;
  __uint128_t tmp3, tmp4;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0, a.d)];
  tmp2 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1, a.d)];
  tmp4 = tmp2;
  tmp3 = tmp1;
  tmp3 <<= 64;
  tmp3 |= tmp4;
  return tmp3;
}



/* High-order downconversions 
   
   Mathematically, comp_downconvert_high_<t1>_<t2>(b,a) computes

   b = floor(2^(t2 - t1) * a)

*/

STATIC INLINE void comp_downconvert_high_64_32(comp_uint32_t * RESTRICT b, comp_uint64_t a) {
  uint64_t tmp;
  tmp = a;
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_96_32(comp_uint32_t * RESTRICT b, comp_uint96_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_128_32(comp_uint32_t * RESTRICT b, comp_uint128_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_160_32(comp_uint32_t * RESTRICT b, comp_uint160_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_192_32(comp_uint32_t * RESTRICT b, comp_uint192_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_224_32(comp_uint32_t * RESTRICT b, comp_uint224_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_256_32(comp_uint32_t * RESTRICT b, comp_uint256_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_288_32(comp_uint32_t * RESTRICT b, comp_uint288_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_320_32(comp_uint32_t * RESTRICT b, comp_uint320_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_352_32(comp_uint32_t * RESTRICT b, comp_uint352_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_384_32(comp_uint32_t * RESTRICT b, comp_uint384_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_416_32(comp_uint32_t * RESTRICT b, comp_uint416_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_448_32(comp_uint32_t * RESTRICT b, comp_uint448_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_480_32(comp_uint32_t * RESTRICT b, comp_uint480_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
}

STATIC INLINE void comp_downconvert_high_512_32(comp_uint32_t * RESTRICT b, comp_uint512_t a) {
  uint64_t tmp;
  tmp = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp >>= 32;
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_96_64(comp_uint32_t * RESTRICT b, comp_uint96_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_128_64(comp_uint32_t * RESTRICT b, comp_uint128_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_160_64(comp_uint32_t * RESTRICT b, comp_uint160_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_192_64(comp_uint32_t * RESTRICT b, comp_uint192_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_224_64(comp_uint32_t * RESTRICT b, comp_uint224_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_256_64(comp_uint32_t * RESTRICT b, comp_uint256_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_288_64(comp_uint32_t * RESTRICT b, comp_uint288_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_320_64(comp_uint32_t * RESTRICT b, comp_uint320_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_352_64(comp_uint32_t * RESTRICT b, comp_uint352_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_384_64(comp_uint32_t * RESTRICT b, comp_uint384_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_416_64(comp_uint32_t * RESTRICT b, comp_uint416_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_448_64(comp_uint32_t * RESTRICT b, comp_uint448_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_480_64(comp_uint32_t * RESTRICT b, comp_uint480_t a) {
  uint64_t tmp;
  tmp = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp <<= 32;
  tmp |= a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  *b = tmp;
}

STATIC INLINE void comp_downconvert_high_512_64(comp_uint32_t * RESTRICT b, comp_uint512_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
}

STATIC INLINE void comp_downconvert_high_128_96(comp_uint96_t * RESTRICT b, comp_uint128_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_160_96(comp_uint96_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_160_128(comp_uint128_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_192_96(comp_uint96_t * RESTRICT b, comp_uint192_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_192_128(comp_uint128_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_192_160(comp_uint160_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_224_96(comp_uint96_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_224_128(comp_uint128_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_224_160(comp_uint160_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_224_192(comp_uint192_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_256_96(comp_uint96_t * RESTRICT b, comp_uint256_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_256_128(comp_uint128_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_256_160(comp_uint160_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_256_192(comp_uint192_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_256_224(comp_uint224_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_288_96(comp_uint96_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_288_128(comp_uint128_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_288_160(comp_uint160_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_288_192(comp_uint192_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_288_224(comp_uint224_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_288_256(comp_uint256_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_320_96(comp_uint96_t * RESTRICT b, comp_uint320_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_320_128(comp_uint128_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_320_160(comp_uint160_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_320_192(comp_uint192_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_320_224(comp_uint224_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_320_256(comp_uint256_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_320_288(comp_uint288_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_352_96(comp_uint96_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_352_128(comp_uint128_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_352_160(comp_uint160_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_352_192(comp_uint192_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_352_224(comp_uint224_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_352_256(comp_uint256_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_352_288(comp_uint288_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_352_320(comp_uint320_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_384_96(comp_uint96_t * RESTRICT b, comp_uint384_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_384_128(comp_uint128_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_384_160(comp_uint160_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_384_192(comp_uint192_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_384_224(comp_uint224_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_384_256(comp_uint256_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_384_288(comp_uint288_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_384_320(comp_uint320_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_384_352(comp_uint352_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_416_96(comp_uint96_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_416_128(comp_uint128_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_416_160(comp_uint160_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_416_192(comp_uint192_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_416_224(comp_uint224_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_416_256(comp_uint256_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_416_288(comp_uint288_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_416_320(comp_uint320_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_416_352(comp_uint352_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_416_384(comp_uint384_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_448_96(comp_uint96_t * RESTRICT b, comp_uint448_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_448_128(comp_uint128_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_448_160(comp_uint160_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_448_192(comp_uint192_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_448_224(comp_uint224_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_448_256(comp_uint256_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_448_288(comp_uint288_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_448_320(comp_uint320_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_448_352(comp_uint352_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_448_384(comp_uint384_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_448_416(comp_uint416_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_480_96(comp_uint96_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_480_128(comp_uint128_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_480_160(comp_uint160_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_480_192(comp_uint192_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_480_224(comp_uint224_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_480_256(comp_uint256_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_480_288(comp_uint288_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_480_320(comp_uint320_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_480_352(comp_uint352_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_480_384(comp_uint384_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_480_416(comp_uint416_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_downconvert_high_480_448(comp_uint448_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_high_512_96(comp_uint96_t * RESTRICT b, comp_uint512_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_512_128(comp_uint128_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_512_160(comp_uint160_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_512_192(comp_uint192_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_512_224(comp_uint224_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_512_256(comp_uint256_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_512_288(comp_uint288_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_512_320(comp_uint320_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_512_352(comp_uint352_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_512_384(comp_uint384_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_512_416(comp_uint416_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_high_512_448(comp_uint448_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_downconvert_high_512_480(comp_uint480_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,a.d)];
  tmp1 >>= 32;
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, b->w)] = tmp2;
}

/* Low-order downconversions 
   
   Mathematically, comp_downconvert_low_<t1>_<t2>(b,a) computes

   b = a mod 2^t2

*/
STATIC INLINE void comp_downconvert_low_64_32(comp_uint32_t * RESTRICT b, comp_uint64_t a) {
  *b = (uint32_t) a;
}

STATIC INLINE void comp_downconvert_low_96_32(comp_uint32_t * RESTRICT b, comp_uint96_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
}

STATIC INLINE void comp_downconvert_low_128_32(comp_uint32_t * RESTRICT b, comp_uint128_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_160_32(comp_uint32_t * RESTRICT b, comp_uint160_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
}

STATIC INLINE void comp_downconvert_low_192_32(comp_uint32_t * RESTRICT b, comp_uint192_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_224_32(comp_uint32_t * RESTRICT b, comp_uint224_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
}

STATIC INLINE void comp_downconvert_low_256_32(comp_uint32_t * RESTRICT b, comp_uint256_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_288_32(comp_uint32_t * RESTRICT b, comp_uint288_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,a.w)];
}

STATIC INLINE void comp_downconvert_low_320_32(comp_uint32_t * RESTRICT b, comp_uint320_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_352_32(comp_uint32_t * RESTRICT b, comp_uint352_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, a.w)];
}

STATIC INLINE void comp_downconvert_low_384_32(comp_uint32_t * RESTRICT b, comp_uint384_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_416_32(comp_uint32_t * RESTRICT b, comp_uint416_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, a.w)];
}

STATIC INLINE void comp_downconvert_low_448_32(comp_uint32_t * RESTRICT b, comp_uint448_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_480_32(comp_uint32_t * RESTRICT b, comp_uint480_t a) {
  *b = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, a.w)];
}

STATIC INLINE void comp_downconvert_low_512_32(comp_uint32_t * RESTRICT b, comp_uint512_t a) {
  uint32_t tmp;
  tmp = (uint32_t) (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,a.d)]);
  *b = tmp;
}

STATIC INLINE void comp_downconvert_low_96_64(comp_uint64_t * RESTRICT b, comp_uint96_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_128_64(comp_uint64_t * RESTRICT b, comp_uint128_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
}

STATIC INLINE void comp_downconvert_low_160_64(comp_uint64_t * RESTRICT b, comp_uint160_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_192_64(comp_uint64_t * RESTRICT b, comp_uint192_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
}

STATIC INLINE void comp_downconvert_low_224_64(comp_uint64_t * RESTRICT b, comp_uint224_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_256_64(comp_uint64_t * RESTRICT b, comp_uint256_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
}

STATIC INLINE void comp_downconvert_low_288_64(comp_uint64_t * RESTRICT b, comp_uint288_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_320_64(comp_uint64_t * RESTRICT b, comp_uint320_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
}

STATIC INLINE void comp_downconvert_low_352_64(comp_uint64_t * RESTRICT b, comp_uint352_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(9,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_384_64(comp_uint64_t * RESTRICT b, comp_uint384_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
}

STATIC INLINE void comp_downconvert_low_416_64(comp_uint64_t * RESTRICT b, comp_uint416_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(11, a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_448_64(comp_uint64_t * RESTRICT b, comp_uint448_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.d)];
}

STATIC INLINE void comp_downconvert_low_480_64(comp_uint64_t * RESTRICT b, comp_uint480_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(13, a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  *b = tmp3;
}

STATIC INLINE void comp_downconvert_low_512_64(comp_uint64_t * RESTRICT b, comp_uint512_t a) {
  *b = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,a.d)];
}

STATIC INLINE void comp_downconvert_low_128_96(comp_uint96_t * RESTRICT b, comp_uint128_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_160_96(comp_uint96_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_160_128(comp_uint128_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 2, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_192_96(comp_uint96_t * RESTRICT b, comp_uint192_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_192_128(comp_uint128_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 1, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_192_160(comp_uint160_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=1;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_224_96(comp_uint96_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_224_128(comp_uint128_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 3, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 4, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_224_160(comp_uint160_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_224_192(comp_uint192_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 2, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_256_96(comp_uint96_t * RESTRICT b, comp_uint256_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_256_128(comp_uint128_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_256_160(comp_uint160_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=2;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_256_192(comp_uint192_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 1, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_256_224(comp_uint224_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=1;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_288_96(comp_uint96_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 6, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_288_128(comp_uint128_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 5, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 6, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_288_160(comp_uint160_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_288_192(comp_uint192_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 3, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 4, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_288_224(comp_uint224_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_288_256(comp_uint256_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 2, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_320_96(comp_uint96_t * RESTRICT b, comp_uint320_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_320_128(comp_uint128_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 3, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_320_160(comp_uint160_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=3;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_320_192(comp_uint192_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_320_224(comp_uint224_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=2;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_320_256(comp_uint256_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 1, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_320_288(comp_uint288_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=1;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_352_96(comp_uint96_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 8, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_352_128(comp_uint128_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 7, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 8, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_352_160(comp_uint160_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 6, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_352_192(comp_uint192_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 5, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 6, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_352_224(comp_uint224_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_352_256(comp_uint256_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 3, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 4, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_352_288(comp_uint288_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_352_320(comp_uint320_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 2, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_384_96(comp_uint96_t * RESTRICT b, comp_uint384_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_384_128(comp_uint128_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_384_160(comp_uint160_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=4;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-7), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-6), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_384_192(comp_uint192_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 3, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_384_224(comp_uint224_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=3;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_384_256(comp_uint256_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_384_288(comp_uint288_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=2;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_384_320(comp_uint320_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 1, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_384_352(comp_uint352_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=1;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_416_96(comp_uint96_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 10, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_416_128(comp_uint128_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 9, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 10, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_416_160(comp_uint160_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 8, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_416_192(comp_uint192_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 7, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 8, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_416_224(comp_uint224_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 6, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_416_256(comp_uint256_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 5, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 6, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_416_288(comp_uint288_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_416_320(comp_uint320_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 3, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 4, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_416_352(comp_uint352_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_416_384(comp_uint384_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 2, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_448_96(comp_uint96_t * RESTRICT b, comp_uint448_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_448_128(comp_uint128_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 5, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_448_160(comp_uint160_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=5;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-9), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-8), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_448_192(comp_uint192_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_448_224(comp_uint224_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=4;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-7), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-6), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_448_256(comp_uint256_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 3, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_448_288(comp_uint288_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=3;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_448_320(comp_uint320_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_448_352(comp_uint352_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=2;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_448_384(comp_uint384_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 1, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_448_416(comp_uint416_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=1;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_480_96(comp_uint96_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 12, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_480_128(comp_uint128_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 11, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 12, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_480_160(comp_uint160_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 10, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_480_192(comp_uint192_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 9, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 10, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_480_224(comp_uint224_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 8, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_480_256(comp_uint256_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 7, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 8, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_480_288(comp_uint288_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 6, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_480_320(comp_uint320_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 5, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 6, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_480_352(comp_uint352_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_480_384(comp_uint384_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 3, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 4, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_480_416(comp_uint416_t * RESTRICT b, comp_uint480_t a) {
  int i;
  for (i=0;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.w)];
  }
}

STATIC INLINE void comp_downconvert_low_480_448(comp_uint448_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 2, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_downconvert_low_512_96(comp_uint96_t * RESTRICT b, comp_uint512_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,a.d)];
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_downconvert_low_512_128(comp_uint128_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 6, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_512_160(comp_uint160_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=6;i<8;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-11), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-10), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_512_192(comp_uint192_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 5, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_512_224(comp_uint224_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=5;i<8;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-9), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-8), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_512_256(comp_uint256_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 4, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_512_288(comp_uint288_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=4;i<8;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-7), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-6), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_512_320(comp_uint320_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 3, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_512_352(comp_uint352_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=3;i<8;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_512_384(comp_uint384_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 2, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_512_416(comp_uint416_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=2;i<8;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), b->w)] = tmp2;
  }
}

STATIC INLINE void comp_downconvert_low_512_448(comp_uint448_t * RESTRICT b, comp_uint512_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + 1, a.d)];
  }
}

STATIC INLINE void comp_downconvert_low_512_480(comp_uint480_t * RESTRICT b, comp_uint512_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)];
  tmp2 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp2;
  for (i=1;i<8;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), b->w)] = tmp2;
  }
}




/* High-order upconversions 
   
   Mathematically, comp_upconvert_high_<t1>_<t2>(b,a) computes

   b = 2^(t2 - t1) * a

*/

STATIC INLINE void comp_upconvert_high_32_64(comp_uint64_t * RESTRICT b, comp_uint32_t a) {
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  *b = tmp;
}

STATIC INLINE void comp_upconvert_high_32_96(comp_uint96_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_128(comp_uint128_t * RESTRICT b, comp_uint32_t a) {
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_32_160(comp_uint160_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_192(comp_uint192_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  for (i=1;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_224(comp_uint224_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_256(comp_uint256_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  for (i=1;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_288(comp_uint288_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_320(comp_uint320_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  for (i=1;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_352(comp_uint352_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_384(comp_uint384_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  for (i=1;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_416(comp_uint416_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_448(comp_uint448_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  for (i=1;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_480(comp_uint480_t * RESTRICT b, comp_uint32_t a) {
  int i;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = a;
  for (i=1;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_32_512(comp_uint512_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  tmp = a;
  tmp <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp;
  for (i=1;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_96(comp_uint96_t * RESTRICT b, comp_uint64_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_64_128(comp_uint128_t * RESTRICT b, comp_uint64_t a) {
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_64_160(comp_uint160_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  for (i=2;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_192(comp_uint192_t * RESTRICT b, comp_uint64_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  for (i=1;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_224(comp_uint224_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  for (i=2;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_256(comp_uint256_t * RESTRICT b, comp_uint64_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  for (i=1;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_288(comp_uint288_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  for (i=2;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_320(comp_uint320_t * RESTRICT b, comp_uint64_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  for (i=1;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_352(comp_uint352_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  for (i=2;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_384(comp_uint384_t * RESTRICT b, comp_uint64_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  for (i=1;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_416(comp_uint416_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  for (i=2;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_448(comp_uint448_t * RESTRICT b, comp_uint64_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  for (i=1;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_480(comp_uint480_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp2;
  for (i=2;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_64_512(comp_uint512_t * RESTRICT b, comp_uint64_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = a;
  for (i=1;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_128(comp_uint128_t * RESTRICT b, comp_uint96_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_high_96_160(comp_uint160_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=3;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_192(comp_uint192_t * RESTRICT b, comp_uint96_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_96_224(comp_uint224_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=3;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_256(comp_uint256_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for (i=2;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_288(comp_uint288_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=3;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_320(comp_uint320_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for (i=2;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_352(comp_uint352_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=3;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_384(comp_uint384_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for (i=2;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_416(comp_uint416_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=3;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_448(comp_uint448_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for (i=2;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_480(comp_uint480_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=3;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_96_512(comp_uint512_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for (i=2;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_160(comp_uint160_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_128_192(comp_uint192_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_128_224(comp_uint224_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=4;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_256(comp_uint256_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=2;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_288(comp_uint288_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=4;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_320(comp_uint320_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=2;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_352(comp_uint352_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=4;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_384(comp_uint384_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=2;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_416(comp_uint416_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=4;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_448(comp_uint448_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=2;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_480(comp_uint480_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=4;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_128_512(comp_uint512_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=2;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_192(comp_uint192_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_high_160_224(comp_uint224_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=5;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_256(comp_uint256_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_160_288(comp_uint288_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=5;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_320(comp_uint320_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for (i=3;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_352(comp_uint352_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=5;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_384(comp_uint384_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for (i=3;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_416(comp_uint416_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=5;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_448(comp_uint448_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for (i=3;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_480(comp_uint480_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=5;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_160_512(comp_uint512_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<2;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for (i=3;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_224(comp_uint224_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_192_256(comp_uint256_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_192_288(comp_uint288_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=6;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_320(comp_uint320_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=3;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_352(comp_uint352_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=6;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_384(comp_uint384_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=3;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_416(comp_uint416_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=6;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_448(comp_uint448_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=3;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_480(comp_uint480_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=6;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_192_512(comp_uint512_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=3;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_256(comp_uint256_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_high_224_288(comp_uint288_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=7;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_320(comp_uint320_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_224_352(comp_uint352_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=7;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_384(comp_uint384_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  for (i=4;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_416(comp_uint416_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=7;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_448(comp_uint448_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  for (i=4;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_480(comp_uint480_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=7;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_224_512(comp_uint512_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  for (i=4;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_256_288(comp_uint288_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_256_320(comp_uint320_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_256_352(comp_uint352_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=8;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_256_384(comp_uint384_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=4;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_256_416(comp_uint416_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=8;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_256_448(comp_uint448_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=4;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_256_480(comp_uint480_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=8;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_256_512(comp_uint512_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=4;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_288_320(comp_uint320_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_high_288_352(comp_uint352_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=9;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_288_384(comp_uint384_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_288_416(comp_uint416_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=9;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_288_448(comp_uint448_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
  for (i=5;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_288_480(comp_uint480_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=9;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_288_512(comp_uint512_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
  for (i=5;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_320_352(comp_uint352_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_320_384(comp_uint384_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_320_416(comp_uint416_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=10;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_320_448(comp_uint448_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=5;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_320_480(comp_uint480_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=10;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_320_512(comp_uint512_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=5;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_352_384(comp_uint384_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_high_352_416(comp_uint416_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=11;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_352_448(comp_uint448_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_352_480(comp_uint480_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=11;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_352_512(comp_uint512_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp3;
  for (i=6;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_384_416(comp_uint416_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_384_448(comp_uint448_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_384_480(comp_uint480_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  for (i=12;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_384_512(comp_uint512_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  for (i=6;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_416_448(comp_uint448_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_high_416_480(comp_uint480_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for (i=0;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
  for (i=13;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
}

STATIC INLINE void comp_upconvert_high_416_512(comp_uint512_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_448_480(comp_uint480_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, b->w)] = tmp2;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, b->w)] = (uint32_t) 0;
}

STATIC INLINE void comp_upconvert_high_448_512(comp_uint512_t * RESTRICT b, comp_uint448_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->d)] = (uint64_t) 0;
}

STATIC INLINE void comp_upconvert_high_480_512(comp_uint512_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i, a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + 1, a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->d)] = tmp3;
}


/* Low-order upconversions 
   
   Mathematically, comp_upconvert_low_<t1>_<t2>(b,a) computes

   b = a

*/

STATIC INLINE void comp_upconvert_low_32_64(comp_uint64_t * RESTRICT b, comp_uint32_t a) {
  *b = a;
}

STATIC INLINE void comp_upconvert_low_32_96(comp_uint96_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_128(comp_uint128_t * RESTRICT b, comp_uint32_t a) {
  uint64_t tmp;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_32_160(comp_uint160_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_192(comp_uint192_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_32_224(comp_uint224_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_256(comp_uint256_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_32_288(comp_uint288_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<8;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_320(comp_uint320_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_32_352(comp_uint352_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<10;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_384(comp_uint384_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_32_416(comp_uint416_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<12;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_448(comp_uint448_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_32_480(comp_uint480_t * RESTRICT b, comp_uint32_t a) {
  int i;
  for (i=0;i<14;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, b->w)] = a;
}

STATIC INLINE void comp_upconvert_low_32_512(comp_uint512_t * RESTRICT b, comp_uint32_t a) {
  int i;
  uint64_t tmp;
  for (i=0;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp = a;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->d)] = tmp;
}

STATIC INLINE void comp_upconvert_low_64_96(comp_uint96_t * RESTRICT b, comp_uint64_t a) {
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_128(comp_uint128_t * RESTRICT b, comp_uint64_t a) {
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_64_160(comp_uint160_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_192(comp_uint192_t * RESTRICT b, comp_uint64_t a) {
  int i;
  for (i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_64_224(comp_uint224_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_256(comp_uint256_t * RESTRICT b, comp_uint64_t a) {
  int i;
  for (i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_64_288(comp_uint288_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8,b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_320(comp_uint320_t * RESTRICT b, comp_uint64_t a) {
  int i;
  for (i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_64_352(comp_uint352_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(9,b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10, b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_384(comp_uint384_t * RESTRICT b, comp_uint64_t a) {
  int i;
  for (i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_64_416(comp_uint416_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(11, b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12, b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_448(comp_uint448_t * RESTRICT b, comp_uint64_t a) {
  int i;
  for (i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_64_480(comp_uint480_t * RESTRICT b, comp_uint64_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  tmp1 = a;
  tmp2 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp3 = (uint32_t) tmp1;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(13, b->w)] = tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(14, b->w)] = tmp2;
}

STATIC INLINE void comp_upconvert_low_64_512(comp_uint512_t * RESTRICT b, comp_uint64_t a) {
  int i;
  for (i=0;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->d)] = a;
}

STATIC INLINE void comp_upconvert_low_96_128(comp_uint128_t * RESTRICT b, comp_uint96_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_96_160(comp_uint160_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_96_192(comp_uint192_t * RESTRICT b, comp_uint96_t a) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_96_224(comp_uint224_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_96_256(comp_uint256_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_96_288(comp_uint288_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for(i=0;i<6;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_96_320(comp_uint320_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_96_352(comp_uint352_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for(i=0;i<8;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_96_384(comp_uint384_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_96_416(comp_uint416_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for(i=0;i<10;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_96_448(comp_uint448_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_96_480(comp_uint480_t * RESTRICT b, comp_uint96_t a) {
  int i;
  for(i=0;i<12;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_96_512(comp_uint512_t * RESTRICT b, comp_uint96_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6,b->d)] = tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,a.w)];
  tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,a.w)];
  tmp3 = tmp1;
  tmp3 <<= 32;
  tmp3 |= tmp2;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(7,b->d)] = tmp3;
}

STATIC INLINE void comp_upconvert_low_128_160(comp_uint160_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  for(i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_128_192(comp_uint192_t * RESTRICT b, comp_uint128_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_128_224(comp_uint224_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_128_256(comp_uint256_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_128_288(comp_uint288_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(5 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_128_320(comp_uint320_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_128_352(comp_uint352_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(7 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_128_384(comp_uint384_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_128_416(comp_uint416_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(9 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_128_448(comp_uint448_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_128_480(comp_uint480_t * RESTRICT b, comp_uint128_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<2;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(11 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(12 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_128_512(comp_uint512_t * RESTRICT b, comp_uint128_t a) {
  int i;
  for(i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_160_192(comp_uint192_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  for(i=1;i<3;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_160_224(comp_uint224_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_160_256(comp_uint256_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for(i=2;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_160_288(comp_uint288_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_160_320(comp_uint320_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for(i=3;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_160_352(comp_uint352_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for(i=0;i<6;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_160_384(comp_uint384_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  for(i=4;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-7), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-6), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_160_416(comp_uint416_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for(i=0;i<8;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_160_448(comp_uint448_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
  for(i=5;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-9), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-8), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_160_480(comp_uint480_t * RESTRICT b, comp_uint160_t a) {
  int i;
  for(i=0;i<10;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_160_512(comp_uint512_t * RESTRICT b, comp_uint160_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5,b->d)] = tmp3;
  for(i=6;i<8;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-11), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-10), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_192_224(comp_uint224_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  for(i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_192_256(comp_uint256_t * RESTRICT b, comp_uint192_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_192_288(comp_uint288_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_192_320(comp_uint320_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_192_352(comp_uint352_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(5 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_192_384(comp_uint384_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_192_416(comp_uint416_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(7 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_192_448(comp_uint448_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_192_480(comp_uint480_t * RESTRICT b, comp_uint192_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<3;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(9 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(10 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_192_512(comp_uint512_t * RESTRICT b, comp_uint192_t a) {
  int i;
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(5 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_224_256(comp_uint256_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  for(i=1;i<4;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_224_288(comp_uint288_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_224_320(comp_uint320_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for(i=2;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_224_352(comp_uint352_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_224_384(comp_uint384_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for(i=3;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_224_416(comp_uint416_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for(i=0;i<6;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_224_448(comp_uint448_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  for(i=4;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-7), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-6), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_224_480(comp_uint480_t * RESTRICT b, comp_uint224_t a) {
  int i;
  for(i=0;i<8;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_224_512(comp_uint512_t * RESTRICT b, comp_uint224_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4,b->d)] = tmp3;
  for(i=5;i<8;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-9), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-8), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_256_288(comp_uint288_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  for(i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_256_320(comp_uint320_t * RESTRICT b, comp_uint256_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_256_352(comp_uint352_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_256_384(comp_uint384_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_256_416(comp_uint416_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(5 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_256_448(comp_uint448_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_256_480(comp_uint480_t * RESTRICT b, comp_uint256_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<4;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(7 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(8 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_256_512(comp_uint512_t * RESTRICT b, comp_uint256_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_288_320(comp_uint320_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  for(i=1;i<5;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_288_352(comp_uint352_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_288_384(comp_uint384_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for(i=2;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_288_416(comp_uint416_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_288_448(comp_uint448_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for(i=3;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_288_480(comp_uint480_t * RESTRICT b, comp_uint288_t a) {
  int i;
  for(i=0;i<6;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_288_512(comp_uint512_t * RESTRICT b, comp_uint288_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3,b->d)] = tmp3;
  for(i=4;i<8;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-7), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-6), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_320_352(comp_uint352_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  for(i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_320_384(comp_uint384_t * RESTRICT b, comp_uint320_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_320_416(comp_uint416_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_320_448(comp_uint448_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_320_480(comp_uint480_t * RESTRICT b, comp_uint320_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<5;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(5 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(6 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_320_512(comp_uint512_t * RESTRICT b, comp_uint320_t a) {
  int i;
  for(i=0;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_352_384(comp_uint384_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  for(i=1;i<6;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_352_416(comp_uint416_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_352_448(comp_uint448_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for(i=2;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_352_480(comp_uint480_t * RESTRICT b, comp_uint352_t a) {
  int i;
  for(i=0;i<4;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_352_512(comp_uint512_t * RESTRICT b, comp_uint352_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b->d)] = tmp3;
  for(i=3;i<8;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-5), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-4), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_384_416(comp_uint416_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  for(i=0;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_384_448(comp_uint448_t * RESTRICT b, comp_uint384_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  for(i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_384_480(comp_uint480_t * RESTRICT b, comp_uint384_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for(i=0;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<6;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(3 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(4 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_384_512(comp_uint512_t * RESTRICT b, comp_uint384_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  for(i=0;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_416_448(comp_uint448_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  for(i=1;i<7;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_416_480(comp_uint480_t * RESTRICT b, comp_uint416_t a) {
  int i;
  for(i=0;i<2;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  for(i=0;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + i, b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)];
  }
}

STATIC INLINE void comp_upconvert_low_416_512(comp_uint512_t * RESTRICT b, comp_uint416_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b->d)] = tmp3;
  for(i=2;i<8;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-3), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-2), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}

STATIC INLINE void comp_upconvert_low_448_480(comp_uint480_t * RESTRICT b, comp_uint448_t a) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->w)] = (uint32_t) 0;
  for(i=0;i<7;i++) {
    tmp1 = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + 2 * i, b->w)] = tmp3;
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 + 2 * i, b->w)] = tmp2;
  }
}

STATIC INLINE void comp_upconvert_low_448_512(comp_uint512_t * RESTRICT b, comp_uint448_t a) {
  int i;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = (uint64_t) 0;
  for(i=0;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1 + i, b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)];
  }
}

STATIC INLINE void comp_upconvert_low_480_512(comp_uint512_t * RESTRICT b, comp_uint480_t a) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)];
  tmp3 = tmp1;
  b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b->d)] = tmp3;
  for(i=1;i<8;i++) {
    tmp1 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (-1), a.w)];
    tmp2 = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2 * i + (0), a.w)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = tmp3;
  }
}


/* Right-shifts
   
   Mathematically, comp_shift_right_<t>(b,a,k) computes

   b = floor(2^(-(k mod t)) * a)

*/
STATIC INLINE void comp_shift_right_32(comp_uint32_t * RESTRICT b, comp_uint32_t a, unsigned int k) {
  unsigned int s;
  s = k & ((unsigned int) 31);
  *b = a >> s;
}

STATIC INLINE void comp_shift_right_64(comp_uint64_t * RESTRICT b, comp_uint64_t a, unsigned int k) {
  unsigned int s;
  s = k & ((unsigned int) 63);
  *b = a >> s;
}

STATIC INLINE void comp_shift_right_96(comp_uint96_t * RESTRICT b, comp_uint96_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 96);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=2;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<3;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_128(comp_uint128_t * RESTRICT b, comp_uint128_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k & ((unsigned int) 127);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=1;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<2;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_160(comp_uint160_t * RESTRICT b, comp_uint160_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 160);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=4;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<5;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_192(comp_uint192_t * RESTRICT b, comp_uint192_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 192);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=2;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<3;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_224(comp_uint224_t * RESTRICT b, comp_uint224_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 224);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=6;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<7;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_256(comp_uint256_t * RESTRICT b, comp_uint256_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k & ((unsigned int) 255);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=3;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<4;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_288(comp_uint288_t * RESTRICT b, comp_uint288_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 288);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=8;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<9;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_320(comp_uint320_t * RESTRICT b, comp_uint320_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 320);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=4;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<5;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_352(comp_uint352_t * RESTRICT b, comp_uint352_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 352);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=10;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<11;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_384(comp_uint384_t * RESTRICT b, comp_uint384_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 384);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=5;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<6;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_416(comp_uint416_t * RESTRICT b, comp_uint416_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 416);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=12;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<13;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_448(comp_uint448_t * RESTRICT b, comp_uint448_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 448);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=6;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<7;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_480(comp_uint480_t * RESTRICT b, comp_uint480_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 480);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=14;i>=w;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.w)];
  }
  for (i=w-1;i>=0;i--) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint32_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint32_t) 1) << r) - ((uint32_t) 1);
    v = ((unsigned int) 32) - r;
    c = (uint32_t) 0;
    for (i=0;i<15;i++) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] >> r) | c;
      c = t << v;
    }
  }
}

STATIC INLINE void comp_shift_right_512(comp_uint512_t * RESTRICT b, comp_uint512_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k & ((unsigned int) 511);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=7;i>=w;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i - w, a.d)];
  }
  for (i=w-1;i>=0;i--) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    m = (((uint64_t) 1) << r) - ((uint64_t) 1);
    v = ((unsigned int) 64) - r;
    c = (uint64_t) 0;
    for (i=0;i<8;i++) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] >> r) | c;
      c = t << v;
    }
  }
}


/* Left-shifts
   
   Mathematically, comp_shift_left_<t>(b,a,k) computes

   b = (2^(k mod t) * a) mod 2^t 

*/
STATIC INLINE void comp_shift_left_32(comp_uint32_t * RESTRICT b, comp_uint32_t a, unsigned int k) {
  unsigned int s;
  s = k & ((unsigned int) 31);
  *b = a << s;
}

STATIC INLINE void comp_shift_left_64(comp_uint64_t * RESTRICT b, comp_uint64_t a, unsigned int k) {
  unsigned int s;
  s = k & ((unsigned int) 63);
  *b = a << s;
}

STATIC INLINE void comp_shift_left_96(comp_uint96_t * RESTRICT b, comp_uint96_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 96);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<3-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=3-w;i<3;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=2;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_128(comp_uint128_t * RESTRICT b, comp_uint128_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k & ((unsigned int) 127);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<2-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=2-w;i<2;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=1;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_160(comp_uint160_t * RESTRICT b, comp_uint160_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 160);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<5-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=5-w;i<5;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=4;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_192(comp_uint192_t * RESTRICT b, comp_uint192_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 192);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<3-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=3-w;i<3;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=2;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_224(comp_uint224_t * RESTRICT b, comp_uint224_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 224);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<7-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=7-w;i<7;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=6;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_256(comp_uint256_t * RESTRICT b, comp_uint256_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k & ((unsigned int) 255);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<4-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=4-w;i<4;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=3;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_288(comp_uint288_t * RESTRICT b, comp_uint288_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 288);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<9-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=9-w;i<9;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=8;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_320(comp_uint320_t * RESTRICT b, comp_uint320_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 320);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<5-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=5-w;i<5;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=4;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_352(comp_uint352_t * RESTRICT b, comp_uint352_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 352);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<11-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=11-w;i<11;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=10;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_384(comp_uint384_t * RESTRICT b, comp_uint384_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 384);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<6-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=6-w;i<6;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=5;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_416(comp_uint416_t * RESTRICT b, comp_uint416_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 416);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<13-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=13-w;i<13;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=12;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_448(comp_uint448_t * RESTRICT b, comp_uint448_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k % ((unsigned int) 448);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<7-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=7-w;i<7;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=6;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_480(comp_uint480_t * RESTRICT b, comp_uint480_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint32_t c, t, m;
  s = k % ((unsigned int) 480);
  w = (int) (s >> 5);
  r = s & ((unsigned int) 31);
  for (i=0;i<15-w;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.w)];
  }
  for (i=15-w;i<15;i++) {
    b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 32) - r;
    m = ((((uint32_t) 1) << r) - ((uint32_t) 1)) << v;
    c = (uint32_t) 0;
    for (i=14;i>=0;i--) {
      t = b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] & m;
      b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] = (b->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->w)] << r) | c;
      c = t >> v;
    }
  }
}

STATIC INLINE void comp_shift_left_512(comp_uint512_t * RESTRICT b, comp_uint512_t a, unsigned int k) {
  unsigned int s, r, v;
  int i, w;
  uint64_t c, t, m;
  s = k & ((unsigned int) 511);
  w = (int) (s >> 6);
  r = s & ((unsigned int) 63);
  for (i=0;i<8-w;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i + w, a.d)];
  }
  for (i=8-w;i<8;i++) {
    b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (uint64_t) 0;
  }
  if (r != (unsigned int) 0) {
    v = ((unsigned int) 64) - r;
    m = ((((uint64_t) 1) << r) - ((uint64_t) 1)) << v;
    c = (uint64_t) 0;
    for (i=7;i>=0;i--) {
      t = b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] & m;
      b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] = (b->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b->d)] << r) | c;
      c = t >> v;
    }
  }
}


/* Leading-zero counter functions 

   If the input is zero, the functions return the input bitsize of the input.

*/

STATIC INLINE unsigned int comp_leading_zeros_32(comp_uint32_t a) {
  unsigned int r;
  uint32_t t;
  if (a == ((uint32_t) 0)) return (unsigned int) 32;
  if (sizeof(uint32_t) == sizeof(unsigned int)) {
    return (unsigned int) __builtin_clz(a);
  } else {
    if (sizeof(uint32_t) == sizeof(unsigned long)) {
      return (unsigned int) __builtin_clzl(a);
    } else {
      if (sizeof(uint32_t) == sizeof(unsigned long long)) {
	return (unsigned int) __builtin_clzll(a);
      } else {
	for (r=0, t=a; ((t & (((uint32_t) 1) << 31)) == ((uint32_t) 0)); r++, t<<=1);
	return r;
      }
    }
  }
  return (unsigned int) 0;
}

STATIC INLINE unsigned int __comp_leading_zeros_uint32(uint32_t a) {
  return comp_leading_zeros_32(comp_uint32_to_comp_uint32(a));
}

STATIC INLINE unsigned int comp_leading_zeros_64(comp_uint64_t a) {
  unsigned int r;
  uint64_t t;
  if (a == ((uint64_t) 0)) return (unsigned int) 64;
  if (sizeof(uint64_t) == sizeof(unsigned int)) {
    return (unsigned int) __builtin_clz(a);
  } else {
    if (sizeof(uint64_t) == sizeof(unsigned long)) {
      return (unsigned int) __builtin_clzl(a);
    } else {
      if (sizeof(uint64_t) == sizeof(unsigned long long)) {
	return (unsigned int) __builtin_clzll(a);
      } else {
	for (r=0, t=a; ((t & (((uint64_t) 1) << 63)) == ((uint32_t) 0)); r++, t<<=1);
	return r;
      }
    }
  }
  return (unsigned int) 0;
}

STATIC INLINE unsigned int __comp_leading_zeros_uint64(uint64_t a) {
  return comp_leading_zeros_64(comp_uint64_to_comp_uint64(a));
}

STATIC INLINE unsigned int comp_leading_zeros_96(comp_uint96_t a) {
  int i;
  for (i=0;((i<2) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_128(comp_uint128_t a) {
  int i;
  for (i=0;((i<1) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}

STATIC INLINE unsigned int __comp_leading_zeros_uint128(__uint128_t a) {
  return comp_leading_zeros_128(comp_uint128_to_comp_uint128(a));
}

STATIC INLINE unsigned int comp_leading_zeros_160(comp_uint160_t a) {
  int i;
  for (i=0;((i<4) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_192(comp_uint192_t a) {
  int i;
  for (i=0;((i<2) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}

STATIC INLINE unsigned int comp_leading_zeros_224(comp_uint224_t a) {
  int i;
  for (i=0;((i<6) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_256(comp_uint256_t a) {
  int i;
  for (i=0;((i<3) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}

STATIC INLINE unsigned int comp_leading_zeros_288(comp_uint288_t a) {
  int i;
  for (i=0;((i<8) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_320(comp_uint320_t a) {
  int i;
  for (i=0;((i<4) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}

STATIC INLINE unsigned int comp_leading_zeros_352(comp_uint352_t a) {
  int i;
  for (i=0;((i<10) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_384(comp_uint384_t a) {
  int i;
  for (i=0;((i<5) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}

STATIC INLINE unsigned int comp_leading_zeros_416(comp_uint416_t a) {
  int i;
  for (i=0;((i<12) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_448(comp_uint448_t a) {
  int i;
  for (i=0;((i<6) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}

STATIC INLINE unsigned int comp_leading_zeros_480(comp_uint480_t a) {
  int i;
  for (i=0;((i<14) && (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] == ((uint32_t) 0)));i++);
  return (((unsigned int) i) << 5) + comp_leading_zeros_32(a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)]);
}

STATIC INLINE unsigned int comp_leading_zeros_512(comp_uint512_t a) {
  int i;
  for (i=0;((i<7) && (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] == ((uint64_t) 0)));i++);
  return (((unsigned int) i) << 6) + comp_leading_zeros_64(a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)]);
}


/* Sign functions

   Let sgn be defined by


   /  +1  if x > 0
   sgn(x) = |    0  if x = 0
   \  -1  if x < 0

   The functions comp_sign_<t>(a) return 

   sgn((a mod 2^(t - 1)) - floor(a / 2^(t - 1)) * 2^(t - 1))

   i.e. 

   if the input is 

   * all zeros,               the functions return 0
   * has its MSB set to one,  the functions return -1
   * has its MSB set to zero, the functions return +1.

   */
STATIC INLINE int comp_sign_32(comp_uint32_t a) {
  if (a == ((uint32_t) 0)) return 0;
  if ((a & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  return 1;
}

STATIC INLINE int comp_sign_64(comp_uint64_t a) {
  if (a == ((uint64_t) 0)) return 0;
  if ((a & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  return 1;
}

STATIC INLINE int comp_sign_96(comp_uint96_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<3;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_128(comp_uint128_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<2;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_160(comp_uint160_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<5;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_192(comp_uint192_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<3;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_224(comp_uint224_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<7;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_256(comp_uint256_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<4;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_288(comp_uint288_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<9;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_320(comp_uint320_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<5;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_352(comp_uint352_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<11;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_384(comp_uint384_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<6;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_416(comp_uint416_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<13;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_448(comp_uint448_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<7;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_480(comp_uint480_t a) {
  int i;
  if ((a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.w)] & (((uint32_t) 1) << 31)) != ((uint32_t) 0)) return -1;
  for (i=0;i<15;i++) {
    if (a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)] != ((uint32_t) 0)) return 1;
  }
  return 0;
}

STATIC INLINE int comp_sign_512(comp_uint512_t a) {
  int i;
  if ((a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,a.d)] & (((uint64_t) 1) << 63)) != ((uint64_t) 0)) return -1;
  for (i=0;i<8;i++) {
    if (a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)] != ((uint64_t) 0)) return 1;
  }
  return 0;
}


/* Addition functions

   The functions comp_add_<t>(c, a, b) compute

   c = (a + b) mod 2^t

*/
STATIC INLINE void __comp_add_full_adder_32(uint32_t * RESTRICT cout, uint32_t * RESTRICT c, 
					    uint32_t a, uint32_t b, uint32_t cin) {
  uint64_t tmp1, tmp2, tmp3, tmp4;
  uint32_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = cin;
  tmp4 = tmp1 + tmp2 + tmp3;
  tmp5 = (uint32_t) tmp4;
  tmp4 >>= 32;
  tmp6 = (uint32_t) tmp4;
  *c = tmp5;
  *cout = tmp6;
}

STATIC INLINE void __comp_add_full_adder_64(uint64_t * RESTRICT cout, uint64_t * RESTRICT c, 
					    uint64_t a, uint64_t b, uint64_t cin) {
  __uint128_t tmp1, tmp2, tmp3, tmp4;
  uint64_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = cin;
  tmp4 = tmp1 + tmp2 + tmp3;
  tmp5 = (uint64_t) tmp4;
  tmp4 >>= 64;
  tmp6 = (uint64_t) tmp4;
  *c = tmp5;
  *cout = tmp6;
}

STATIC INLINE void comp_add_32(comp_uint32_t * RESTRICT c, comp_uint32_t a, comp_uint32_t b) {
  *c = a + b;
}

STATIC INLINE void comp_add_64(comp_uint64_t * RESTRICT c, comp_uint64_t a, comp_uint64_t b) {
  *c = a + b;
}

STATIC INLINE void comp_add_96(comp_uint96_t * RESTRICT c, comp_uint96_t a, comp_uint96_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=2;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_128(comp_uint128_t * RESTRICT c, comp_uint128_t a, comp_uint128_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=1;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_160(comp_uint160_t * RESTRICT c, comp_uint160_t a, comp_uint160_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=4;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_192(comp_uint192_t * RESTRICT c, comp_uint192_t a, comp_uint192_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=2;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_224(comp_uint224_t * RESTRICT c, comp_uint224_t a, comp_uint224_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=6;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_256(comp_uint256_t * RESTRICT c, comp_uint256_t a, comp_uint256_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=3;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_288(comp_uint288_t * RESTRICT c, comp_uint288_t a, comp_uint288_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=8;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_320(comp_uint320_t * RESTRICT c, comp_uint320_t a, comp_uint320_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=4;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_352(comp_uint352_t * RESTRICT c, comp_uint352_t a, comp_uint352_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=10;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_384(comp_uint384_t * RESTRICT c, comp_uint384_t a, comp_uint384_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=5;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_416(comp_uint416_t * RESTRICT c, comp_uint416_t a, comp_uint416_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=12;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_448(comp_uint448_t * RESTRICT c, comp_uint448_t a, comp_uint448_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=6;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_480(comp_uint480_t * RESTRICT c, comp_uint480_t a, comp_uint480_t b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for(i=14;i>=0;i--) {
    __comp_add_full_adder_32(&cout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], cin);
    cin = cout;
  }
}

STATIC INLINE void comp_add_512(comp_uint512_t * RESTRICT c, comp_uint512_t a, comp_uint512_t b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for(i=7;i>=0;i--) {
    __comp_add_full_adder_64(&cout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], cin);
    cin = cout;
  }
}


/* Subtraction functions

   The functions comp_sub_<t>(c, a, b) compute

   c = (a - b) mod 2^t

*/
STATIC INLINE void __comp_sub_full_subtractor_32(uint32_t * RESTRICT bout, uint32_t * RESTRICT c, 
						 uint32_t a, uint32_t b, uint32_t bin) {
  uint64_t tmp1, tmp2, tmp3, tmp4;
  uint32_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = bin;
  tmp4 = tmp1 - tmp2 - tmp3;
  tmp5 = (uint32_t) tmp4;
  tmp4 >>= 32;
  tmp4 &= (uint64_t) 1;
  tmp6 = (uint32_t) tmp4;
  *c = tmp5;
  *bout = tmp6;
}

STATIC INLINE void __comp_sub_full_subtractor_64(uint64_t * RESTRICT bout, uint64_t * RESTRICT c, 
						 uint64_t a, uint64_t b, uint64_t bin) {
  __uint128_t tmp1, tmp2, tmp3, tmp4;
  uint64_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = bin;
  tmp4 = tmp1 - tmp2 - tmp3;
  tmp5 = (uint64_t) tmp4;
  tmp4 >>= 64;
  tmp4 &= (__uint128_t) 1;
  tmp6 = (uint64_t) tmp4;
  *c = tmp5;
  *bout = tmp6;
}

STATIC INLINE void comp_sub_32(comp_uint32_t * RESTRICT c, comp_uint32_t a, comp_uint32_t b) {
  *c = a + b;
}

STATIC INLINE void comp_sub_64(comp_uint64_t * RESTRICT c, comp_uint64_t a, comp_uint64_t b) {
  *c = a + b;
}

STATIC INLINE void comp_sub_96(comp_uint96_t * RESTRICT c, comp_uint96_t a, comp_uint96_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=2;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_128(comp_uint128_t * RESTRICT c, comp_uint128_t a, comp_uint128_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=1;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_160(comp_uint160_t * RESTRICT c, comp_uint160_t a, comp_uint160_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=4;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_192(comp_uint192_t * RESTRICT c, comp_uint192_t a, comp_uint192_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=2;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_224(comp_uint224_t * RESTRICT c, comp_uint224_t a, comp_uint224_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=6;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_256(comp_uint256_t * RESTRICT c, comp_uint256_t a, comp_uint256_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=3;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_288(comp_uint288_t * RESTRICT c, comp_uint288_t a, comp_uint288_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=8;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_320(comp_uint320_t * RESTRICT c, comp_uint320_t a, comp_uint320_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=4;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_352(comp_uint352_t * RESTRICT c, comp_uint352_t a, comp_uint352_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=10;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_384(comp_uint384_t * RESTRICT c, comp_uint384_t a, comp_uint384_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=5;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_416(comp_uint416_t * RESTRICT c, comp_uint416_t a, comp_uint416_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=12;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_448(comp_uint448_t * RESTRICT c, comp_uint448_t a, comp_uint448_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=6;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_480(comp_uint480_t * RESTRICT c, comp_uint480_t a, comp_uint480_t b) {
  int i;
  uint32_t bin, bout;
  bin = (uint32_t) 0;
  for(i=14;i>=0;i--) {
    __comp_sub_full_subtractor_32(&bout, &(c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->w)]), a.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.w)], b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.w)], bin);
    bin = bout;
  }
}

STATIC INLINE void comp_sub_512(comp_uint512_t * RESTRICT c, comp_uint512_t a, comp_uint512_t b) {
  int i;
  uint64_t bin, bout;
  bin = (uint64_t) 0;
  for(i=7;i>=0;i--) {
    __comp_sub_full_subtractor_64(&bout, &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,c->d)]), a.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,a.d)], b.d[__COMP_ENDIANNESS_IDX_ARR_SIZE(i,b.d)], bin);
    bin = bout;
  }
}


/* Multiplication functions

   The functions comp_mul_<t1>_<t2>_<t3>(c, a, b) compute

   c = a * b

   Only the functions where t3 = t1 + t2 exist, so the 
   products never provoke overflow or modulo arithmetic.

*/

STATIC INLINE void __comp_mul_basic_mul_32(uint32_t * RESTRICT hi, uint32_t * RESTRICT lo, uint32_t a, uint32_t b) {
  uint64_t tmp1, tmp2, tmp3;
  uint32_t tmp4, tmp5;
  tmp1 = a;
  tmp2 = b;
  tmp3 = tmp1 * tmp2;
  tmp4 = (uint32_t) tmp3;
  tmp3 >>= 32;
  tmp5 = (uint32_t) tmp3;
  *hi = tmp5;
  *lo = tmp4;
}

STATIC INLINE void __comp_mul_basic_mul_64(uint64_t * RESTRICT hi, uint64_t * RESTRICT lo, uint64_t a, uint64_t b) {
  __uint128_t tmp1, tmp2, tmp3;
  uint64_t tmp4, tmp5;
  tmp1 = a;
  tmp2 = b;
  tmp3 = tmp1 * tmp2;
  tmp4 = (uint64_t) tmp3;
  tmp3 >>= 64;
  tmp5 = (uint64_t) tmp3;
  *hi = tmp5;
  *lo = tmp4;
}

STATIC INLINE void __comp_mul_carry_save_basic_mul_32(uint64_t * RESTRICT c, uint32_t a, uint32_t b) {
  uint32_t tmp1, tmp2;
  uint64_t tmp3, tmp4;
  __comp_mul_basic_mul_32(&tmp1, &tmp2, a, b);
  tmp3 = tmp1;
  tmp4 = tmp2;
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(0, 2)] = tmp3;
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(1, 2)] = tmp4;
}

STATIC INLINE void __comp_mul_carry_save_basic_mul_64(uint64_t * RESTRICT c, uint64_t a, uint64_t b) {
  uint64_t tmp1, tmp2;
  uint32_t tmp3, tmp4;
  uint64_t tmp5, tmp6;
  __comp_mul_basic_mul_64(&tmp1, &tmp2, a, b);
  tmp3 = (uint32_t) tmp1;
  tmp1 >>= 32;
  tmp4 = (uint32_t) tmp1;
  tmp5 = tmp3;
  tmp6 = tmp4;
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(0, 4)] = tmp6;
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(1, 4)] = tmp5;
  tmp3 = (uint32_t) tmp2;
  tmp2 >>= 32;
  tmp4 = (uint32_t) tmp2;
  tmp5 = tmp3;
  tmp6 = tmp4;
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2, 4)] = tmp6;
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(3, 4)] = tmp5;    
}

STATIC INLINE void __comp_mul_convert_to_32_bit_limb(uint32_t * RESTRICT b, CONST uint64_t * RESTRICT a, int size_in_64_bit_limbs) {
  int i;
  uint64_t tmp1;
  uint32_t tmp2, tmp3;
  for (i=0;i<size_in_64_bit_limbs;i++) {
    tmp1 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i, size_in_64_bit_limbs)];
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2 * i, 2 * size_in_64_bit_limbs)] = tmp3;
    b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2 * i + 1, 2 * size_in_64_bit_limbs)] = tmp2;
  }
}

STATIC INLINE void __comp_mul_convert_from_32_bit_limb(uint64_t * RESTRICT b, CONST uint32_t * RESTRICT a, int size_in_64_bit_limbs) {
  int i;
  uint32_t tmp1, tmp2;
  uint64_t tmp3;
  for (i=0;i<size_in_64_bit_limbs;i++) {
    tmp1 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2 * i, size_in_64_bit_limbs)];
    tmp2 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2 * i + 1, size_in_64_bit_limbs)];
    tmp3 = tmp1;
    tmp3 <<= 32;
    tmp3 |= tmp2;
    b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i, size_in_64_bit_limbs)] = tmp3;
  }
}

STATIC INLINE void __comp_mul_acc_carry_save(uint64_t * RESTRICT acc, int k, CONST uint64_t * RESTRICT a, int size_in_32_bit_limbs_acc, int size_in_32_bit_limbs_a) {
  int i;
  (void) size_in_32_bit_limbs_acc;
  for (i=k;i>k-size_in_32_bit_limbs_a;i--) {
    acc[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i,size_in_32_bit_limbs_acc)] += a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i-k+size_in_32_bit_limbs_a-1,size_in_32_bit_limbs_a)];
  }
}

/* The output array b is supposed to have size_in_32_bit_limbs_in_b
   elements, a is supposed to have size_in_32_bit_limbs_in_b
   elements.
*/
STATIC INLINE void __comp_mul_reduce_carry_save_32(uint32_t * RESTRICT b, CONST uint64_t * RESTRICT a, int size_in_32_bit_limbs_in_b) {
  int i;
  uint64_t cin, cout, tmp1;
  uint32_t tmp2, tmp3;
  cin = (uint64_t) 0;
  for (i=size_in_32_bit_limbs_in_b-1;i>=0;i--) {
    tmp1 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i,size_in_32_bit_limbs_in_b)];
    tmp1 += cin;
    tmp2 = (uint32_t) tmp1;
    tmp1 >>= 32;
    tmp3 = (uint32_t) tmp1;
    b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i,size_in_32_bit_limbs_in_b)] = tmp2;
    cout = tmp3;
    cin = cout;
  }
}

/* The output array b is supposed to have size_in_64_bit_limbs_in_b
   elements, a is supposed to have 2 * size_in_64_bit_limbs_in_b
   elements.
*/
STATIC INLINE void __comp_mul_reduce_carry_save_64(uint64_t * RESTRICT b, CONST uint64_t * RESTRICT a, int size_in_64_bit_limbs_in_b) {
  int i;
  uint64_t tmp1, tmp2;
  __uint128_t cin, cout, tmp3, tmp4, tmp5;
  uint64_t tmp6, tmp7;
  __uint128_t tmp8;
  cin = (__uint128_t) 0;
  for (i=0;i<size_in_64_bit_limbs_in_b;i++) {
    tmp1 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2 * size_in_64_bit_limbs_in_b - 1 - 2 * i,2 * size_in_64_bit_limbs_in_b)];
    tmp2 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(2 * size_in_64_bit_limbs_in_b - 2 - 2 * i,2 * size_in_64_bit_limbs_in_b)];
    tmp3 = tmp1;
    tmp4 = tmp2;
    tmp4 <<= 32;
    tmp5 = tmp4 + tmp3 + cin;
    tmp6 = (uint64_t) tmp5;
    tmp5 >>= 64;
    tmp7 = (uint64_t) tmp5;
    tmp8 = tmp7;
    cout = tmp8;
    b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_in_64_bit_limbs_in_b-1-i, size_in_64_bit_limbs_in_b)] = tmp6;
    cin = cout;
  }
}

/* The array c is supposed to have size_a + size_b elements */
STATIC INLINE void __comp_mul_carry_save_mul_32(uint64_t * RESTRICT c, CONST uint32_t * RESTRICT a, int size_a, CONST uint32_t * RESTRICT b, int size_b) {
  int i, j, k;
  uint32_t tmp1, tmp2;
  uint64_t tmp3[2];
  for (i=0;i<(size_a+size_b);i++) {
    c[i] = (uint64_t) 0;
  }
  for (i=0;i<size_a;i++) {
    tmp1 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a - 1 - i,size_a)];
    for (j=0;j<size_b;j++) {
      tmp2 = b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - 1 - j,size_b)];
      __comp_mul_carry_save_basic_mul_32(tmp3, tmp1, tmp2);
      k = i + j;
      __comp_mul_acc_carry_save(c, size_a + size_b - 1 - k, tmp3, size_a + size_b, 2);
    }
  }
}

/* The array c is supposed to have 2 * (size_a + size_b) elements */
STATIC INLINE void __comp_mul_carry_save_mul_64(uint64_t * RESTRICT c, CONST uint64_t * RESTRICT a, int size_a, CONST uint64_t * RESTRICT b, int size_b) {
  int i, j, k;
  uint64_t tmp1, tmp2;
  uint64_t tmp3[4];
  for (i=0;i<(2*(size_a+size_b));i++) {
    c[i] = (uint64_t) 0;
  }
  for (i=0;i<size_a;i++) {
    tmp1 = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a - 1 - i,size_a)];
    for (j=0;j<size_b;j++) {
      tmp2 = b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - 1 - j,size_b)];
      __comp_mul_carry_save_basic_mul_64(tmp3, tmp1, tmp2);
      k = i + j;
      __comp_mul_acc_carry_save(c, 2 * (size_a + size_b) - 1 - 2 * k, tmp3, 2 * (size_a + size_b), 4);
    }
  }
}

/* Computes hi * 2^32 + lo = a * b + c */
STATIC INLINE void __comp_mul_and_add_32(uint32_t * RESTRICT hi, uint32_t * RESTRICT lo, uint32_t a, uint32_t b, uint32_t c) {
  uint64_t tmp1, tmp2, tmp3, tmp4;
  uint32_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = c;
  tmp4 = tmp1 * tmp2 + tmp3;
  tmp5 = (uint32_t) tmp4;
  tmp4 >>= 32;
  tmp6 = (uint32_t) tmp4;
  *hi = tmp6;
  *lo = tmp5;
}

/* Computes hi * 2^64 + lo = a * b + c */
STATIC INLINE void __comp_mul_and_add_64(uint64_t * RESTRICT hi, uint64_t * RESTRICT lo, uint64_t a, uint64_t b, uint64_t c) {
  __uint128_t tmp1, tmp2, tmp3, tmp4;
  uint64_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = c;
  tmp4 = tmp1 * tmp2 + tmp3;
  tmp5 = (uint64_t) tmp4;
  tmp4 >>= 64;
  tmp6 = (uint64_t) tmp4;
  *hi = tmp6;
  *lo = tmp5;
}

STATIC INLINE void __comp_mul_basic_32_array(uint32_t * RESTRICT c, uint32_t a, CONST uint32_t * RESTRICT b, int size_b) {
  int i;
  uint32_t cin, cout;
  cin = (uint32_t) 0;
  for (i=0;i<size_b;i++) {
    __comp_mul_and_add_32(&cout, &(c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - i, size_b + 1)]), 
			  a, b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - 1 - i, size_b)], cin);
    cin = cout;
  }
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(0, size_b + 1)] = cin;
}

STATIC INLINE void __comp_mul_basic_64_array(uint64_t * RESTRICT c, uint64_t a, CONST uint64_t * RESTRICT b, int size_b) {
  int i;
  uint64_t cin, cout;
  cin = (uint64_t) 0;
  for (i=0;i<size_b;i++) {
    __comp_mul_and_add_64(&cout, &(c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - i, size_b + 1)]), 
			  a, b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - 1 - i, size_b)], cin);
    cin = cout;
  }
  c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(0, size_b + 1)] = cin;
}

/* Computes hi * 2^64 + lo = a * b + c + d */
STATIC INLINE void __comp_mul_and_twice_add_64(uint64_t * RESTRICT hi, uint64_t * RESTRICT lo, uint64_t a, uint64_t b, uint64_t c, uint64_t d) {
  __uint128_t tmp1, tmp2, tmp3, tmp4, tmp7;
  uint64_t tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = c;
  tmp7 = d;
  tmp4 = tmp1 * tmp2 + tmp3 + tmp7;
  tmp5 = (uint64_t) tmp4;
  tmp4 >>= 64;
  tmp6 = (uint64_t) tmp4;
  *hi = tmp6;
  *lo = tmp5;
}

STATIC INLINE void __comp_mul_array_array_64(uint64_t * RESTRICT c, uint64_t * RESTRICT a, int size_a, CONST uint64_t * RESTRICT b, int size_b) {
  int i, j;
  uint64_t elem_a, elem_b, elem_c, cin, cout, hi, lo;
  __uint128_t tmp1, tmp2, tmp3;
  for (i=0;i<size_a+size_b;i++) {
    c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(i, size_a + size_b)] = (uint64_t) 0;
  }
  for (i=0;i<size_a;i++) {
    elem_a = a[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a - 1 - i, size_a)];
    cin = (uint64_t) 0;
    for (j=0;j<size_b;j++) {
      elem_b = b[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_b - 1 - j, size_b)];
      elem_c = c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a + size_b - 1 - i - j, size_a + size_b)];
      __comp_mul_and_twice_add_64(&hi, &lo, elem_a, elem_b, elem_c, cin);
      elem_c = lo;
      cout = hi;
      c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a + size_b - 1 - i - j, size_a + size_b)] = elem_c;
      cin = cout;
    }
    for (;j<size_a + size_b - i;j++) {
      elem_c = c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a + size_b - 1 - i - j, size_a + size_b)];
      tmp1 = elem_c;
      tmp2 = cin;
      tmp3 = tmp1 + tmp2;
      elem_c = (uint64_t) tmp3;
      tmp3 >>= 64;
      cout = (uint64_t) tmp3;
      c[__COMP_ENDIANNESS_IDX_NUMBER_ELEMS(size_a + size_b - 1 - i - j, size_a + size_b)] = elem_c;
      cin = cout;
    }
  }
}

STATIC INLINE void comp_mul_32_32_64(comp_uint64_t * RESTRICT c, comp_uint32_t a, comp_uint32_t b) {
  uint64_t tmp1, tmp2, tmp3;
  tmp1 = a;
  tmp2 = b;
  tmp3 = tmp1 * tmp2;
  *c = tmp3;
}

STATIC INLINE void comp_mul_32_64_96(comp_uint96_t * RESTRICT c, comp_uint32_t a, comp_uint64_t b) {
  __uint128_t tmp1, tmp2, tmp3;
  uint32_t tmp4, tmp5, tmp6;
  tmp1 = a;
  tmp2 = b;
  tmp3 = tmp1 * tmp2;
  tmp4 = (uint32_t) tmp3;
  tmp3 >>= 32;
  tmp5 = (uint32_t) tmp3;
  tmp3 >>= 32;
  tmp6 = (uint32_t) tmp3;
  c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,c->w)] = tmp6;
  c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,c->w)] = tmp5;
  c->w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,c->w)] = tmp4;
}

STATIC INLINE void comp_mul_32_96_128(comp_uint128_t * RESTRICT c, comp_uint32_t a, comp_uint96_t b) {
  __uint128_t tmp1, tmp2, tmp3;
  uint64_t tmp4, tmp5;
  tmp1 = b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,b.w)];
  tmp1 <<= 32;
  tmp1 |= b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,b.w)];
  tmp1 <<= 32;
  tmp1 |= b.w[__COMP_ENDIANNESS_IDX_ARR_SIZE(2,b.w)];
  tmp2 = a;
  tmp3 = tmp1 * tmp2;
  tmp4 = (uint64_t) tmp3;
  tmp3 >>= 64;
  tmp5 = (uint64_t) tmp3;
  c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,c->d)] = tmp5;
  c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,c->d)] = tmp4;
}

STATIC INLINE void comp_mul_64_64_128(comp_uint128_t * RESTRICT c, comp_uint64_t a, comp_uint64_t b) {
  __comp_mul_basic_mul_64(&(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(0,c->d)]), &(c->d[__COMP_ENDIANNESS_IDX_ARR_SIZE(1,c->d)]), a, b);
}

STATIC INLINE void comp_mul_32_160_192(comp_uint192_t * RESTRICT c, comp_uint32_t a, comp_uint160_t b) {
    uint32_t csplit[6];
    __comp_mul_basic_32_array(csplit, a, b.w, 5);
    __comp_mul_convert_from_32_bit_limb(c->d, csplit, 3);
}

STATIC INLINE void comp_mul_32_224_256(comp_uint256_t * RESTRICT c, comp_uint32_t a, comp_uint224_t b) {
    uint32_t csplit[8];
    __comp_mul_basic_32_array(csplit, a, b.w, 7);
    __comp_mul_convert_from_32_bit_limb(c->d, csplit, 4);
}

STATIC INLINE void comp_mul_32_288_320(comp_uint320_t * RESTRICT c, comp_uint32_t a, comp_uint288_t b) {
    uint32_t csplit[10];
    __comp_mul_basic_32_array(csplit, a, b.w, 9);
    __comp_mul_convert_from_32_bit_limb(c->d, csplit, 5);
}

STATIC INLINE void comp_mul_32_352_384(comp_uint384_t * RESTRICT c, comp_uint32_t a, comp_uint352_t b) {
    uint32_t csplit[12];
    __comp_mul_basic_32_array(csplit, a, b.w, 11);
    __comp_mul_convert_from_32_bit_limb(c->d, csplit, 6);
}

STATIC INLINE void comp_mul_32_416_448(comp_uint448_t * RESTRICT c, comp_uint32_t a, comp_uint416_t b) {
    uint32_t csplit[14];
    __comp_mul_basic_32_array(csplit, a, b.w, 13);
    __comp_mul_convert_from_32_bit_limb(c->d, csplit, 7);
}

STATIC INLINE void comp_mul_32_480_512(comp_uint512_t * RESTRICT c, comp_uint32_t a, comp_uint480_t b) {
    uint32_t csplit[16];
    __comp_mul_basic_32_array(csplit, a, b.w, 15);
    __comp_mul_convert_from_32_bit_limb(c->d, csplit, 8);
}

STATIC INLINE void comp_mul_64_128_192(comp_uint192_t * RESTRICT c, comp_uint64_t a, comp_uint128_t b) {
    __comp_mul_basic_64_array(c->d, a, b.d, 2);
}

STATIC INLINE void comp_mul_64_96_160(comp_uint160_t * RESTRICT c, comp_uint64_t a, comp_uint96_t b) {
    comp_uint128_t tmp1;
    comp_uint192_t tmp2;
    comp_upconvert_low_96_128(&tmp1, b);
    comp_mul_64_128_192(&tmp2, a, tmp1);
    comp_downconvert_low_192_160(c, tmp2);
}

STATIC INLINE void comp_mul_64_192_256(comp_uint256_t * RESTRICT c, comp_uint64_t a, comp_uint192_t b) {
    __comp_mul_basic_64_array(c->d, a, b.d, 3);
}

STATIC INLINE void comp_mul_64_160_224(comp_uint224_t * RESTRICT c, comp_uint64_t a, comp_uint160_t b) {
    comp_uint192_t tmp1;
    comp_uint256_t tmp2;
    comp_upconvert_low_160_192(&tmp1, b);
    comp_mul_64_192_256(&tmp2, a, tmp1);
    comp_downconvert_low_256_224(c, tmp2);
}

STATIC INLINE void comp_mul_64_256_320(comp_uint320_t * RESTRICT c, comp_uint64_t a, comp_uint256_t b) {
    __comp_mul_basic_64_array(c->d, a, b.d, 4);
}

STATIC INLINE void comp_mul_64_224_288(comp_uint288_t * RESTRICT c, comp_uint64_t a, comp_uint224_t b) {
    comp_uint256_t tmp1;
    comp_uint320_t tmp2;
    comp_upconvert_low_224_256(&tmp1, b);
    comp_mul_64_256_320(&tmp2, a, tmp1);
    comp_downconvert_low_320_288(c, tmp2);
}

STATIC INLINE void comp_mul_64_320_384(comp_uint384_t * RESTRICT c, comp_uint64_t a, comp_uint320_t b) {
    __comp_mul_basic_64_array(c->d, a, b.d, 5);
}

STATIC INLINE void comp_mul_64_288_352(comp_uint352_t * RESTRICT c, comp_uint64_t a, comp_uint288_t b) {
    comp_uint320_t tmp1;
    comp_uint384_t tmp2;
    comp_upconvert_low_288_320(&tmp1, b);
    comp_mul_64_320_384(&tmp2, a, tmp1);
    comp_downconvert_low_384_352(c, tmp2);
}

STATIC INLINE void comp_mul_64_384_448(comp_uint448_t * RESTRICT c, comp_uint64_t a, comp_uint384_t b) {
    __comp_mul_basic_64_array(c->d, a, b.d, 6);
}

STATIC INLINE void comp_mul_64_352_416(comp_uint416_t * RESTRICT c, comp_uint64_t a, comp_uint352_t b) {
    comp_uint384_t tmp1;
    comp_uint448_t tmp2;
    comp_upconvert_low_352_384(&tmp1, b);
    comp_mul_64_384_448(&tmp2, a, tmp1);
    comp_downconvert_low_448_416(c, tmp2);
}

STATIC INLINE void comp_mul_64_448_512(comp_uint512_t * RESTRICT c, comp_uint64_t a, comp_uint448_t b) {
    __comp_mul_basic_64_array(c->d, a, b.d, 7);
}

STATIC INLINE void comp_mul_64_416_480(comp_uint480_t * RESTRICT c, comp_uint64_t a, comp_uint416_t b) {
    comp_uint448_t tmp1;
    comp_uint512_t tmp2;
    comp_upconvert_low_416_448(&tmp1, b);
    comp_mul_64_448_512(&tmp2, a, tmp1);
    comp_downconvert_low_512_480(c, tmp2);
}

STATIC INLINE void comp_mul_32_128_160(comp_uint160_t * RESTRICT c, comp_uint32_t a, comp_uint128_t b) {
    comp_uint64_t tmp1;
    comp_uint192_t tmp2;
    comp_upconvert_low_32_64(&tmp1, a);
    comp_mul_64_128_192(&tmp2,tmp1,b);
    comp_downconvert_low_192_160(c, tmp2);
}

STATIC INLINE void comp_mul_32_192_224(comp_uint224_t * RESTRICT c, comp_uint32_t a, comp_uint192_t b) {
    comp_uint64_t tmp1;
    comp_uint256_t tmp2;
    comp_upconvert_low_32_64(&tmp1, a);
    comp_mul_64_192_256(&tmp2,tmp1,b);
    comp_downconvert_low_256_224(c, tmp2);
}

STATIC INLINE void comp_mul_32_256_288(comp_uint288_t * RESTRICT c, comp_uint32_t a, comp_uint256_t b) {
    comp_uint64_t tmp1;
    comp_uint320_t tmp2;
    comp_upconvert_low_32_64(&tmp1, a);
    comp_mul_64_256_320(&tmp2,tmp1,b);
    comp_downconvert_low_320_288(c, tmp2);
}

STATIC INLINE void comp_mul_32_320_352(comp_uint352_t * RESTRICT c, comp_uint32_t a, comp_uint320_t b) {
    comp_uint64_t tmp1;
    comp_uint384_t tmp2;
    comp_upconvert_low_32_64(&tmp1, a);
    comp_mul_64_320_384(&tmp2,tmp1,b);
    comp_downconvert_low_384_352(c, tmp2);
}

STATIC INLINE void comp_mul_32_384_416(comp_uint416_t * RESTRICT c, comp_uint32_t a, comp_uint384_t b) {
    comp_uint64_t tmp1;
    comp_uint448_t tmp2;
    comp_upconvert_low_32_64(&tmp1, a);
    comp_mul_64_384_448(&tmp2,tmp1,b);
    comp_downconvert_low_448_416(c, tmp2);
}

STATIC INLINE void comp_mul_32_448_480(comp_uint480_t * RESTRICT c, comp_uint32_t a, comp_uint448_t b) {
    comp_uint64_t tmp1;
    comp_uint512_t tmp2;
    comp_upconvert_low_32_64(&tmp1, a);
    comp_mul_64_448_512(&tmp2,tmp1,b);
    comp_downconvert_low_512_480(c, tmp2);
}

STATIC INLINE void comp_mul_96_96_192(comp_uint192_t * RESTRICT c, comp_uint96_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[6];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 3);
}

STATIC INLINE void comp_mul_96_128_224(comp_uint224_t * RESTRICT c, comp_uint96_t a, comp_uint128_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[4];
  uint64_t cs[7];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 2);
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 4);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 7);
}

STATIC INLINE void comp_mul_96_160_256(comp_uint256_t * RESTRICT c, comp_uint96_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[8];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 5);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 4);
}

STATIC INLINE void comp_mul_96_192_288(comp_uint288_t * RESTRICT c, comp_uint96_t a, comp_uint192_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[6];
  uint64_t cs[9];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 3);
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 6);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 9);
}

STATIC INLINE void comp_mul_96_224_320(comp_uint320_t * RESTRICT c, comp_uint96_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[10];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 7);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 5);
}

STATIC INLINE void comp_mul_96_256_352(comp_uint352_t * RESTRICT c, comp_uint96_t a, comp_uint256_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[8];
  uint64_t cs[11];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 4);
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 8);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 11);
}

STATIC INLINE void comp_mul_96_288_384(comp_uint384_t * RESTRICT c, comp_uint96_t a, comp_uint288_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[12];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 9);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 6);
}

STATIC INLINE void comp_mul_96_320_416(comp_uint416_t * RESTRICT c, comp_uint96_t a, comp_uint320_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[10];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 5);
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 10);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_96_352_448(comp_uint448_t * RESTRICT c, comp_uint96_t a, comp_uint352_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 11);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_96_384_480(comp_uint480_t * RESTRICT c, comp_uint96_t a, comp_uint384_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[12];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 6);
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 12);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_96_416_512(comp_uint512_t * RESTRICT c, comp_uint96_t a, comp_uint416_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 3, bptr, 13);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_128_96_224(comp_uint224_t * RESTRICT c, comp_uint128_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[4];
  uint64_t cs[7];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 2);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 4, bptr, 3);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 7);
}

STATIC INLINE void comp_mul_128_128_256(comp_uint256_t * RESTRICT c, comp_uint128_t a, comp_uint128_t b) {
  __comp_mul_array_array_64(c->d, a.d, 2, b.d, 2);
}

STATIC INLINE void comp_mul_128_160_288(comp_uint288_t * RESTRICT c, comp_uint128_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[4];
  uint64_t cs[9];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 2);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 4, bptr, 5);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 9);
}

STATIC INLINE void comp_mul_128_192_320(comp_uint320_t * RESTRICT c, comp_uint128_t a, comp_uint192_t b) {
  __comp_mul_array_array_64(c->d, a.d, 2, b.d, 3);
}

STATIC INLINE void comp_mul_128_224_352(comp_uint352_t * RESTRICT c, comp_uint128_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[4];
  uint64_t cs[11];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 2);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 4, bptr, 7);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 11);
}

STATIC INLINE void comp_mul_128_256_384(comp_uint384_t * RESTRICT c, comp_uint128_t a, comp_uint256_t b) {
  __comp_mul_array_array_64(c->d, a.d, 2, b.d, 4);
}

STATIC INLINE void comp_mul_128_288_416(comp_uint416_t * RESTRICT c, comp_uint128_t a, comp_uint288_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[4];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 2);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 4, bptr, 9);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_128_320_448(comp_uint448_t * RESTRICT c, comp_uint128_t a, comp_uint320_t b) {
  __comp_mul_array_array_64(c->d, a.d, 2, b.d, 5);
}

STATIC INLINE void comp_mul_128_352_480(comp_uint480_t * RESTRICT c, comp_uint128_t a, comp_uint352_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[4];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 2);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 4, bptr, 11);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_128_384_512(comp_uint512_t * RESTRICT c, comp_uint128_t a, comp_uint384_t b) {
  __comp_mul_array_array_64(c->d, a.d, 2, b.d, 6);
}

STATIC INLINE void comp_mul_160_96_256(comp_uint256_t * RESTRICT c, comp_uint160_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[8];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 4);
}

STATIC INLINE void comp_mul_160_128_288(comp_uint288_t * RESTRICT c, comp_uint160_t a, comp_uint128_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[4];
  uint64_t cs[9];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 2);
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 4);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 9);
}

STATIC INLINE void comp_mul_160_160_320(comp_uint320_t * RESTRICT c, comp_uint160_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[10];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 5);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 5);
}

STATIC INLINE void comp_mul_160_192_352(comp_uint352_t * RESTRICT c, comp_uint160_t a, comp_uint192_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[6];
  uint64_t cs[11];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 3);
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 6);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 11);
}

STATIC INLINE void comp_mul_160_224_384(comp_uint384_t * RESTRICT c, comp_uint160_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[12];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 7);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 6);
}

STATIC INLINE void comp_mul_160_256_416(comp_uint416_t * RESTRICT c, comp_uint160_t a, comp_uint256_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[8];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 4);
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 8);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_160_288_448(comp_uint448_t * RESTRICT c, comp_uint160_t a, comp_uint288_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 9);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_160_320_480(comp_uint480_t * RESTRICT c, comp_uint160_t a, comp_uint320_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[10];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 5);
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 10);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_160_352_512(comp_uint512_t * RESTRICT c, comp_uint160_t a, comp_uint352_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 5, bptr, 11);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_192_96_288(comp_uint288_t * RESTRICT c, comp_uint192_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[6];
  uint64_t cs[9];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 3);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 6, bptr, 3);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 9);
}

STATIC INLINE void comp_mul_192_128_320(comp_uint320_t * RESTRICT c, comp_uint192_t a, comp_uint128_t b) {
  comp_mul_128_192_320(c, b, a);
}

STATIC INLINE void comp_mul_192_160_352(comp_uint352_t * RESTRICT c, comp_uint192_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[6];
  uint64_t cs[11];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 3);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 6, bptr, 5);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 11);
}

STATIC INLINE void comp_mul_192_192_384(comp_uint384_t * RESTRICT c, comp_uint192_t a, comp_uint192_t b) {
  uint64_t *aptr, *bptr;
  uint64_t cs[12];
  uint64_t *cptr;
  aptr = a.d;
  bptr = b.d;
  __comp_mul_carry_save_mul_64(cs, aptr, 3, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 6);
}

STATIC INLINE void comp_mul_192_224_416(comp_uint416_t * RESTRICT c, comp_uint192_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[6];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 3);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 6, bptr, 7);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_192_256_448(comp_uint448_t * RESTRICT c, comp_uint192_t a, comp_uint256_t b) {
  uint64_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.d;
  bptr = b.d;
  __comp_mul_carry_save_mul_64(cs, aptr, 3, bptr, 4);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_192_288_480(comp_uint480_t * RESTRICT c, comp_uint192_t a, comp_uint288_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[6];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 3);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 6, bptr, 9);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_192_320_512(comp_uint512_t * RESTRICT c, comp_uint192_t a, comp_uint320_t b) {
  uint64_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.d;
  bptr = b.d;
  __comp_mul_carry_save_mul_64(cs, aptr, 3, bptr, 5);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_224_96_320(comp_uint320_t * RESTRICT c, comp_uint224_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[10];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 5);
}

STATIC INLINE void comp_mul_224_128_352(comp_uint352_t * RESTRICT c, comp_uint224_t a, comp_uint128_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[4];
  uint64_t cs[11];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 2);
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 4);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 11);
}

STATIC INLINE void comp_mul_224_160_384(comp_uint384_t * RESTRICT c, comp_uint224_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[12];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 5);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 6);
}

STATIC INLINE void comp_mul_224_192_416(comp_uint416_t * RESTRICT c, comp_uint224_t a, comp_uint192_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[6];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 3);
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 6);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_224_224_448(comp_uint448_t * RESTRICT c, comp_uint224_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 7);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_224_256_480(comp_uint480_t * RESTRICT c, comp_uint224_t a, comp_uint256_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[8];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 4);
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 8);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_224_288_512(comp_uint512_t * RESTRICT c, comp_uint224_t a, comp_uint288_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 7, bptr, 9);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_256_96_352(comp_uint352_t * RESTRICT c, comp_uint256_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[8];
  uint64_t cs[11];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 4);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 8, bptr, 3);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 11);
}

STATIC INLINE void comp_mul_256_128_384(comp_uint384_t * RESTRICT c, comp_uint256_t a, comp_uint128_t b) {
  comp_mul_128_256_384(c, b, a);
}

STATIC INLINE void comp_mul_256_160_416(comp_uint416_t * RESTRICT c, comp_uint256_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[8];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 4);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 8, bptr, 5);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_256_192_448(comp_uint448_t * RESTRICT c, comp_uint256_t a, comp_uint192_t b) {
  uint64_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.d;
  bptr = b.d;
  __comp_mul_carry_save_mul_64(cs, aptr, 4, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_256_224_480(comp_uint480_t * RESTRICT c, comp_uint256_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[8];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 4);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 8, bptr, 7);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_256_256_512(comp_uint512_t * RESTRICT c, comp_uint256_t a, comp_uint256_t b) {
  __comp_mul_array_array_64(c->d, a.d, 4, b.d, 4);
}

STATIC INLINE void comp_mul_288_96_384(comp_uint384_t * RESTRICT c, comp_uint288_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[12];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 9, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 6);
}

STATIC INLINE void comp_mul_288_128_416(comp_uint416_t * RESTRICT c, comp_uint288_t a, comp_uint128_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[4];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 2);
  __comp_mul_carry_save_mul_32(cs, aptr, 9, bptr, 4);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_288_160_448(comp_uint448_t * RESTRICT c, comp_uint288_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 9, bptr, 5);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_288_192_480(comp_uint480_t * RESTRICT c, comp_uint288_t a, comp_uint192_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[6];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 3);
  __comp_mul_carry_save_mul_32(cs, aptr, 9, bptr, 6);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_288_224_512(comp_uint512_t * RESTRICT c, comp_uint288_t a, comp_uint224_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 9, bptr, 7);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_320_96_416(comp_uint416_t * RESTRICT c, comp_uint320_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[10];
  uint64_t cs[13];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 5);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 10, bptr, 3);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 13);
}

STATIC INLINE void comp_mul_320_128_448(comp_uint448_t * RESTRICT c, comp_uint320_t a, comp_uint128_t b) {
  comp_mul_128_320_448(c, b, a);
}

STATIC INLINE void comp_mul_320_160_480(comp_uint480_t * RESTRICT c, comp_uint320_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[10];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 5);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 10, bptr, 5);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_320_192_512(comp_uint512_t * RESTRICT c, comp_uint320_t a, comp_uint192_t b) {
  uint64_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.d;
  bptr = b.d;
  __comp_mul_carry_save_mul_64(cs, aptr, 5, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_352_96_448(comp_uint448_t * RESTRICT c, comp_uint352_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[14];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 11, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 7);
}

STATIC INLINE void comp_mul_352_128_480(comp_uint480_t * RESTRICT c, comp_uint352_t a, comp_uint128_t b) {
  uint32_t *aptr, *bptr;
  uint32_t bsplit[4];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = a.w;
  bptr = bsplit;
  __comp_mul_convert_to_32_bit_limb(bsplit, b.d, 2);
  __comp_mul_carry_save_mul_32(cs, aptr, 11, bptr, 4);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_352_160_512(comp_uint512_t * RESTRICT c, comp_uint352_t a, comp_uint160_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 11, bptr, 5);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_384_96_480(comp_uint480_t * RESTRICT c, comp_uint384_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint32_t asplit[12];
  uint64_t cs[15];
  uint32_t *cptr;
  aptr = asplit;
  __comp_mul_convert_to_32_bit_limb(asplit, a.d, 6);
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 12, bptr, 3);
  cptr = c->w;
  __comp_mul_reduce_carry_save_32(cptr, cs, 15);
}

STATIC INLINE void comp_mul_384_128_512(comp_uint512_t * RESTRICT c, comp_uint384_t a, comp_uint128_t b) {
  comp_mul_128_384_512(c, b, a);
}

STATIC INLINE void comp_mul_416_96_512(comp_uint512_t * RESTRICT c, comp_uint416_t a, comp_uint96_t b) {
  uint32_t *aptr, *bptr;
  uint64_t cs[16];
  uint64_t *cptr;
  aptr = a.w;
  bptr = b.w;
  __comp_mul_carry_save_mul_32(cs, aptr, 13, bptr, 3);
  cptr = c->d;
  __comp_mul_reduce_carry_save_64(cptr, cs, 8);
}

STATIC INLINE void comp_mul_64_32_96(comp_uint96_t * RESTRICT c, comp_uint64_t a, comp_uint32_t b) {
    comp_mul_32_64_96(c,b,a);
}

STATIC INLINE void comp_mul_96_32_128(comp_uint128_t * RESTRICT c, comp_uint96_t a, comp_uint32_t b) {
    comp_mul_32_96_128(c,b,a);
}

STATIC INLINE void comp_mul_96_64_160(comp_uint160_t * RESTRICT c, comp_uint96_t a, comp_uint64_t b) {
    comp_mul_64_96_160(c,b,a);
}

STATIC INLINE void comp_mul_128_32_160(comp_uint160_t * RESTRICT c, comp_uint128_t a, comp_uint32_t b) {
    comp_mul_32_128_160(c,b,a);
}

STATIC INLINE void comp_mul_128_64_192(comp_uint192_t * RESTRICT c, comp_uint128_t a, comp_uint64_t b) {
    comp_mul_64_128_192(c,b,a);
}

STATIC INLINE void comp_mul_160_32_192(comp_uint192_t * RESTRICT c, comp_uint160_t a, comp_uint32_t b) {
    comp_mul_32_160_192(c,b,a);
}

STATIC INLINE void comp_mul_160_64_224(comp_uint224_t * RESTRICT c, comp_uint160_t a, comp_uint64_t b) {
    comp_mul_64_160_224(c,b,a);
}

STATIC INLINE void comp_mul_192_32_224(comp_uint224_t * RESTRICT c, comp_uint192_t a, comp_uint32_t b) {
    comp_mul_32_192_224(c,b,a);
}

STATIC INLINE void comp_mul_192_64_256(comp_uint256_t * RESTRICT c, comp_uint192_t a, comp_uint64_t b) {
    comp_mul_64_192_256(c,b,a);
}

STATIC INLINE void comp_mul_224_32_256(comp_uint256_t * RESTRICT c, comp_uint224_t a, comp_uint32_t b) {
    comp_mul_32_224_256(c,b,a);
}

STATIC INLINE void comp_mul_224_64_288(comp_uint288_t * RESTRICT c, comp_uint224_t a, comp_uint64_t b) {
    comp_mul_64_224_288(c,b,a);
}

STATIC INLINE void comp_mul_256_32_288(comp_uint288_t * RESTRICT c, comp_uint256_t a, comp_uint32_t b) {
    comp_mul_32_256_288(c,b,a);
}

STATIC INLINE void comp_mul_256_64_320(comp_uint320_t * RESTRICT c, comp_uint256_t a, comp_uint64_t b) {
    comp_mul_64_256_320(c,b,a);
}

STATIC INLINE void comp_mul_288_32_320(comp_uint320_t * RESTRICT c, comp_uint288_t a, comp_uint32_t b) {
    comp_mul_32_288_320(c,b,a);
}

STATIC INLINE void comp_mul_288_64_352(comp_uint352_t * RESTRICT c, comp_uint288_t a, comp_uint64_t b) {
    comp_mul_64_288_352(c,b,a);
}

STATIC INLINE void comp_mul_320_32_352(comp_uint352_t * RESTRICT c, comp_uint320_t a, comp_uint32_t b) {
    comp_mul_32_320_352(c,b,a);
}

STATIC INLINE void comp_mul_320_64_384(comp_uint384_t * RESTRICT c, comp_uint320_t a, comp_uint64_t b) {
    comp_mul_64_320_384(c,b,a);
}

STATIC INLINE void comp_mul_352_32_384(comp_uint384_t * RESTRICT c, comp_uint352_t a, comp_uint32_t b) {
    comp_mul_32_352_384(c,b,a);
}

STATIC INLINE void comp_mul_352_64_416(comp_uint416_t * RESTRICT c, comp_uint352_t a, comp_uint64_t b) {
    comp_mul_64_352_416(c,b,a);
}

STATIC INLINE void comp_mul_384_32_416(comp_uint416_t * RESTRICT c, comp_uint384_t a, comp_uint32_t b) {
    comp_mul_32_384_416(c,b,a);
}

STATIC INLINE void comp_mul_384_64_448(comp_uint448_t * RESTRICT c, comp_uint384_t a, comp_uint64_t b) {
    comp_mul_64_384_448(c,b,a);
}

STATIC INLINE void comp_mul_416_32_448(comp_uint448_t * RESTRICT c, comp_uint416_t a, comp_uint32_t b) {
    comp_mul_32_416_448(c,b,a);
}

STATIC INLINE void comp_mul_416_64_480(comp_uint480_t * RESTRICT c, comp_uint416_t a, comp_uint64_t b) {
    comp_mul_64_416_480(c,b,a);
}

STATIC INLINE void comp_mul_448_32_480(comp_uint480_t * RESTRICT c, comp_uint448_t a, comp_uint32_t b) {
    comp_mul_32_448_480(c,b,a);
}

STATIC INLINE void comp_mul_448_64_512(comp_uint512_t * RESTRICT c, comp_uint448_t a, comp_uint64_t b) {
    comp_mul_64_448_512(c,b,a);
}

STATIC INLINE void comp_mul_480_32_512(comp_uint512_t * RESTRICT c, comp_uint480_t a, comp_uint32_t b) {
    comp_mul_32_480_512(c,b,a);
}

/* External (dynamically linkable, non-static) versions of the functions */

comp_uint32_t comp_ext_uint32_to_comp_uint32(uint32_t a);
uint32_t comp_ext_uint32_from_comp_uint32(comp_uint32_t a);
comp_uint64_t comp_ext_uint64_to_comp_uint64(uint64_t a);
uint64_t comp_ext_uint64_from_comp_uint64(comp_uint64_t a);
comp_uint128_t comp_ext_uint128_to_comp_uint128(__uint128_t a);
__uint128_t comp_ext_uint128_from_comp_uint128(comp_uint128_t a);
void comp_ext_downconvert_high_64_32(comp_uint32_t * RESTRICT b, comp_uint64_t a);
void comp_ext_downconvert_high_96_32(comp_uint32_t * RESTRICT b, comp_uint96_t a);
void comp_ext_downconvert_high_128_32(comp_uint32_t * RESTRICT b, comp_uint128_t a);
void comp_ext_downconvert_high_160_32(comp_uint32_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_high_192_32(comp_uint32_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_high_224_32(comp_uint32_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_high_256_32(comp_uint32_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_288_32(comp_uint32_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_320_32(comp_uint32_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_352_32(comp_uint32_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_384_32(comp_uint32_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_416_32(comp_uint32_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_448_32(comp_uint32_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_480_32(comp_uint32_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_512_32(comp_uint32_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_96_64(comp_uint32_t * RESTRICT b, comp_uint96_t a);
void comp_ext_downconvert_high_128_64(comp_uint32_t * RESTRICT b, comp_uint128_t a);
void comp_ext_downconvert_high_160_64(comp_uint32_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_high_192_64(comp_uint32_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_high_224_64(comp_uint32_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_high_256_64(comp_uint32_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_288_64(comp_uint32_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_320_64(comp_uint32_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_352_64(comp_uint32_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_384_64(comp_uint32_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_416_64(comp_uint32_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_448_64(comp_uint32_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_480_64(comp_uint32_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_512_64(comp_uint32_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_128_96(comp_uint96_t * RESTRICT b, comp_uint128_t a);
void comp_ext_downconvert_high_160_96(comp_uint96_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_high_160_128(comp_uint128_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_high_192_96(comp_uint96_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_high_192_128(comp_uint128_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_high_192_160(comp_uint160_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_high_224_96(comp_uint96_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_high_224_128(comp_uint128_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_high_224_160(comp_uint160_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_high_224_192(comp_uint192_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_high_256_96(comp_uint96_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_256_128(comp_uint128_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_256_160(comp_uint160_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_256_192(comp_uint192_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_256_224(comp_uint224_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_high_288_96(comp_uint96_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_288_128(comp_uint128_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_288_160(comp_uint160_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_288_192(comp_uint192_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_288_224(comp_uint224_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_288_256(comp_uint256_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_high_320_96(comp_uint96_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_320_128(comp_uint128_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_320_160(comp_uint160_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_320_192(comp_uint192_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_320_224(comp_uint224_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_320_256(comp_uint256_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_320_288(comp_uint288_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_high_352_96(comp_uint96_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_128(comp_uint128_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_160(comp_uint160_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_192(comp_uint192_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_224(comp_uint224_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_256(comp_uint256_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_288(comp_uint288_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_352_320(comp_uint320_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_high_384_96(comp_uint96_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_128(comp_uint128_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_160(comp_uint160_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_192(comp_uint192_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_224(comp_uint224_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_256(comp_uint256_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_288(comp_uint288_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_320(comp_uint320_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_384_352(comp_uint352_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_high_416_96(comp_uint96_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_128(comp_uint128_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_160(comp_uint160_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_192(comp_uint192_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_224(comp_uint224_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_256(comp_uint256_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_288(comp_uint288_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_320(comp_uint320_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_352(comp_uint352_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_416_384(comp_uint384_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_high_448_96(comp_uint96_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_128(comp_uint128_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_160(comp_uint160_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_192(comp_uint192_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_224(comp_uint224_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_256(comp_uint256_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_288(comp_uint288_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_320(comp_uint320_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_352(comp_uint352_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_384(comp_uint384_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_448_416(comp_uint416_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_high_480_96(comp_uint96_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_128(comp_uint128_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_160(comp_uint160_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_192(comp_uint192_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_224(comp_uint224_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_256(comp_uint256_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_288(comp_uint288_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_320(comp_uint320_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_352(comp_uint352_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_384(comp_uint384_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_416(comp_uint416_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_480_448(comp_uint448_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_high_512_96(comp_uint96_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_128(comp_uint128_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_160(comp_uint160_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_192(comp_uint192_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_224(comp_uint224_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_256(comp_uint256_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_288(comp_uint288_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_320(comp_uint320_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_352(comp_uint352_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_384(comp_uint384_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_416(comp_uint416_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_448(comp_uint448_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_high_512_480(comp_uint480_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_64_32(comp_uint32_t * RESTRICT b, comp_uint64_t a);
void comp_ext_downconvert_low_96_32(comp_uint32_t * RESTRICT b, comp_uint96_t a);
void comp_ext_downconvert_low_128_32(comp_uint32_t * RESTRICT b, comp_uint128_t a);
void comp_ext_downconvert_low_160_32(comp_uint32_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_low_192_32(comp_uint32_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_low_224_32(comp_uint32_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_low_256_32(comp_uint32_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_288_32(comp_uint32_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_320_32(comp_uint32_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_352_32(comp_uint32_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_384_32(comp_uint32_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_416_32(comp_uint32_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_448_32(comp_uint32_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_480_32(comp_uint32_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_512_32(comp_uint32_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_96_64(comp_uint64_t * RESTRICT b, comp_uint96_t a);
void comp_ext_downconvert_low_128_64(comp_uint64_t * RESTRICT b, comp_uint128_t a);
void comp_ext_downconvert_low_160_64(comp_uint64_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_low_192_64(comp_uint64_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_low_224_64(comp_uint64_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_low_256_64(comp_uint64_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_288_64(comp_uint64_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_320_64(comp_uint64_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_352_64(comp_uint64_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_384_64(comp_uint64_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_416_64(comp_uint64_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_448_64(comp_uint64_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_480_64(comp_uint64_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_512_64(comp_uint64_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_128_96(comp_uint96_t * RESTRICT b, comp_uint128_t a);
void comp_ext_downconvert_low_160_96(comp_uint96_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_low_160_128(comp_uint128_t * RESTRICT b, comp_uint160_t a);
void comp_ext_downconvert_low_192_96(comp_uint96_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_low_192_128(comp_uint128_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_low_192_160(comp_uint160_t * RESTRICT b, comp_uint192_t a);
void comp_ext_downconvert_low_224_96(comp_uint96_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_low_224_128(comp_uint128_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_low_224_160(comp_uint160_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_low_224_192(comp_uint192_t * RESTRICT b, comp_uint224_t a);
void comp_ext_downconvert_low_256_96(comp_uint96_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_256_128(comp_uint128_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_256_160(comp_uint160_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_256_192(comp_uint192_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_256_224(comp_uint224_t * RESTRICT b, comp_uint256_t a);
void comp_ext_downconvert_low_288_96(comp_uint96_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_288_128(comp_uint128_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_288_160(comp_uint160_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_288_192(comp_uint192_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_288_224(comp_uint224_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_288_256(comp_uint256_t * RESTRICT b, comp_uint288_t a);
void comp_ext_downconvert_low_320_96(comp_uint96_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_320_128(comp_uint128_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_320_160(comp_uint160_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_320_192(comp_uint192_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_320_224(comp_uint224_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_320_256(comp_uint256_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_320_288(comp_uint288_t * RESTRICT b, comp_uint320_t a);
void comp_ext_downconvert_low_352_96(comp_uint96_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_128(comp_uint128_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_160(comp_uint160_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_192(comp_uint192_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_224(comp_uint224_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_256(comp_uint256_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_288(comp_uint288_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_352_320(comp_uint320_t * RESTRICT b, comp_uint352_t a);
void comp_ext_downconvert_low_384_96(comp_uint96_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_128(comp_uint128_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_160(comp_uint160_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_192(comp_uint192_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_224(comp_uint224_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_256(comp_uint256_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_288(comp_uint288_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_320(comp_uint320_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_384_352(comp_uint352_t * RESTRICT b, comp_uint384_t a);
void comp_ext_downconvert_low_416_96(comp_uint96_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_128(comp_uint128_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_160(comp_uint160_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_192(comp_uint192_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_224(comp_uint224_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_256(comp_uint256_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_288(comp_uint288_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_320(comp_uint320_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_352(comp_uint352_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_416_384(comp_uint384_t * RESTRICT b, comp_uint416_t a);
void comp_ext_downconvert_low_448_96(comp_uint96_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_128(comp_uint128_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_160(comp_uint160_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_192(comp_uint192_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_224(comp_uint224_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_256(comp_uint256_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_288(comp_uint288_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_320(comp_uint320_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_352(comp_uint352_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_384(comp_uint384_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_448_416(comp_uint416_t * RESTRICT b, comp_uint448_t a);
void comp_ext_downconvert_low_480_96(comp_uint96_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_128(comp_uint128_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_160(comp_uint160_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_192(comp_uint192_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_224(comp_uint224_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_256(comp_uint256_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_288(comp_uint288_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_320(comp_uint320_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_352(comp_uint352_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_384(comp_uint384_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_416(comp_uint416_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_480_448(comp_uint448_t * RESTRICT b, comp_uint480_t a);
void comp_ext_downconvert_low_512_96(comp_uint96_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_128(comp_uint128_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_160(comp_uint160_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_192(comp_uint192_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_224(comp_uint224_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_256(comp_uint256_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_288(comp_uint288_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_320(comp_uint320_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_352(comp_uint352_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_384(comp_uint384_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_416(comp_uint416_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_448(comp_uint448_t * RESTRICT b, comp_uint512_t a);
void comp_ext_downconvert_low_512_480(comp_uint480_t * RESTRICT b, comp_uint512_t a);
void comp_ext_upconvert_high_32_64(comp_uint64_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_96(comp_uint96_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_128(comp_uint128_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_160(comp_uint160_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_192(comp_uint192_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_224(comp_uint224_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_256(comp_uint256_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_288(comp_uint288_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_320(comp_uint320_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_352(comp_uint352_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_384(comp_uint384_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_416(comp_uint416_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_448(comp_uint448_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_480(comp_uint480_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_32_512(comp_uint512_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_high_64_96(comp_uint96_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_128(comp_uint128_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_160(comp_uint160_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_192(comp_uint192_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_224(comp_uint224_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_256(comp_uint256_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_288(comp_uint288_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_320(comp_uint320_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_352(comp_uint352_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_384(comp_uint384_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_416(comp_uint416_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_448(comp_uint448_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_480(comp_uint480_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_64_512(comp_uint512_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_high_96_128(comp_uint128_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_160(comp_uint160_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_192(comp_uint192_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_224(comp_uint224_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_256(comp_uint256_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_288(comp_uint288_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_320(comp_uint320_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_352(comp_uint352_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_384(comp_uint384_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_416(comp_uint416_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_448(comp_uint448_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_480(comp_uint480_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_96_512(comp_uint512_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_high_128_160(comp_uint160_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_192(comp_uint192_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_224(comp_uint224_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_256(comp_uint256_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_288(comp_uint288_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_320(comp_uint320_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_352(comp_uint352_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_384(comp_uint384_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_416(comp_uint416_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_448(comp_uint448_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_480(comp_uint480_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_128_512(comp_uint512_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_high_160_192(comp_uint192_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_224(comp_uint224_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_256(comp_uint256_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_288(comp_uint288_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_320(comp_uint320_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_352(comp_uint352_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_384(comp_uint384_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_416(comp_uint416_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_448(comp_uint448_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_480(comp_uint480_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_160_512(comp_uint512_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_high_192_224(comp_uint224_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_256(comp_uint256_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_288(comp_uint288_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_320(comp_uint320_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_352(comp_uint352_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_384(comp_uint384_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_416(comp_uint416_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_448(comp_uint448_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_480(comp_uint480_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_192_512(comp_uint512_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_high_224_256(comp_uint256_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_288(comp_uint288_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_320(comp_uint320_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_352(comp_uint352_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_384(comp_uint384_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_416(comp_uint416_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_448(comp_uint448_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_480(comp_uint480_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_224_512(comp_uint512_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_high_256_288(comp_uint288_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_320(comp_uint320_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_352(comp_uint352_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_384(comp_uint384_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_416(comp_uint416_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_448(comp_uint448_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_480(comp_uint480_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_256_512(comp_uint512_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_high_288_320(comp_uint320_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_288_352(comp_uint352_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_288_384(comp_uint384_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_288_416(comp_uint416_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_288_448(comp_uint448_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_288_480(comp_uint480_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_288_512(comp_uint512_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_high_320_352(comp_uint352_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_high_320_384(comp_uint384_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_high_320_416(comp_uint416_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_high_320_448(comp_uint448_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_high_320_480(comp_uint480_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_high_320_512(comp_uint512_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_high_352_384(comp_uint384_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_high_352_416(comp_uint416_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_high_352_448(comp_uint448_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_high_352_480(comp_uint480_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_high_352_512(comp_uint512_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_high_384_416(comp_uint416_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_high_384_448(comp_uint448_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_high_384_480(comp_uint480_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_high_384_512(comp_uint512_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_high_416_448(comp_uint448_t * RESTRICT b, comp_uint416_t a);
void comp_ext_upconvert_high_416_480(comp_uint480_t * RESTRICT b, comp_uint416_t a);
void comp_ext_upconvert_high_416_512(comp_uint512_t * RESTRICT b, comp_uint416_t a);
void comp_ext_upconvert_high_448_480(comp_uint480_t * RESTRICT b, comp_uint448_t a);
void comp_ext_upconvert_high_448_512(comp_uint512_t * RESTRICT b, comp_uint448_t a);
void comp_ext_upconvert_high_480_512(comp_uint512_t * RESTRICT b, comp_uint480_t a);
void comp_ext_upconvert_low_32_64(comp_uint64_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_96(comp_uint96_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_128(comp_uint128_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_160(comp_uint160_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_192(comp_uint192_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_224(comp_uint224_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_256(comp_uint256_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_288(comp_uint288_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_320(comp_uint320_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_352(comp_uint352_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_384(comp_uint384_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_416(comp_uint416_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_448(comp_uint448_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_480(comp_uint480_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_32_512(comp_uint512_t * RESTRICT b, comp_uint32_t a);
void comp_ext_upconvert_low_64_96(comp_uint96_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_128(comp_uint128_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_160(comp_uint160_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_192(comp_uint192_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_224(comp_uint224_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_256(comp_uint256_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_288(comp_uint288_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_320(comp_uint320_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_352(comp_uint352_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_384(comp_uint384_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_416(comp_uint416_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_448(comp_uint448_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_480(comp_uint480_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_64_512(comp_uint512_t * RESTRICT b, comp_uint64_t a);
void comp_ext_upconvert_low_96_128(comp_uint128_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_160(comp_uint160_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_192(comp_uint192_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_224(comp_uint224_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_256(comp_uint256_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_288(comp_uint288_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_320(comp_uint320_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_352(comp_uint352_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_384(comp_uint384_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_416(comp_uint416_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_448(comp_uint448_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_480(comp_uint480_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_96_512(comp_uint512_t * RESTRICT b, comp_uint96_t a);
void comp_ext_upconvert_low_128_160(comp_uint160_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_192(comp_uint192_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_224(comp_uint224_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_256(comp_uint256_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_288(comp_uint288_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_320(comp_uint320_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_352(comp_uint352_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_384(comp_uint384_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_416(comp_uint416_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_448(comp_uint448_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_480(comp_uint480_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_128_512(comp_uint512_t * RESTRICT b, comp_uint128_t a);
void comp_ext_upconvert_low_160_192(comp_uint192_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_224(comp_uint224_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_256(comp_uint256_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_288(comp_uint288_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_320(comp_uint320_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_352(comp_uint352_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_384(comp_uint384_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_416(comp_uint416_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_448(comp_uint448_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_480(comp_uint480_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_160_512(comp_uint512_t * RESTRICT b, comp_uint160_t a);
void comp_ext_upconvert_low_192_224(comp_uint224_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_256(comp_uint256_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_288(comp_uint288_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_320(comp_uint320_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_352(comp_uint352_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_384(comp_uint384_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_416(comp_uint416_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_448(comp_uint448_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_480(comp_uint480_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_192_512(comp_uint512_t * RESTRICT b, comp_uint192_t a);
void comp_ext_upconvert_low_224_256(comp_uint256_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_288(comp_uint288_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_320(comp_uint320_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_352(comp_uint352_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_384(comp_uint384_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_416(comp_uint416_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_448(comp_uint448_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_480(comp_uint480_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_224_512(comp_uint512_t * RESTRICT b, comp_uint224_t a);
void comp_ext_upconvert_low_256_288(comp_uint288_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_320(comp_uint320_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_352(comp_uint352_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_384(comp_uint384_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_416(comp_uint416_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_448(comp_uint448_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_480(comp_uint480_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_256_512(comp_uint512_t * RESTRICT b, comp_uint256_t a);
void comp_ext_upconvert_low_288_320(comp_uint320_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_288_352(comp_uint352_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_288_384(comp_uint384_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_288_416(comp_uint416_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_288_448(comp_uint448_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_288_480(comp_uint480_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_288_512(comp_uint512_t * RESTRICT b, comp_uint288_t a);
void comp_ext_upconvert_low_320_352(comp_uint352_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_low_320_384(comp_uint384_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_low_320_416(comp_uint416_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_low_320_448(comp_uint448_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_low_320_480(comp_uint480_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_low_320_512(comp_uint512_t * RESTRICT b, comp_uint320_t a);
void comp_ext_upconvert_low_352_384(comp_uint384_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_low_352_416(comp_uint416_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_low_352_448(comp_uint448_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_low_352_480(comp_uint480_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_low_352_512(comp_uint512_t * RESTRICT b, comp_uint352_t a);
void comp_ext_upconvert_low_384_416(comp_uint416_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_low_384_448(comp_uint448_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_low_384_480(comp_uint480_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_low_384_512(comp_uint512_t * RESTRICT b, comp_uint384_t a);
void comp_ext_upconvert_low_416_448(comp_uint448_t * RESTRICT b, comp_uint416_t a);
void comp_ext_upconvert_low_416_480(comp_uint480_t * RESTRICT b, comp_uint416_t a);
void comp_ext_upconvert_low_416_512(comp_uint512_t * RESTRICT b, comp_uint416_t a);
void comp_ext_upconvert_low_448_480(comp_uint480_t * RESTRICT b, comp_uint448_t a);
void comp_ext_upconvert_low_448_512(comp_uint512_t * RESTRICT b, comp_uint448_t a);
void comp_ext_upconvert_low_480_512(comp_uint512_t * RESTRICT b, comp_uint480_t a);
void comp_ext_shift_right_32(comp_uint32_t * RESTRICT b, comp_uint32_t a, unsigned int k);
void comp_ext_shift_right_64(comp_uint64_t * RESTRICT b, comp_uint64_t a, unsigned int k);
void comp_ext_shift_right_96(comp_uint96_t * RESTRICT b, comp_uint96_t a, unsigned int k);
void comp_ext_shift_right_128(comp_uint128_t * RESTRICT b, comp_uint128_t a, unsigned int k);
void comp_ext_shift_right_160(comp_uint160_t * RESTRICT b, comp_uint160_t a, unsigned int k);
void comp_ext_shift_right_192(comp_uint192_t * RESTRICT b, comp_uint192_t a, unsigned int k);
void comp_ext_shift_right_224(comp_uint224_t * RESTRICT b, comp_uint224_t a, unsigned int k);
void comp_ext_shift_right_256(comp_uint256_t * RESTRICT b, comp_uint256_t a, unsigned int k);
void comp_ext_shift_right_288(comp_uint288_t * RESTRICT b, comp_uint288_t a, unsigned int k);
void comp_ext_shift_right_320(comp_uint320_t * RESTRICT b, comp_uint320_t a, unsigned int k);
void comp_ext_shift_right_352(comp_uint352_t * RESTRICT b, comp_uint352_t a, unsigned int k);
void comp_ext_shift_right_384(comp_uint384_t * RESTRICT b, comp_uint384_t a, unsigned int k);
void comp_ext_shift_right_416(comp_uint416_t * RESTRICT b, comp_uint416_t a, unsigned int k);
void comp_ext_shift_right_448(comp_uint448_t * RESTRICT b, comp_uint448_t a, unsigned int k);
void comp_ext_shift_right_480(comp_uint480_t * RESTRICT b, comp_uint480_t a, unsigned int k);
void comp_ext_shift_right_512(comp_uint512_t * RESTRICT b, comp_uint512_t a, unsigned int k);
void comp_ext_shift_left_32(comp_uint32_t * RESTRICT b, comp_uint32_t a, unsigned int k);
void comp_ext_shift_left_64(comp_uint64_t * RESTRICT b, comp_uint64_t a, unsigned int k);
void comp_ext_shift_left_96(comp_uint96_t * RESTRICT b, comp_uint96_t a, unsigned int k);
void comp_ext_shift_left_128(comp_uint128_t * RESTRICT b, comp_uint128_t a, unsigned int k);
void comp_ext_shift_left_160(comp_uint160_t * RESTRICT b, comp_uint160_t a, unsigned int k);
void comp_ext_shift_left_192(comp_uint192_t * RESTRICT b, comp_uint192_t a, unsigned int k);
void comp_ext_shift_left_224(comp_uint224_t * RESTRICT b, comp_uint224_t a, unsigned int k);
void comp_ext_shift_left_256(comp_uint256_t * RESTRICT b, comp_uint256_t a, unsigned int k);
void comp_ext_shift_left_288(comp_uint288_t * RESTRICT b, comp_uint288_t a, unsigned int k);
void comp_ext_shift_left_320(comp_uint320_t * RESTRICT b, comp_uint320_t a, unsigned int k);
void comp_ext_shift_left_352(comp_uint352_t * RESTRICT b, comp_uint352_t a, unsigned int k);
void comp_ext_shift_left_384(comp_uint384_t * RESTRICT b, comp_uint384_t a, unsigned int k);
void comp_ext_shift_left_416(comp_uint416_t * RESTRICT b, comp_uint416_t a, unsigned int k);
void comp_ext_shift_left_448(comp_uint448_t * RESTRICT b, comp_uint448_t a, unsigned int k);
void comp_ext_shift_left_480(comp_uint480_t * RESTRICT b, comp_uint480_t a, unsigned int k);
void comp_ext_shift_left_512(comp_uint512_t * RESTRICT b, comp_uint512_t a, unsigned int k);
unsigned int comp_ext_leading_zeros_64(comp_uint64_t a);
unsigned int comp_ext_leading_zeros_128(comp_uint128_t a);
unsigned int comp_ext_leading_zeros_192(comp_uint192_t a);
unsigned int comp_ext_leading_zeros_256(comp_uint256_t a);
unsigned int comp_ext_leading_zeros_320(comp_uint320_t a);
unsigned int comp_ext_leading_zeros_384(comp_uint384_t a);
unsigned int comp_ext_leading_zeros_448(comp_uint448_t a);
unsigned int comp_ext_leading_zeros_512(comp_uint512_t a);
int comp_ext_sign_32(comp_uint32_t a);
int comp_ext_sign_64(comp_uint64_t a);
int comp_ext_sign_96(comp_uint96_t a);
int comp_ext_sign_128(comp_uint128_t a);
int comp_ext_sign_160(comp_uint160_t a);
int comp_ext_sign_192(comp_uint192_t a);
int comp_ext_sign_224(comp_uint224_t a);
int comp_ext_sign_256(comp_uint256_t a);
int comp_ext_sign_288(comp_uint288_t a);
int comp_ext_sign_320(comp_uint320_t a);
int comp_ext_sign_352(comp_uint352_t a);
int comp_ext_sign_384(comp_uint384_t a);
int comp_ext_sign_416(comp_uint416_t a);
int comp_ext_sign_448(comp_uint448_t a);
int comp_ext_sign_480(comp_uint480_t a);
int comp_ext_sign_512(comp_uint512_t a);
void comp_ext_add_32(comp_uint32_t * RESTRICT c, comp_uint32_t a, comp_uint32_t b);
void comp_ext_add_64(comp_uint64_t * RESTRICT c, comp_uint64_t a, comp_uint64_t b);
void comp_ext_add_96(comp_uint96_t * RESTRICT c, comp_uint96_t a, comp_uint96_t b);
void comp_ext_add_128(comp_uint128_t * RESTRICT c, comp_uint128_t a, comp_uint128_t b);
void comp_ext_add_160(comp_uint160_t * RESTRICT c, comp_uint160_t a, comp_uint160_t b);
void comp_ext_add_192(comp_uint192_t * RESTRICT c, comp_uint192_t a, comp_uint192_t b);
void comp_ext_add_224(comp_uint224_t * RESTRICT c, comp_uint224_t a, comp_uint224_t b);
void comp_ext_add_256(comp_uint256_t * RESTRICT c, comp_uint256_t a, comp_uint256_t b);
void comp_ext_add_288(comp_uint288_t * RESTRICT c, comp_uint288_t a, comp_uint288_t b);
void comp_ext_add_320(comp_uint320_t * RESTRICT c, comp_uint320_t a, comp_uint320_t b);
void comp_ext_add_352(comp_uint352_t * RESTRICT c, comp_uint352_t a, comp_uint352_t b);
void comp_ext_add_384(comp_uint384_t * RESTRICT c, comp_uint384_t a, comp_uint384_t b);
void comp_ext_add_416(comp_uint416_t * RESTRICT c, comp_uint416_t a, comp_uint416_t b);
void comp_ext_add_448(comp_uint448_t * RESTRICT c, comp_uint448_t a, comp_uint448_t b);
void comp_ext_add_480(comp_uint480_t * RESTRICT c, comp_uint480_t a, comp_uint480_t b);
void comp_ext_add_512(comp_uint512_t * RESTRICT c, comp_uint512_t a, comp_uint512_t b);
void comp_ext_sub_32(comp_uint32_t * RESTRICT c, comp_uint32_t a, comp_uint32_t b);
void comp_ext_sub_64(comp_uint64_t * RESTRICT c, comp_uint64_t a, comp_uint64_t b);
void comp_ext_sub_96(comp_uint96_t * RESTRICT c, comp_uint96_t a, comp_uint96_t b);
void comp_ext_sub_128(comp_uint128_t * RESTRICT c, comp_uint128_t a, comp_uint128_t b);
void comp_ext_sub_160(comp_uint160_t * RESTRICT c, comp_uint160_t a, comp_uint160_t b);
void comp_ext_sub_192(comp_uint192_t * RESTRICT c, comp_uint192_t a, comp_uint192_t b);
void comp_ext_sub_224(comp_uint224_t * RESTRICT c, comp_uint224_t a, comp_uint224_t b);
void comp_ext_sub_256(comp_uint256_t * RESTRICT c, comp_uint256_t a, comp_uint256_t b);
void comp_ext_sub_288(comp_uint288_t * RESTRICT c, comp_uint288_t a, comp_uint288_t b);
void comp_ext_sub_320(comp_uint320_t * RESTRICT c, comp_uint320_t a, comp_uint320_t b);
void comp_ext_sub_352(comp_uint352_t * RESTRICT c, comp_uint352_t a, comp_uint352_t b);
void comp_ext_sub_384(comp_uint384_t * RESTRICT c, comp_uint384_t a, comp_uint384_t b);
void comp_ext_sub_416(comp_uint416_t * RESTRICT c, comp_uint416_t a, comp_uint416_t b);
void comp_ext_sub_448(comp_uint448_t * RESTRICT c, comp_uint448_t a, comp_uint448_t b);
void comp_ext_sub_480(comp_uint480_t * RESTRICT c, comp_uint480_t a, comp_uint480_t b);
void comp_ext_sub_512(comp_uint512_t * RESTRICT c, comp_uint512_t a, comp_uint512_t b);
void comp_ext_mul_32_32_64(comp_uint64_t * RESTRICT c, comp_uint32_t a, comp_uint32_t b);
void comp_ext_mul_32_64_96(comp_uint96_t * RESTRICT c, comp_uint32_t a, comp_uint64_t b);
void comp_ext_mul_32_96_128(comp_uint128_t * RESTRICT c, comp_uint32_t a, comp_uint96_t b);
void comp_ext_mul_32_128_160(comp_uint160_t * RESTRICT c, comp_uint32_t a, comp_uint128_t b);
void comp_ext_mul_32_160_192(comp_uint192_t * RESTRICT c, comp_uint32_t a, comp_uint160_t b);
void comp_ext_mul_32_192_224(comp_uint224_t * RESTRICT c, comp_uint32_t a, comp_uint192_t b);
void comp_ext_mul_32_224_256(comp_uint256_t * RESTRICT c, comp_uint32_t a, comp_uint224_t b);
void comp_ext_mul_32_256_288(comp_uint288_t * RESTRICT c, comp_uint32_t a, comp_uint256_t b);
void comp_ext_mul_32_288_320(comp_uint320_t * RESTRICT c, comp_uint32_t a, comp_uint288_t b);
void comp_ext_mul_32_320_352(comp_uint352_t * RESTRICT c, comp_uint32_t a, comp_uint320_t b);
void comp_ext_mul_32_352_384(comp_uint384_t * RESTRICT c, comp_uint32_t a, comp_uint352_t b);
void comp_ext_mul_32_384_416(comp_uint416_t * RESTRICT c, comp_uint32_t a, comp_uint384_t b);
void comp_ext_mul_32_416_448(comp_uint448_t * RESTRICT c, comp_uint32_t a, comp_uint416_t b);
void comp_ext_mul_32_448_480(comp_uint480_t * RESTRICT c, comp_uint32_t a, comp_uint448_t b);
void comp_ext_mul_32_480_512(comp_uint512_t * RESTRICT c, comp_uint32_t a, comp_uint480_t b);
void comp_ext_mul_64_32_96(comp_uint96_t * RESTRICT c, comp_uint64_t a, comp_uint32_t b);
void comp_ext_mul_64_64_128(comp_uint128_t * RESTRICT c, comp_uint64_t a, comp_uint64_t b);
void comp_ext_mul_64_96_160(comp_uint160_t * RESTRICT c, comp_uint64_t a, comp_uint96_t b);
void comp_ext_mul_64_128_192(comp_uint192_t * RESTRICT c, comp_uint64_t a, comp_uint128_t b);
void comp_ext_mul_64_160_224(comp_uint224_t * RESTRICT c, comp_uint64_t a, comp_uint160_t b);
void comp_ext_mul_64_192_256(comp_uint256_t * RESTRICT c, comp_uint64_t a, comp_uint192_t b);
void comp_ext_mul_64_224_288(comp_uint288_t * RESTRICT c, comp_uint64_t a, comp_uint224_t b);
void comp_ext_mul_64_256_320(comp_uint320_t * RESTRICT c, comp_uint64_t a, comp_uint256_t b);
void comp_ext_mul_64_288_352(comp_uint352_t * RESTRICT c, comp_uint64_t a, comp_uint288_t b);
void comp_ext_mul_64_320_384(comp_uint384_t * RESTRICT c, comp_uint64_t a, comp_uint320_t b);
void comp_ext_mul_64_352_416(comp_uint416_t * RESTRICT c, comp_uint64_t a, comp_uint352_t b);
void comp_ext_mul_64_384_448(comp_uint448_t * RESTRICT c, comp_uint64_t a, comp_uint384_t b);
void comp_ext_mul_64_416_480(comp_uint480_t * RESTRICT c, comp_uint64_t a, comp_uint416_t b);
void comp_ext_mul_64_448_512(comp_uint512_t * RESTRICT c, comp_uint64_t a, comp_uint448_t b);
void comp_ext_mul_96_32_128(comp_uint128_t * RESTRICT c, comp_uint96_t a, comp_uint32_t b);
void comp_ext_mul_96_64_160(comp_uint160_t * RESTRICT c, comp_uint96_t a, comp_uint64_t b);
void comp_ext_mul_96_96_192(comp_uint192_t * RESTRICT c, comp_uint96_t a, comp_uint96_t b);
void comp_ext_mul_96_128_224(comp_uint224_t * RESTRICT c, comp_uint96_t a, comp_uint128_t b);
void comp_ext_mul_96_160_256(comp_uint256_t * RESTRICT c, comp_uint96_t a, comp_uint160_t b);
void comp_ext_mul_96_192_288(comp_uint288_t * RESTRICT c, comp_uint96_t a, comp_uint192_t b);
void comp_ext_mul_96_224_320(comp_uint320_t * RESTRICT c, comp_uint96_t a, comp_uint224_t b);
void comp_ext_mul_96_256_352(comp_uint352_t * RESTRICT c, comp_uint96_t a, comp_uint256_t b);
void comp_ext_mul_96_288_384(comp_uint384_t * RESTRICT c, comp_uint96_t a, comp_uint288_t b);
void comp_ext_mul_96_320_416(comp_uint416_t * RESTRICT c, comp_uint96_t a, comp_uint320_t b);
void comp_ext_mul_96_352_448(comp_uint448_t * RESTRICT c, comp_uint96_t a, comp_uint352_t b);
void comp_ext_mul_96_384_480(comp_uint480_t * RESTRICT c, comp_uint96_t a, comp_uint384_t b);
void comp_ext_mul_96_416_512(comp_uint512_t * RESTRICT c, comp_uint96_t a, comp_uint416_t b);
void comp_ext_mul_128_32_160(comp_uint160_t * RESTRICT c, comp_uint128_t a, comp_uint32_t b);
void comp_ext_mul_128_64_192(comp_uint192_t * RESTRICT c, comp_uint128_t a, comp_uint64_t b);
void comp_ext_mul_128_96_224(comp_uint224_t * RESTRICT c, comp_uint128_t a, comp_uint96_t b);
void comp_ext_mul_128_128_256(comp_uint256_t * RESTRICT c, comp_uint128_t a, comp_uint128_t b);
void comp_ext_mul_128_160_288(comp_uint288_t * RESTRICT c, comp_uint128_t a, comp_uint160_t b);
void comp_ext_mul_128_192_320(comp_uint320_t * RESTRICT c, comp_uint128_t a, comp_uint192_t b);
void comp_ext_mul_128_224_352(comp_uint352_t * RESTRICT c, comp_uint128_t a, comp_uint224_t b);
void comp_ext_mul_128_256_384(comp_uint384_t * RESTRICT c, comp_uint128_t a, comp_uint256_t b);
void comp_ext_mul_128_288_416(comp_uint416_t * RESTRICT c, comp_uint128_t a, comp_uint288_t b);
void comp_ext_mul_128_320_448(comp_uint448_t * RESTRICT c, comp_uint128_t a, comp_uint320_t b);
void comp_ext_mul_128_352_480(comp_uint480_t * RESTRICT c, comp_uint128_t a, comp_uint352_t b);
void comp_ext_mul_128_384_512(comp_uint512_t * RESTRICT c, comp_uint128_t a, comp_uint384_t b);
void comp_ext_mul_160_32_192(comp_uint192_t * RESTRICT c, comp_uint160_t a, comp_uint32_t b);
void comp_ext_mul_160_64_224(comp_uint224_t * RESTRICT c, comp_uint160_t a, comp_uint64_t b);
void comp_ext_mul_160_96_256(comp_uint256_t * RESTRICT c, comp_uint160_t a, comp_uint96_t b);
void comp_ext_mul_160_128_288(comp_uint288_t * RESTRICT c, comp_uint160_t a, comp_uint128_t b);
void comp_ext_mul_160_160_320(comp_uint320_t * RESTRICT c, comp_uint160_t a, comp_uint160_t b);
void comp_ext_mul_160_192_352(comp_uint352_t * RESTRICT c, comp_uint160_t a, comp_uint192_t b);
void comp_ext_mul_160_224_384(comp_uint384_t * RESTRICT c, comp_uint160_t a, comp_uint224_t b);
void comp_ext_mul_160_256_416(comp_uint416_t * RESTRICT c, comp_uint160_t a, comp_uint256_t b);
void comp_ext_mul_160_288_448(comp_uint448_t * RESTRICT c, comp_uint160_t a, comp_uint288_t b);
void comp_ext_mul_160_320_480(comp_uint480_t * RESTRICT c, comp_uint160_t a, comp_uint320_t b);
void comp_ext_mul_160_352_512(comp_uint512_t * RESTRICT c, comp_uint160_t a, comp_uint352_t b);
void comp_ext_mul_192_32_224(comp_uint224_t * RESTRICT c, comp_uint192_t a, comp_uint32_t b);
void comp_ext_mul_192_64_256(comp_uint256_t * RESTRICT c, comp_uint192_t a, comp_uint64_t b);
void comp_ext_mul_192_96_288(comp_uint288_t * RESTRICT c, comp_uint192_t a, comp_uint96_t b);
void comp_ext_mul_192_128_320(comp_uint320_t * RESTRICT c, comp_uint192_t a, comp_uint128_t b);
void comp_ext_mul_192_160_352(comp_uint352_t * RESTRICT c, comp_uint192_t a, comp_uint160_t b);
void comp_ext_mul_192_192_384(comp_uint384_t * RESTRICT c, comp_uint192_t a, comp_uint192_t b);
void comp_ext_mul_192_224_416(comp_uint416_t * RESTRICT c, comp_uint192_t a, comp_uint224_t b);
void comp_ext_mul_192_256_448(comp_uint448_t * RESTRICT c, comp_uint192_t a, comp_uint256_t b);
void comp_ext_mul_192_288_480(comp_uint480_t * RESTRICT c, comp_uint192_t a, comp_uint288_t b);
void comp_ext_mul_192_320_512(comp_uint512_t * RESTRICT c, comp_uint192_t a, comp_uint320_t b);
void comp_ext_mul_224_32_256(comp_uint256_t * RESTRICT c, comp_uint224_t a, comp_uint32_t b);
void comp_ext_mul_224_64_288(comp_uint288_t * RESTRICT c, comp_uint224_t a, comp_uint64_t b);
void comp_ext_mul_224_96_320(comp_uint320_t * RESTRICT c, comp_uint224_t a, comp_uint96_t b);
void comp_ext_mul_224_128_352(comp_uint352_t * RESTRICT c, comp_uint224_t a, comp_uint128_t b);
void comp_ext_mul_224_160_384(comp_uint384_t * RESTRICT c, comp_uint224_t a, comp_uint160_t b);
void comp_ext_mul_224_192_416(comp_uint416_t * RESTRICT c, comp_uint224_t a, comp_uint192_t b);
void comp_ext_mul_224_224_448(comp_uint448_t * RESTRICT c, comp_uint224_t a, comp_uint224_t b);
void comp_ext_mul_224_256_480(comp_uint480_t * RESTRICT c, comp_uint224_t a, comp_uint256_t b);
void comp_ext_mul_224_288_512(comp_uint512_t * RESTRICT c, comp_uint224_t a, comp_uint288_t b);
void comp_ext_mul_256_32_288(comp_uint288_t * RESTRICT c, comp_uint256_t a, comp_uint32_t b);
void comp_ext_mul_256_64_320(comp_uint320_t * RESTRICT c, comp_uint256_t a, comp_uint64_t b);
void comp_ext_mul_256_96_352(comp_uint352_t * RESTRICT c, comp_uint256_t a, comp_uint96_t b);
void comp_ext_mul_256_128_384(comp_uint384_t * RESTRICT c, comp_uint256_t a, comp_uint128_t b);
void comp_ext_mul_256_160_416(comp_uint416_t * RESTRICT c, comp_uint256_t a, comp_uint160_t b);
void comp_ext_mul_256_192_448(comp_uint448_t * RESTRICT c, comp_uint256_t a, comp_uint192_t b);
void comp_ext_mul_256_224_480(comp_uint480_t * RESTRICT c, comp_uint256_t a, comp_uint224_t b);
void comp_ext_mul_256_256_512(comp_uint512_t * RESTRICT c, comp_uint256_t a, comp_uint256_t b);
void comp_ext_mul_288_32_320(comp_uint320_t * RESTRICT c, comp_uint288_t a, comp_uint32_t b);
void comp_ext_mul_288_64_352(comp_uint352_t * RESTRICT c, comp_uint288_t a, comp_uint64_t b);
void comp_ext_mul_288_96_384(comp_uint384_t * RESTRICT c, comp_uint288_t a, comp_uint96_t b);
void comp_ext_mul_288_128_416(comp_uint416_t * RESTRICT c, comp_uint288_t a, comp_uint128_t b);
void comp_ext_mul_288_160_448(comp_uint448_t * RESTRICT c, comp_uint288_t a, comp_uint160_t b);
void comp_ext_mul_288_192_480(comp_uint480_t * RESTRICT c, comp_uint288_t a, comp_uint192_t b);
void comp_ext_mul_288_224_512(comp_uint512_t * RESTRICT c, comp_uint288_t a, comp_uint224_t b);
void comp_ext_mul_320_32_352(comp_uint352_t * RESTRICT c, comp_uint320_t a, comp_uint32_t b);
void comp_ext_mul_320_64_384(comp_uint384_t * RESTRICT c, comp_uint320_t a, comp_uint64_t b);
void comp_ext_mul_320_96_416(comp_uint416_t * RESTRICT c, comp_uint320_t a, comp_uint96_t b);
void comp_ext_mul_320_128_448(comp_uint448_t * RESTRICT c, comp_uint320_t a, comp_uint128_t b);
void comp_ext_mul_320_160_480(comp_uint480_t * RESTRICT c, comp_uint320_t a, comp_uint160_t b);
void comp_ext_mul_320_192_512(comp_uint512_t * RESTRICT c, comp_uint320_t a, comp_uint192_t b);
void comp_ext_mul_352_32_384(comp_uint384_t * RESTRICT c, comp_uint352_t a, comp_uint32_t b);
void comp_ext_mul_352_64_416(comp_uint416_t * RESTRICT c, comp_uint352_t a, comp_uint64_t b);
void comp_ext_mul_352_96_448(comp_uint448_t * RESTRICT c, comp_uint352_t a, comp_uint96_t b);
void comp_ext_mul_352_128_480(comp_uint480_t * RESTRICT c, comp_uint352_t a, comp_uint128_t b);
void comp_ext_mul_352_160_512(comp_uint512_t * RESTRICT c, comp_uint352_t a, comp_uint160_t b);
void comp_ext_mul_384_32_416(comp_uint416_t * RESTRICT c, comp_uint384_t a, comp_uint32_t b);
void comp_ext_mul_384_64_448(comp_uint448_t * RESTRICT c, comp_uint384_t a, comp_uint64_t b);
void comp_ext_mul_384_96_480(comp_uint480_t * RESTRICT c, comp_uint384_t a, comp_uint96_t b);
void comp_ext_mul_384_128_512(comp_uint512_t * RESTRICT c, comp_uint384_t a, comp_uint128_t b);
void comp_ext_mul_416_32_448(comp_uint448_t * RESTRICT c, comp_uint416_t a, comp_uint32_t b);
void comp_ext_mul_416_64_480(comp_uint480_t * RESTRICT c, comp_uint416_t a, comp_uint64_t b);
void comp_ext_mul_416_96_512(comp_uint512_t * RESTRICT c, comp_uint416_t a, comp_uint96_t b);
void comp_ext_mul_448_32_480(comp_uint480_t * RESTRICT c, comp_uint448_t a, comp_uint32_t b);
void comp_ext_mul_448_64_512(comp_uint512_t * RESTRICT c, comp_uint448_t a, comp_uint64_t b);
void comp_ext_mul_480_32_512(comp_uint512_t * RESTRICT c, comp_uint480_t a, comp_uint32_t b);

#endif
